Package: eks
Version: 1.1.3
Date: 2026-05-10
Title: Tidy and Geospatial Kernel Smoothing
Authors@R: c(person("Tarn", "Duong", role=c("aut","cre"), email="tarn.duong@gmail.com", comment=c(ORCID="0000-0002-1198-3482"))) 
Maintainer: Tarn Duong <tarn.duong@gmail.com>
Depends: R (>= 2.10.0)
Imports: colorspace, dplyr, ggplot2 (>= 3.0.0), isoband, ks (>=
        1.15.0), sf
Suggests: ggquiver, knitr, mapsf, MASS, rmarkdown
VignetteBuilder: knitr
Encoding: UTF-8
Description: Extensions of the kernel smoothing functions from the 'ks' package for compatibility with the tidyverse and geospatial ecosystems <doi:10.1007/s00180-024-01543-9>.
License: GPL-2 | GPL-3
URL: https://mvstat.net/eks/, https://mvstat.net/mvksa/,
        https://github.com/cran/eks/
NeedsCompilation: no
Packaged: 2026-05-10 10:48:15 UTC; tduong
Author: Tarn Duong [aut, cre] (ORCID: <https://orcid.org/0000-0002-1198-3482>)
Repository: CRAN
Date/Publication: 2026-05-10 11:40:03 UTC
Built: R 4.6.0; ; 2026-05-10 13:25:39 UTC; unix

Package: eks
Version: 1.1.3
Date: 2026-05-10
Title: Tidy and Geospatial Kernel Smoothing
Authors@R: c(person("Tarn", "Duong", role=c("aut","cre"), email="tarn.duong@gmail.com", comment=c(ORCID="0000-0002-1198-3482"))) 
Maintainer: Tarn Duong <tarn.duong@gmail.com>
Depends: R (>= 2.10.0)
Imports: colorspace, dplyr, ggplot2 (>= 3.0.0), isoband, ks (>=
        1.15.0), sf
Suggests: ggquiver, knitr, mapsf, MASS, rmarkdown
VignetteBuilder: knitr
Encoding: UTF-8
Description: Extensions of the kernel smoothing functions from the 'ks' package for compatibility with the tidyverse and geospatial ecosystems <doi:10.1007/s00180-024-01543-9>.
License: GPL-2 | GPL-3
URL: https://mvstat.net/eks/, https://mvstat.net/mvksa/,
        https://github.com/cran/eks/
NeedsCompilation: no
Packaged: 2026-05-10 10:48:15 UTC; tduong
Author: Tarn Duong [aut, cre] (ORCID: <https://orcid.org/0000-0002-1198-3482>)
Repository: CRAN
Date/Publication: 2026-05-10 11:40:03 UTC

importFrom(colorspace, diverging_hcl, qualitative_hcl, scale_colour_discrete_sequential, scale_colour_discrete_qualitative, scale_fill_discrete_diverging, sequential_hcl)
importFrom(dplyr, .data, arrange, as_tibble, enquo, group_by, group_indices, filter, group_modify, groups, left_join, mutate, one_of, pull, relocate, rename, select, slice_head, summarise, sym, tibble, ungroup, vars)
importFrom(ggplot2, ggplot, Geom, GeomContour, GeomContourFilled, GeomPoint, GeomRug, StatContour, StatContourFilled, aes, ggplot_add, guide_legend, guides, labs, scale_colour_manual, scale_colour_viridis_d, scale_fill_manual, scale_fill_viridis_d, standardise_aes_names, unit)
importFrom(geos, geos_area, geos_length, geos_intersects)
importFrom(graphics, image)
importFrom(grDevices, contourLines)
importFrom(isoband, isobands, isolines)
importFrom(ks, as.kde, contourBreaks, contourLevels, kcde, kcopula, kcurv, kda, kdcde, kde, kde.balloon, kde.boundary, kde.local.test, kde.sp, kde.test, kde.truncate, kdde.truncate, kdde, kdr, kfe, kfs, kms, kms.part, kroc, ksupp)
importFrom(lwgeom, st_split)
importFrom(sf, st_cast, st_as_sfc, st_as_sf, st_coordinates, st_crs, st_difference, st_drop_geometry, st_geometry, st_is, st_set_crs, st_sf, st_sfc)
importFrom(stats, cutree, dist, hclust, median, na.omit, predict, setNames, weighted.mean)
importFrom(tools, toTitleCase)
importFrom(utils, data, head, tail)

export(add_contour_breaks)
export(aes_ks, guides_ks, labs_ks)
export(as_tidy_ks)
export(dim_ks)
export(contour_breaks)
export(geom_contour_filled_ks, stat_contour_filled_ks)
export(geom_contour_ks, stat_contour_ks)            
export(geom_point_ks, stat_point_ks)
export(geom_rug_ks, stat_rug_ks)
export(GeomContourKs, StatContourKs)
export(GeomContourFilledKs, StatContourFilledKs)
export(GeomPointKs, StatPointKs)
export(GeomRugKs, StatRugKs)
export(label_percent)
export(scale_asymmetric) 
export(scale_transp_neutral)
export(st_add_coordinates)
export(st_get_contour)
export(st_remove_long_segment)
export(tidy_as_kde, st_as_kde)
export(tidy_intergrid, st_intergrid)
export(tidy_kcde, st_kcde)
export(tidy_kcopula)
export(tidy_kcurv, st_kcurv)
export(tidy_kda, st_kda)
export(tidy_kdcde, st_kdcde)
export(tidy_kdde, st_kdde)    
export(tidy_kde, st_kde)
export(tidy_kde_balloon, st_kde_balloon)   
export(tidy_kde_boundary, st_kde_boundary)
export(tidy_kde_local_test, st_kde_local_test)
export(tidy_kde_sp, st_kde_sp)
export(tidy_kde_truncate, st_kde_truncate)       
export(tidy_kdr, st_kdr)            
export(tidy_kfs, st_kfs)           
export(tidy_kms, st_kms)
export(tidy_kquiver, st_kquiver)
export(tidy_kroc, st_kroc)
export(tidy_ks, st_ks)           
export(tidy_ksupp, st_ksupp)
export(unfactor)  

S3method(c, sf_ks)
S3method(c, tidy_ks)
S3method(contourLevels, tidy_ks)
S3method(ggplot, sf_ks)
S3method(ggplot, tidy_ks)
S3method(ggplot_add, scale_rem)
S3method(plot, sf_ks)
S3method(rbind, sf_ks)
S3method(rbind, tidy_ks)

## Introduction 

This package extends the functionality of the kernel smoothing functions from the `ks` package in base `R` to the tidyverse and to GIS (Geographical Information Systems) ecosystems. 

As the kernel smoothers from the `ks` package are prefixed as `k*`, their equivalents in `eks` are systematically named as follows:   

* `tidy_k*` for 1- and 2-d tidy data 
* `st_k*` for 2-d geospatial data.
    
The output data tibbles (tidy data frames provided by the `tibble` package) from `tidy_k*` can be visualised within the `ggplot2` graphical interface, using the usual layer functions and the custom ones supplied in this package. These `tidy_k*` functions are analogous to those in the `broom` and related packages, though the latter tend to focus on tidying the summary diagnostic output from model fitting (and not on tidying the underlying estimates themselves), whereas `tidy_k*` are more substantive since they do compute tidy estimates.   
  
The output simple feature geometries (provided by the `sf` package) from `st_k*` can be visualised in the (i) `ggplot2` graphical interface using primarily the `ggplot2::geom_sf` layer function, or (ii) in the base `R` graphical interface using the `plot` method supplied in this package. These simple feature geometries can also be exported as standard geospatial formats (e.g. shapefile, GEOS geometry) for use in external GIS software such as [ArcGIS](https://www.arcgis.com/index.html) and [QGIS](https://qgis.org/). 

For an illustration of kernel density estimates for tidy and geospatial data, see `vignette("eks")`.

## Installation

Install from CRAN:

```r
install.packages("eks")
```

## Further reading

Chacon, J. E. and Duong, T. (2018) [_Multivariate Kernel Smoothing and Its Applications_](https://mvstat.net/mvksa/) Chapman & Hall/CRC Press, Boca Raton.

Duong, T. (2025) [Statistical visualisation for tidy and geospatial data in R via kernel smoothing methods in the eks package](https://mvstat.net/tduong/research/publications/duong-2025-compstat.pdf) _Computational Statistics_ **40**, 2825--2847.

## ----echo=FALSE, message=FALSE------------------------------------------------
knitr::opts_chunk$set(global.par=TRUE, collapse=TRUE, comment="#>", fig.width=5, fig.height=5, fig.align="center", dpi=96)
options(tibble.print_min=3L, tibble.print_max=3L)

## ----message=FALSE------------------------------------------------------------
library(eks)
library(colorspace)
library(ggplot2)
library(dplyr)

## crabs data 
data(crabs, package="MASS")
crabs2 <- select(crabs, FL, CW)
xlab <- "Frontal lobe size (mm)"
ylab <- "Carapace width (mm)"

## -----------------------------------------------------------------------------
## KDE contour plot + scatter plot
tkde2 <- tidy_kde(crabs2)
gkde2 <- ggplot(tkde2, aes(x=FL, y=CW)) + labs(x=xlab, y=ylab)
gkde2 + geom_point_ks(colour=8) + geom_contour_ks(colour=1)

## -----------------------------------------------------------------------------
## geom_density_2d KDE contour plot + scatter plot
mkde2 <- ggplot(crabs2, aes(x=FL, y=CW))
mkde2 + geom_point(colour=8) + geom_density_2d(colour=1, bins=4)

## -----------------------------------------------------------------------------
## KDE filled contour plot
gkde2 + geom_contour_filled_ks(colour=1) 

## -----------------------------------------------------------------------------
## KDE continuous colour scale plot
gkde2 + geom_raster(aes(fill=estimate), interpolate=TRUE) + 
   scale_fill_continuous_sequential(palette="Heat")

## ----fig.width=7--------------------------------------------------------------
crabs2g <- select(crabs, FL, CW, sp)
crabs2g <- group_by(crabs2g, sp)
tkde2g <- tidy_kde(crabs2g)
gkde2g <- ggplot(tkde2g, aes(x=FL, y=CW, group=sp)) + labs(x=xlab, y=ylab, colour="Species") + 
   scale_colour_manual(values=c(4, 7)) + 
   guides(colour=guide_legend(title="Species"))

## facetted KDE contour plots + scatter plots
gkde2g + geom_point_ks(colour=8) + 
   geom_contour_ks(aes(colour=sp)) + facet_wrap(~sp) 

## ----fig.width=7--------------------------------------------------------------
## facetted KDE filled contour plots with fixed contour levels for all facets
bkde2g <- contour_breaks(tkde2g)
gkde2g + geom_contour_filled_ks(breaks=bkde2g, colour=1) + facet_wrap(~sp)

## ----message=FALSE------------------------------------------------------------
library(sf)

## Grevillea data
data(grevilleasf, package="eks")
grevilleasf <- mutate(grevilleasf, species=factor(species))
paradoxa <- filter(grevilleasf, name %in% "Grevillea paradoxa")
eryngioides <- filter(grevilleasf, name %in% "Grevillea eryngioides")
grevillea_ep <- filter(grevilleasf, name %in% c("Grevillea eryngioides", 
   "Grevillea paradoxa"))
grevillea_ep <- group_by(grevillea_ep, name)
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)

## WA polygon
data(wa, package="eks")
gwa <- geom_sf(data=wa, fill=NA, colour=1)

## -----------------------------------------------------------------------------
## base R scatter plot
plot(st_geometry(wa), xlim=xlim, ylim=ylim)
plot(st_geometry(eryngioides), add=TRUE, col=3, pch=16, cex=0.5)
plot(st_geometry(paradoxa), add=TRUE, col=6, pch=17, cex=0.5)
mapsf::mf_legend(type="symb", val=c("Grevillea eryngioides", "Grevillea paradoxa"), 
   pal=c(3,6), pch=16:17, cex=c(1,1), title="Species", pos="bottomleft")
## geom_sf scatter plot 
theme_set(ggthemes::theme_map())
ggplot() + gwa + 
   geom_sf(data=grevillea_ep, aes(colour=name, shape=name)) + 
   coord_sf(xlim=xlim, ylim=ylim) + scale_colour_manual(values=c(3, 6)) + 
   guides(colour=guide_legend(title="Species"), shape=guide_legend(title="Species"))

## -----------------------------------------------------------------------------
skde1 <- st_kde(paradoxa)

## base R contour plot
plot(st_geometry(wa), xlim=xlim, ylim=ylim)
plot(st_geometry(paradoxa), add=TRUE, pch=16, col=8, cex=0.5)
plot(skde1, add=TRUE, col=NA, border=1, legend=FALSE)

## -----------------------------------------------------------------------------
## geom_sf contour plot
gs <- ggplot(skde1) + gwa + ggthemes::theme_map()
gs + geom_sf(data=paradoxa, col=8, size=0.5) + 
   geom_sf(data=st_get_contour(skde1), colour=1, fill=NA) + 
   coord_sf(xlim=xlim, ylim=ylim)

## -----------------------------------------------------------------------------
## R base filled contour plot
plot(st_geometry(wa), xlim=xlim, ylim=ylim)
plot(skde1, add=TRUE)
## geom_sf filled contour
gs + geom_sf(data=st_get_contour(skde1), aes(fill=contperc)) + 
  coord_sf(xlim=xlim, ylim=ylim) 

## ----fig.width=7--------------------------------------------------------------
skde1g <- st_kde(grevillea_ep)

## facetted geom_sf filled contour
gs + geom_sf(data=st_get_contour(skde1g), aes(fill=contperc)) + 
   facet_wrap(~name) + coord_sf(xlim=xlim, ylim=ylim) 
## facetted geom_sf filled contour with fixed contour levels for all facets
bkde1g <- contour_breaks(skde1g, group=TRUE)
gs + geom_sf(data=st_get_contour(skde1g, breaks=bkde1g), aes(fill=estimate)) +
   facet_wrap(~name) + coord_sf(xlim=xlim, ylim=ylim) 

#####################################################################
## Auxiliary functions for tidy_* functions
#####################################################################

ggplot.tidy_ks <- function(data=NULL, mapping=aes(), ...)
{
    data1 <- data
    if (!is.null(data)) 
    {
        class(data1) <- class(data1)[class(data1)!="tidy_ks"]
        transp_neutral <- any(object_class(data, type="ks") %in% "kdde")
    }
    else transp_neutral <- FALSE
    p <- ggplot2::ggplot(data=data1, mapping=mapping, ...)
    p <- p + aes_ks(data) + labs_ks(data) + guides_ks(data)
    p <- p + scale_colour_ks(data) + scale_fill_ks(data, transp_neutral=transp_neutral)
    
    return(p)
}

## extract ks object from tidy_ks object
## object = tidy_ks object
untidy_ks <- function(object, name, collapse=FALSE, var_ks="ks")
{
    ge <- dplyr::pull(dplyr::slice_head(object), var_ks)
    n <- length(ge)

    if (n==1)
    {
        ge <- ge[[1]]
        if (!missing(name)) ge <- getElement(ge, name)
    }
    else
    {
        if (!missing(name)) ge <- lapply(ge, getElement, name)
    }

    return(ge)
}

## extract data matrix from ks object from tidy_ks object
## object = tidy_ks object
get_data_ks <- function(object, var_ks="ks")
{
    ge <- untidy_ks(object, "x", var_ks=var_ks)
    if (is.null(ge)) ge <- rbind(untidy_ks(object, "fhat1", var_ks=var_ks)$x, untidy_ks(object, "fhat2", var_ks=var_ks)$x)
    oc <- class(getElement(object, var_ks)[[1]])
    
    if (oc %in% "kda")
    {
        gr <- unique(object$group)
        ge <- data.frame(ge[[gr]])
        colnames(ge) <- c("x","y","z")[1:ncol(ge)]
        ge <- data.frame(ge, group=gr)
    }
    else
    {
        if (!is.list(ge))
        {
            ge <- data.frame(ge)
            colnames(ge) <- c("x","y","z")[1:ncol(ge)]
        }
        else
        {
            gr <- rep(unique(object$group), lengths(ge))
            d <- unlist(tail(getElement(dplyr::ungroup(object), "ks"),n=1))
            if (d==1) ge <- data.frame(do.call("c",ge))
            else ge <- do.call("rbind",ge)
            colnames(ge) <- c("x","y","z")[1:ncol(ge)]
            rownames(ge) <- NULL
            ge <- data.frame(ge, group=gr)
        }
    }

    return(ge)
}

## dimension of tidy_ks/sf_ks objects
dim_ks <- function(object)
{
    if (inherits(object, "sf_ks"))
    {
        object <- object$tidy_ks[1,]
        if (object$tks %in% "kda") d <- ncol(untidy_ks(object)$H[[1]])
        else if (object$tks %in% "ksupp") d <- ncol(untidy_ks(object))
        else d <- ncol(get_data_ks(object))
    }
    else
        d <- unlist(tail(getElement(dplyr::ungroup(object), "ks"),n=1))

    return(as.integer(d))
}

## renaming function
rename_ks <- function(data, gg, d)
{
    gv <- dplyr::group_vars(data)
    if (length(gv)>0) nd <- names(dplyr::select(dplyr::ungroup(data), -dplyr::all_of(gv)))
    else nd <- names(data)

    if (d>=1) names(gg)[names(gg)=="x"] <- nd[1]
    if (d>=2) names(gg)[names(gg)=="y"] <- nd[2]

    return(gg)
}

## cast data as "tidy_ks" object
as_tidy_ks <- function(data)
{
    class(data) <- unique(c("tidy_ks", class(data)))

    return(data)
}

## concatenate tidy_ks objects
c.tidy_ks <- function (..., add_group=TRUE, labels) 
{
    args <- list(...)
    nargs <- length(args)
    
    if (nargs>=1) 
    {
        if (add_group)
        {
        
            tt <- dplyr::mutate(args[[1]], group=1L)
            if (nargs>=2)
            {
                for (i in 2:nargs) 
                    tt <- dplyr::bind_rows(tt, dplyr::mutate(args[[i]], group=i))
            }
            if (any(names(tt) %in% "group"))
            {
                if (missing(labels)) tt$group <- factor(tt$group)
                else tt$group <- factor(tt$group, labels=labels)
                tt <- dplyr::group_by(tt, .data$group)
            }
            tt <- as_tidy_ks(tt)
        }
        else 
        {   
            tt <- dplyr::bind_rows(...)
            tt <- as_tidy_ks(tt)
        }
    }
    else tt <- NULL

    return(tt)
}

rbind.tidy_ks <- c.tidy_ks

#####################################################################
## Probability contour regions
#####################################################################
## compute contour breaks for ks objects
## data = tidy ks/sf_ks object
contour_breaks <- function(data, cont=c(25,50,75), n=3, group=FALSE, type="density") 
{ 
    cont.aug <- cont[cont>=1 & cont<=99]
    oct <- object_class(data, type="tks")
    type <- match.arg(type, c("density", "length", "quantile", "natural"))

    ## sf_ks object
    if (inherits(data, "sf_ks")) 
    {
        data$tidy_ks <- as_tidy_ks(data$tidy_ks)
        is.kdde <- oct %in% "kdde"
       
        if (type=="density") 
        {
            ## contour levels only w/o min, w/o max 
            cb <- contourLevels(x=data, cont=cont.aug)
            if (is.kdde) cb <- dplyr::group_modify(cb, .f=~head(tail(.x, n=-1), n=-1))
            else cb <- dplyr::group_modify(cb, .f=~head(.x, n=-1))
        }
        else 
        {
            ## contour levels, kde: with min, w/o max, kdde: w/o min, w/o max
            cb <- dplyr::group_modify(data$grid, .f=~data.frame(breaks=ks::.contourBreaks(.x, cont=cont.aug, n=n, type=type, is.kdde=is.kdde)))
            if (is.kdde) cb <- dplyr::group_modify(cb, .f=~head(tail(.x, n=-1), n=-1))
        }
    }
    ## tidy_ks object
    else
    {
        data <- as_tidy_ks(data) 
        is.kdde <- oct %in% "kdde"
       
        if (type=="density") 
        {
            cb <- ks::contourLevels(x=data, cont=cont.aug)
            cb <- dplyr::group_modify(cb, .f=~head(.x, n=-1))
        }
        else 
        {
            cb <- dplyr::group_modify(data, .f=~data.frame(breaks=ks::.contourBreaks(.x, cont=cont.aug, n=n, type=type, is.kdde=is.kdde)))
            if (is.kdde) cb <- dplyr::group_modify(cb, .f=~head(.x, n=-1))
        }    
    }
    
    if (group)
    {
        cb.summary <- dplyr::mutate(cb, contgr=dplyr::row_number(), .before=1)
        cb.summary <- dplyr::select(cb.summary, dplyr::all_of(c("contgr", "breaks", dplyr::group_vars(cb.summary))))
        if (inherits(data, "sf_ks")) data2 <- sf::st_drop_geometry(data$sf) else data2 <- data
        if (is.kdde & any(names(data2) %in% "deriv_ind")) 
        { 
            cb.summary <- dplyr::left_join(cb.summary, dplyr::distinct(dplyr::ungroup(data2)[,c("deriv_ind", "deriv_group")]), by="deriv_group")
            cb.summary <- dplyr::relocate(cb.summary, "contgr", "breaks", .before=1)
        }
        cb.summary <- dplyr::arrange(cb.summary, .data$breaks, .by_group=TRUE)  
    }
    else 
    {    
        cb <- dplyr::arrange(cb, .data$breaks)
        cb <- dplyr::ungroup(cb) 
        cb <- dplyr::distinct(cb, .data$breaks)

        cont.prob.diff <- function(cb)
        {
            if (inherits(data, "sf_ks")) data2 <- data$tidy_ks
            else data2 <- dplyr::slice_head(data, n=1) 
            cb.mat <- matrix(0, ncol=nrow(data2), nrow=nrow(cb))
            for (i in 1:ncol(cb.mat)) 
            {
                if (inherits(untidy_ks(data2[i,]), "kda"))
                {
                    kdei <- untidy_ks(data2[i,])
                    kdei$x <- kdei$x[[i]]
                    kdei$estimate <- kdei$estimate[[i]]
                    kdei$H <- kdei$H[[i]]
                    kdei$cont <- kdei$cont[[i]]   
                    kdei$w <- kdei$w[[i]]
                    class(kdei) <- "kde"
                    data2[i,]$ks <- list(kdei)
                }
                if (inherits(untidy_ks(data2[i,]), "kdde")) cb.mat[,i] <- cb$breaks
                else cb.mat[,i] <- ks::contourProbs(untidy_ks(data2[i,]), abs.cont=as.vector(unlist(cb)))
            }
            return(as.vector(cb.mat))
        }

        if (nrow(cb)>2) 
        {
            ## always keep min value
            if (inherits(data, "tidy_ks")) 
            {
                cb1 <- head(cb, n=1); cb1$contgr <- 0;
                cb <- tail(cb, n=-1)
            }
            cb2 <- cont.prob.diff(cb)
            cont.cutoff <- abs(head(diff(cont.aug),n=1))
            if (inherits(data, "sf_ks")) data2 <- data$tidy_ks else data2 <- data
            if (any(data2$tks %in% "kdde")) {cb2 <- (cb2-min(cb2))/(max(cb2)-min(cb2))} 
            clust <- cutree(hclust(dist(cb2), method="complete"), h=cont.cutoff/100)
            if (length(clust)==0) clust <- 1:length(cb2)
            cb <- data.frame(breaks=rep(cb$breaks, times=length(cb2)/nrow(cb)))
            cb$contgr <- clust
            cb <- dplyr::distinct(cb)
            if (inherits(data, "tidy_ks")) { cb <- rbind(cb1, cb); cb$contgr <- cb$contgr+1 }
        }
        else 
            cb$contgr <- 1:nrow(cb)

        cb <- dplyr::group_by(cb, .data$contgr)
        cb.summary <- dplyr::summarise(cb, min.breaks=min(.data$breaks), mean.breaks=mean(.data$breaks), max.breaks=max(.data$breaks))
        cb.summary <- dplyr::mutate(cb.summary, breaks=ifelse(dplyr::row_number()==1, .data$min.breaks, ifelse(dplyr::row_number()==dplyr::n(), .data$max.breaks, .data$mean.breaks)))
        cb.summary <- dplyr::select(cb.summary, dplyr::all_of(c("contgr", "breaks"))) 
        cb.summary <- dplyr::arrange(cb.summary, .data$breaks) 
        cb.summary <- dplyr::distinct(cb.summary, .data$breaks, .keep_all=TRUE)
        cb.summary$contgr <- 1:nrow(cb.summary)  
    }
    
    return(cb.summary)    
}

## contour levels method for tidy_ks opjects
contourLevels.tidy_ks <- function(x, ..., cont=c(25,50,75))
{
    oct <- object_class(x, type="tks")
    if (oct %in% "kcurv") 
    { 
        ftemp <- x$ks[lengths(x$ks)>1]
        ftemp <- lapply(ftemp, function(.) { class(.) <- "kdecurv"; . })
        x$ks[lengths(x$ks)>1] <- ftemp
    }
    data <- x
    data$.group <- dplyr::group_indices(data)
    gg <- dplyr::slice_head(data)
    add_minmax <- TRUE 

    if (oct %in% "kdde")
        gg <- dplyr::group_modify(.data=gg, .f=~{ breaks <- .contourLevels(untidy_ks(.x), cont=cont, which.deriv.ind=unique(.x$deriv.ind), add_minmax=add_minmax); data.frame(breaks, cont=as.numeric(names(breaks))) })
    else if (oct %in% "kda")
        gg <- dplyr::group_modify(.data=gg, .f=~{ breaks <- .contourLevels(untidy_ks(.x), cont=cont, add_minmax=add_minmax)[[.x$.group]]; data.frame(breaks, cont=as.numeric(names(breaks))) })
    else
        gg <- dplyr::group_modify(.data=gg, .f=~{ breaks <- .contourLevels(untidy_ks(.x), cont=cont, add_minmax=add_minmax); data.frame(breaks, cont=as.numeric(names(breaks))) })
    
    return(gg)
}

## fhat = untidy ks object
## add_minmax = add min and/or max values of fhat$estimate to output 
.contourLevels <- function(fhat, cont, which.deriv.ind=1, approx.cont=TRUE, add_minmax=TRUE)
{
    cont <- sort(cont)
    oc <- head(class(fhat),n=1) ## equivalent to object_class(type="ks")
    
    if (oc %in% c("kdde","kqdde"))
    {
        hts <- ks::contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont, which.deriv.ind=fhat$deriv.ind.scalar)
    }
    else if (oc %in% "kdecurv")
    {
        class(fhat) <- "kde"
        fhat.temp <- fhat
        fhat.temp$x <- fhat$x[predict(fhat, x=fhat$x)>0, ]
        hts <- ks::contourLevels(fhat.temp, prob=(100-cont)/100, approx=approx.cont)
    }
    else if (oc %in% c("kcde", "kcopula"))
    {
        hts <- sort(cont/100)
        names(hts) <- paste0(cont,"%")
    }
    else if (oc %in% "kde.loctest")
        hts <- c(-10,-0.5, 0.5, 10)
    else if (oc %in% "kfs")
        hts <- c(0.5, 10)
    else
        hts <- ks::contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)

    if (oc %in% "kda")
    {
        hts <- lapply(hts, sort)
        pp <- fhat$prior.prob
        if (add_minmax) 
        {
            for (i in 1:length(hts)) 
            { 
                hts[[i]] <- c(pp[i]*hts[[i]], max(c(fhat$estimate[[i]], pp[i]*hts[[i]])) + 0.01 * max(abs(fhat$estimate[[i]])))
                names(hts[[i]])[length(hts[[i]])] <- "101%"
            }
        }
    }
    else if (oc %in% c("kdde","kqdde"))
    {
        hts1 <- hts[1,]; hts2 <- hts[2,]
        names(hts1) <- paste0("-", names(hts1))
        hts <- c(sort(hts1), sort(hts2))
        if (add_minmax) 
        {
            hts <- c(min(c(sapply(fhat$estimate, min), hts)) - 0.01 * max(abs(sapply(fhat$estimate, max))), hts, max(c(sapply(fhat$estimate, max), hts)) + 0.01 * max(abs(sapply(fhat$estimate, max))))
            names(hts)[1] <- "-101%"
            names(hts)[length(hts)] <- "101%"
        }
    }
    else 
    {
        hts <- sort(hts)
        if (add_minmax) 
        {
            hts <- c(hts, max(c(fhat$estimate, hts)) + 0.01 * max(abs(fhat$estimate)))
            names(hts)[length(hts)] <- "101%"
        }
    }
    names(hts) <- gsub("%", "", names(hts))

    return (hts)
}

#####################################################################
## stat_* and geom_* auxiliary functions for ggplot2 layers
#####################################################################

## geom and stat function for extracting data points from tidy_ks object
## required for adding scatter plots of original data in geom_point and geom_rug
geom_point_ks <- function(mapping=NULL, data=NULL, stat="point_ks", position="identity", ..., na.rm=FALSE, jitter=FALSE, show.legend=NA, inherit.aes=TRUE)
{
    ggplot2::layer(data=data, mapping=mapping, stat=stat, geom=GeomPointKs, position=position, show.legend=show.legend, inherit.aes=inherit.aes, params=list(na.rm=na.rm, jitter=jitter, ...))
}

GeomPointKs <- ggplot2::ggproto("GeomPointKs", GeomPoint)

stat_point_ks <- function(mapping=NULL, data=NULL, geom="point_ks", position="identity", ..., na.rm=FALSE, show.legend=NA, inherit.aes=TRUE)
{
    ggplot2::layer(data=data, mapping=mapping, stat=StatPointKs, geom=geom, position=position, show.legend=show.legend, inherit.aes=inherit.aes, params=list(na.rm=na.rm, jitter=jitter, ...))
}

StatPointKs <- ggplot2::ggproto("StatPointKs", ggplot2::Stat,
compute_group=function(self, data, scales, jitter=FALSE)
{
    x <- get_data_ks(data, var_ks="weight")
    if (jitter) x <- data.frame(do.call(cbind,lapply(x, jitter, fac=1)))
    return(x)
},
required_aes=c("x","y"),
dropped_aes=c("z", "weight")
)

geom_rug_ks <- function(mapping=NULL, data=NULL, stat="rug_ks", position="identity", ..., outside=FALSE, sides="bl", length=unit(0.03, "npc"), na.rm=FALSE, jitter=FALSE, show.legend=NA, inherit.aes=TRUE)
{
    ggplot2::layer(data=data, mapping=mapping, stat=stat, geom=GeomRugKs, position=position, show.legend=show.legend, inherit.aes=inherit.aes, params=list(outside=outside, sides=sides, length=length, na.rm=na.rm, jitter=jitter, ...))
}

GeomRugKs <- ggplot2::ggproto("GeomRug", GeomRug)

stat_rug_ks <- function(mapping=NULL, data=NULL, geom="rug_ks", position="identity", ..., na.rm=FALSE, show.legend=NA, inherit.aes=TRUE)
{
    ggplot2::layer(data=data, mapping=mapping, stat=StatRugKs, geom=geom, position=position, show.legend=show.legend, inherit.aes=inherit.aes, params=list(na.rm=na.rm, ...))
}

StatRugKs <- ggplot2::ggproto("StatPointKs", ggplot2::Stat,
compute_group=function(self, data, scales, jitter=FALSE)
{
    x <- get_data_ks(data, var_ks="weight")
    if (jitter) x <- data.frame(do.call(cbind,lapply(x, jitter, fac=1)))
     return(x)
},
required_aes=c("x"),
dropped_aes=c("y", "z", "weight") 
)

## stat and geom functions for computing probability contours
## replace ggplot2::geom_contour and ggplot2::stat_contour
geom_contour_ks <- function(mapping=NULL, data=NULL, stat="contour_ks", position="identity", ..., cont=c(25,50,75), contperc=TRUE, breaks=NULL, digits=NULL, show.legend=NA, inherit.aes=TRUE)
{
    aest <- dplyr::as_label(mapping$colour)
    if (aest!="NULL") aest <- gsub("\\)", "", gsub("after_stat\\(", "", aest))
    
    ## breaks_panel = set of common breaks for all panels
    breaks_panel <- NULL
    glayer <- ggplot2::layer(data=data, mapping=mapping, stat=stat, geom=GeomContourKs, position=position, show.legend=show.legend, inherit.aes=inherit.aes, params=list(cont=cont, contperc=contperc, breaks=breaks, digits=digits, breaks_panel=breaks_panel, aest=aest, ...))
    
    ## clear previous fill scale + create new fill scale with breaks
    if (!is.null(data) & !is.null(breaks))
    {
        if (object_class(data, type="tks") %in% "kdde") 
        {
            if (any(names(breaks) %in% "breaks")) breaks1 <- breaks$breaks else breaks1 <- breaks
            if (inherits(breaks, c("matrix","data.frame","tbl_df"))) 
                breaks <- breaks[breaks1 > min(data$estimate) & breaks1 < max(data$estimate),]
            else
                breaks <- breaks[breaks1 > min(data$estimate) & breaks1 < max(data$estimate)] 
            transp_neutral <- object_class(data, type="ks") %in% "kdde" 
            glayer <- c(scale_asymmetric(data, breaks=breaks, transp_neutral=transp_neutral), glayer)
        }
    }
    
    return(glayer)
}

GeomContourKs <- ggplot2::ggproto("GeomContourFilled", GeomContour)

stat_contour_ks <- function(mapping=NULL, data=NULL, geom="contour_ks", position="identity", ..., show.legend=NA, inherit.aes=TRUE)
{
    ggplot2::layer(data=data, mapping=mapping, stat=StatContourKs, geom=geom, position=position, show.legend=show.legend, inherit.aes=inherit.aes, params=list(...))
}

## replaces ggplot2::StatContour
## data = tidy_ks object
## default values needed here since this StatContourKs is invoked directly from geom_contour(stat="contour_ks") rather than reading in the parameters from geom_contour_ks()
StatContourKs <- ggplot2::ggproto("StatContourKs", ggplot2::StatContour,
dropped_aes="weight",
compute_group=function(self, data, scales, cont=c(25,50,75), contperc=FALSE, breaks=NULL, breaks_panel=NULL, digits=NULL, aest=NULL, bins=NULL, binwidth=NULL)
{
    fhat <- untidy_ks(data, var_ks="weight")
    oc <- head(class(fhat),n=1) ## equivalent to object_class(fhat, type="ks")
    null_breaks <- is.null(breaks)
    if (is.null(digits)) digits <- ifelse(oc %in% c("kcde", "kcopula", "kde.loctest", "kfs", "kdr"), 2, 4)
   
    ## compute ks probability contour levels
    if (is.null(breaks))
    {
        if (oc %in% "kde.loctest") 
        {
            breaks <- c(-10,-0.5,0.5,10)
            breaks.label <- fhat$label
        }
        else if (oc %in% "kfs")
        {
            breaks <- c(-10, 0.5, 10)
            breaks.label <- fhat$label
        }
        else 
        {
            if (isTRUE(fhat$type=="kcurv")) class(fhat) <- "kdecurv"
            breaks <- .contourLevels(fhat=fhat, cont=cont)
            if (is.list(breaks)) breaks <- breaks[[unique(data$group)]]
        }

        if (oc %in% "kcopula") names(breaks)[length(breaks)] <- "100%"
        else if (oc %in% "kde.loctest")
        {
            fhat.diff.lab <- fhat$fhat1
            fhat.diff.lab$estimate <- fhat$fhat.diff.pos$estimate - fhat$fhat.diff.neg$estimate
            d <- ncol(fhat$fhat1$H)
            diff.lab <- compute.tidy(fhat.diff.lab, d=d, tidy=FALSE)$estimate
            data$z <- diff.lab     
        }
        else if (oc %in% "kfs")
        {
            d <- ncol(fhat$H)
            data$z <- compute.tidy(fhat, d=d, tidy=FALSE)$estimate
        }
           
        isolines <- .ggplot2_xyz_to_isolines(data, breaks)
        path_df <- .ggplot2_iso_to_path(isolines, data$group[1])
        path_df$estimate <- path_df$level <- as.numeric(path_df$level)
        
        ## if #distinct values in path_df$level != #names.isolines    
        if (oc %in% c("kdde", "kqdde"))
        {
            nc <- length(cont)
            names.isolines <- c(-100 - as.numeric(gsub("%", "",tail(head(names(breaks), n=nc+1),n=-1))), 100 - as.numeric(gsub("%", "",head(tail(names(breaks), n=nc+1),n=-1))))
            if (length(unique(path_df$estimate)) != length(names.isolines)) 
            {
                breaks2 <- signif(breaks[names(breaks) %in% names.isolines],8)
                names.isolines <- names.isolines[!duplicated(breaks2)]
                breaks2 <- breaks2[!duplicated(breaks2)]
                levels2 <- as.character(signif(unique(path_df$level),8))
                path_df$level <- ordered(path_df$level, labels=names.isolines[breaks2 %in% levels2])
            }
            else  
                path_df$level <- ordered(path_df$level, labels=names.isolines)
        }
        else if (oc %in% "kde.loctest")
        {
            names.isolines <- c("-0.5", "0.5")
            path_df$level <- ordered(path_df$level, labels=names.isolines)
        }
        else if (oc %in% "kfs")
        {
            names.isolines <- c("0.5")
            path_df$level <- ordered(path_df$level, labels=names.isolines)
        }
        else
        {
            path_df$level <- findInterval(path_df$level, as.numeric(names(table(path_df$level))))
            names.isolines <- 100-as.numeric(gsub("%", "", head(names(breaks), n=-1)))
            path_df$level <- ordered(path_df$level, labels=names.isolines)
        }
    }
    else
    {
        if (any(names(breaks) %in% "breaks")) breaks <- breaks$breaks
        else if (is.list(breaks)) breaks <- breaks[[unique(data$group)]]
        breaks <- sort(breaks)

        isolines <- .ggplot2_xyz_to_isolines(data, breaks)
        path_df <- .ggplot2_iso_to_path(isolines, data$group[1])
        path_df$estimate <- path_df$level <- as.numeric(path_df$level)
        path_df$level <- round_signif(path_df$level, digits=digits) 
        path_df$level <- ordered(path_df$level)
    }
    
    ## ordered factor used in original ggplot2::StatContourFilled
    ## can cause issues when combining with unordered factors in ggplots 
    class(path_df$level) <- "factor"
    path_df$contlabel <- path_df$level
     
    ## create more informative labels
    path_df <- create_label(path_df, digits=digits, add_contperc=null_breaks, is_kdde=oc %in% c("kdde", "kqdde", "kde.loctest"), is_filled_contour=FALSE)   
    path_df$estimate <- round_signif(path_df$estimate, digits=digits)

    ## breaks_panel = common breaks for all panels   
    if (is.null(breaks_panel))
        path_df$estimate <- factor(path_df$estimate, levels=unique(path_df$estimate)) 
    else
    {    
        breaks_panel_br <- sort(unique(breaks_panel$breaks))
        breaks_panel_df <- create_label(data.frame(estimate=breaks_panel_br, contlabel=breaks_panel_br), digits=digits, add_contperc=null_breaks, is_kdde=oc %in% c("kdde", "kde.loctest"), is_filled_contour=FALSE) 
        breaks_panel_df$estimate <- round_signif(breaks_panel_df$estimate, digits=digits)
        path_df$estimate <- factor(path_df$estimate, levels=breaks_panel_df$estimate)
        path_df$contregion <- factor(path_df$contregion, levels=breaks_panel_df$contregion)
    }

    if (oc %in% "kde.loctest") 
    {
        levels(path_df$contlabel) <- breaks.label
        levels(path_df$contperc) <- c("-50%", "50%")
        path_df$contregion <- path_df$contlabel
    }
    else if (oc %in% "kfs")
    {
        levels(path_df$contperc) <- "50%"
        levels(path_df$contlabel) <- breaks.label
        path_df$contregion <- path_df$contlabel
    }
    if (aest=="NULL") path_df$level <- path_df$contperc
    else if (aest %in% c("contregion", "contlabel", "contline", "estimate"))path_df$level <- path_df[[aest]]
    
    return(path_df)
},
required_aes=c("x","y","z"),
dropped_aes=c("z","weight") 
)

## stat and geom functions for computing filled probability contours
## replace ggplot2::geom_contour_filled & ggplot2::stat_contour_filled
geom_contour_filled_ks <- function(mapping=NULL, data=NULL, stat="contour_filled_ks", position="identity", ..., cont=c(25,50,75), contperc=TRUE, breaks=NULL, transp_neutral=NULL, digits=NULL, show.legend=NA, inherit.aes=TRUE)
{
    aest <- dplyr::as_label(mapping$fill)
    if (aest!="NULL") aest <- gsub("\\)", "", gsub("after_stat\\(", "", aest))
    
    breaks_panel <- NULL
    glayer <- ggplot2::layer(data=data, mapping=mapping, stat=stat, geom=GeomContourFilled, position=position, show.legend=show.legend, inherit.aes=inherit.aes, params=list(cont=cont, contperc=contperc, digits=digits, breaks=breaks, breaks_panel=breaks_panel, aest=aest, ...))
 
    ## clear previous fill scale + create new fill scale with breaks
    if (!is.null(data) & !is.null(breaks))
    {
        if (object_class(data, type="tks") %in% "kdde") 
        {
            if (any(names(breaks) %in% "breaks")) breaks1 <- breaks$breaks else breaks1 <- breaks
            if (inherits(breaks, c("matrix","data.frame","tbl_df"))) 
                breaks <- breaks[breaks1 > min(data$estimate) & breaks1 < max(data$estimate),]
            else
                breaks <- breaks[breaks1 > min(data$estimate) & breaks1 < max(data$estimate)] 
            transp_neutral <- object_class(data, type="ks") %in% "kdde" 
            glayer <- c(scale_asymmetric(data, breaks=breaks, transp_neutral=transp_neutral), glayer)
        }
    }
    
    ## if contline is aesthetic then don't display its legend
    if (any(names(mapping) %in% "linetype")) 
    {
        if (length(grep("contline", dplyr::as_label(mapping$linetype)))>0)
            glayer <- c(glayer, ggplot2::guides(linetype="none"))
    }

    return(glayer)
}

GeomContourFilledKs <- ggplot2::ggproto("GeomContourFilled", GeomContourFilled)

## replace ggplot2::stat_contour_filled & ggplot2::StatContourFilled
stat_contour_filled_ks <- function(mapping=NULL, data=NULL, geom="contour_filled_ks", position="identity", ..., show.legend=NA, inherit.aes=TRUE)
{
    ggplot2::layer(data=data, mapping=mapping, stat=StatContourFilledKs, geom=geom, position=position, show.legend=show.legend, inherit.aes=inherit.aes, params=list(na.rm=FALSE, ...))
}

## default values needed here since this StatContourFilledKs is invoked directly from geom_contour_filled(stat="contour_filled_ks") rather than reading in the parameters from geom_contour_filled_ks()
StatContourFilledKs <- ggplot2::ggproto("StatContourFilledKs", StatContourFilled,
compute_group=function(data, scales, cont=c(25,50,75), contperc=TRUE, breaks=NULL, breaks_panel=NULL, digits=NULL, aest=NULL, bins=NULL, binwidth=NULL)
{
    fhat <- untidy_ks(data, var_ks="weight")
    oc <- head(class(fhat),n=1) ## equivalent to object_class(type="ks")
    null_breaks <- is.null(breaks)
    if (is.null(digits)) digits <- ifelse(oc %in% c("kcde", "kcopula", "kde.loctest", "kfs", "kdr"), 2, 4)
  
    ## compute ks probability contour levels
    if (is.null(breaks))
    {
        if (oc %in% "kde.loctest")
        {
            d <- ncol(fhat$fhat1$H)
            fhat.diff.lab <- fhat$fhat1
            fhat.diff.lab$estimate <- fhat$fhat.diff.pos$estimate - fhat$fhat.diff.neg$estimate
            diff.lab <- compute.tidy(fhat.diff.lab, d=d, tidy=FALSE)$estimate
            data$z <- diff.lab
            breaks <- c(-1, 0, 1, 10)
        }
        else if (oc %in% "kfs")
        {
            d <- ncol(fhat$H)
            data$z <- compute.tidy(fhat, d=d, tidy=FALSE)$estimate
            breaks <- c(-1.0, 10, 10.1)
        }

        if (!(oc %in% "kde.loctest")) 
        {
            if (isTRUE(fhat$type=="kcurv")) class(fhat) <- "kdecurv"
            breaks <- .contourLevels(fhat=fhat, cont=cont)
        }
        if (oc %in% "kda") breaks <- breaks[[unique(data$PANEL)]]
       
        isobands <- .ggplot2_xyz_to_isobands(data, breaks)
        
        if (oc %in% c("kdde", "kqdde"))
        {
            nc <- length(cont)
            names(isobands) <- c(-100 - as.numeric(gsub("%", "",tail(head(names(breaks), n=nc+1),n=-1))), 100, 100 - as.numeric(gsub("%", "",head(tail(names(breaks), n=nc+1),n=-1))))
        }
        else if (oc %in% "kde.loctest")
        {
            contperc <- FALSE
            labels <- na.omit(data$weight[[1]]$label)
            names(isobands) <- c(labels[1], "", labels[2]) 
        }
        else if (oc %in% "kfs")
        {
            contperc <- FALSE
            labels <- na.omit(data$weight[[1]]$label)
            names(isobands) <- head(breaks, n=-1)
        }
        else
        {
            names(isobands) <- 100-as.numeric(gsub("%", "", head(names(breaks), n=-1)), "%")
        }
    }
    else
    {
        if (any(names(breaks) %in% "breaks")) breaks <- breaks$breaks
        else if (is.list(breaks)) breaks <- breaks[[unique(data$group)]]
        breaks <- sort(breaks)
       
        if ((oc %in% c("kdde","kqdde","kda")))
            breaks <- c(sort(breaks), max(c(sapply(fhat$estimate,max), breaks)) + 0.01 * max(abs(sapply(fhat$estimate,max))))
        else
           breaks <- c(sort(breaks), max(c(fhat$estimate, breaks)) + 0.01 * max(abs(fhat$estimate)))
        isobands <- .ggplot2_xyz_to_isobands(data, breaks)
        names(isobands) <- round_signif(as.numeric(head(breaks, n=-1)), digits=digits)
    }

    ## local copy of unexported function from scales
    .scales_rescale_max <- function (x, to=c(0, 1), from=range(x, na.rm=TRUE)) { x/from[2] * to[2] }
    
    ## copied from original StatContourFilled
    path_df <- .ggplot2_iso_to_polygon(isobands, data$group[1])
    path_df$level <- ordered(path_df$level, levels=names(isobands))    
    path_df$level_low <- breaks[as.numeric(path_df$level)]
    path_df$level_high <- breaks[as.numeric(path_df$level) + 1]
    path_df$level_mid <- 0.5 * (path_df$level_low + path_df$level_high)
    path_df$nlevel <- .scales_rescale_max(path_df$level_high)
    
    ## create more informative labels
    ## contlabel = contour percentage (without %)
    ## contperc = contour percentage (with %)
    ## estimate/level = contour level
    ## contregion = contour level (with >= or <=)
    path_df$estimate <- as.numeric(path_df$level_low)
    path_df$estimate_high <- as.numeric(path_df$level_high)
    path_df$contlabel <- path_df$level
    path_df.temp <- create_label(path_df, digits=digits, add_contperc=null_breaks, is_kdde=oc %in% c("kdde", "kqdde", "kde.loctest"))
    
    if (oc %in% "kde.loctest")
    {   
        levels(path_df$contlabel)[levels(path_df$contlabel)==""] <- NA
        path_df$contperc <- factor(path_df$estimate)
        levels(path_df$contperc) <- c("-50%", NA, "50%")
        path_df$contregion <- path_df$contlabel
        path_df$estimate <- factor(path_df$estimate, labels=c(-0.5, 0, 0.5))
        levels(path_df$estimate)[2] <- NA
    }
    else if (oc %in% "kfs")
    {
        path_df$contperc <- path_df$contlabel
        levels(path_df$contperc) <- c("50%",NA)
        path_df$contlabel <- labels
        path_df$contlabel <- factor(path_df$contlabel)
        path_df$contregion <- path_df$contlabel
        path_df$estimate <- factor(path_df.temp$estimate)
    }
    else 
    {
        path_df$contperc <- path_df.temp$contperc 
        path_df$contline <- path_df.temp$contline
        path_df$contregion <- path_df.temp$contregion
        path_df$estimate <- round_signif(path_df$estimate, digits=digits)
     
        ## breaks_panel = common breaks for all panels   
        if (is.null(breaks_panel))
            path_df$estimate <- factor(path_df$estimate, levels=unique(path_df$estimate)) 
        else
        {    
            breaks_panel_br <- sort(unique(breaks_panel$breaks))
            breaks_panel_df <- create_label(data.frame(estimate=breaks_panel_br, contlabel=breaks_panel_br), digits=digits, add_contperc=null_breaks, is_kdde=oc %in% c("kdde", "kde.loctest"), is_filled_contour=FALSE) 
            breaks_panel_df$estimate <- round_signif(breaks_panel_df$estimate, digits=digits)
            path_df$estimate <- factor(path_df$estimate, levels=breaks_panel_df$estimate)
            path_df$contregion <- factor(path_df$contregion, levels=breaks_panel_df$contregion)
        }
    }

    if (aest=="NULL") 
    {
        if (null_breaks) path_df$level <- path_df$contperc
        else path_df$level <- path_df$estimate
    }
    else if (aest %in% c("contregion", "contlabel", "contline", "estimate")) path_df$level <- path_df[[aest]]

    return(path_df)
},
required_aes=c("x","y","z"),
dropped_aes=c("z","weight") 
)

## local copy of ggtheme::theme_map
theme_sf <- function() 
{
    theme(axis.line = element_blank(), axis.text = element_blank(), axis.ticks = element_blank(), axis.title = element_blank(), panel.background = element_blank(), panel.border = element_blank(), panel.grid = element_blank(), panel.spacing = unit(0, "lines"), plot.background = element_blank(), legend.justification = c(0,0), legend.position = c(0,0))
}

## local copies of unexported functions from ggplot2
.ggplot2_new_data_frame <- function(x=list(), n=NULL)
{
    if (length(x) != 0 && is.null(names(x))) { stop("Elements must be named") }
    lengths <- vapply(x, length, integer(1))
    if (is.null(n)) 
    {
        n <- if (length(x)==0 || min(lengths)==0) 0
        else max(lengths)
    }
    for (i in seq_along(x)) 
    {
        if (lengths[i] == n) next
        if (lengths[i] != 1) { stop("Elements must equal the number of rows or 1") }
        x[[i]] <- rep(x[[i]], n)
    }
    class(x) <- "data.frame"
    attr(x, "row.names") <- .set_row_names(n)
    
    return(x)
}

.ggplot2_xyz_to_isolines <- function(data, breaks)
{
    isoband::isolines(x=sort(unique(data$x)), y=sort(unique(data$y)), z=.ggplot2_isoband_z_matrix(data), levels=breaks)
}

.ggplot2_xyz_to_isobands <- function (data, breaks)
{
    isoband::isobands(x=sort(unique(data$x)), y=sort(unique(data$y)), z=.ggplot2_isoband_z_matrix(data), levels_low=breaks[-length(breaks)], levels_high=breaks[-1])
}

.ggplot2_isoband_z_matrix <- function (data)
{
    x_pos <- as.integer(factor(data$x, levels=sort(unique(data$x))))
    y_pos <- as.integer(factor(data$y, levels=sort(unique(data$y))))
    nrow <- max(y_pos)
    ncol <- max(x_pos)
    raster <- matrix(NA_real_, nrow=nrow, ncol=ncol)
    raster[cbind(y_pos, x_pos)] <- data$z
    
    return(raster)
}

.ggplot2_iso_to_path <- function (iso, group=1)
{
    lengths <- vapply(iso, function(x) length(x$x), integer(1))
    if (all(lengths == 0)) 
    {
        warning("stat_contour(): Zero contours were generated")
        return(.ggplot2_new_data_frame())
    }
    levels <- names(iso)
    xs <- unlist(lapply(iso, "[[", "x"), use.names=FALSE)
    ys <- unlist(lapply(iso, "[[", "y"), use.names=FALSE)
    ids <- unlist(lapply(iso, "[[", "id"), use.names=FALSE)
    item_id <- rep(seq_along(iso), lengths)
    groups <- paste(group, sprintf("%03d", item_id), sprintf("%03d", ids), sep="-")
    groups <- factor(groups)
    .ggplot2_new_data_frame(list(level=rep(levels, lengths), x=xs, y=ys, piece=as.integer(groups), group=groups), n=length(xs))
}

.ggplot2_iso_to_polygon <- function (iso, group=1)
{
    lengths <- vapply(iso, function(x) length(x$x), integer(1))
    if (all(lengths == 0)) {
        warning("stat_contour(): Zero contours were generated")
        return(.ggplot2_new_data_frame())
    }
    levels <- names(iso)
    xs <- unlist(lapply(iso, "[[", "x"), use.names=FALSE)
    ys <- unlist(lapply(iso, "[[", "y"), use.names=FALSE)
    ids <- unlist(lapply(iso, "[[", "id"), use.names=FALSE)
    item_id <- rep(seq_along(iso), lengths)
    groups <- paste(group, sprintf("%03d", item_id), sep="-")
    groups <- factor(groups)
    .ggplot2_new_data_frame(list(level=rep(levels, lengths), x=xs, y=ys, piece=as.integer(groups), group=groups, subgroup=ids), n=length(xs))
}

#####################################################################
## Compute tidy versions of ks function
#####################################################################

## default auxiliary function to compute ks object
## y = data frame
compute.ks <- function(y, args_ks, fun_ks)
{
    x <- as.matrix(dplyr::ungroup(y))
    if (length(args_ks)==0) args_ks <- list(x=x)
    else args_ks <- c(list(x=x), args_ks)
    fhat <- do.call(fun_ks, args_ks)
    if (is.matrix(fhat[["x"]])) 
        if (ncol(fhat[["x"]])==1) fhat[["x"]] <- as.vector(fhat[["x"]])
    fhat[["group"]] <- dplyr::group_vars(y)

    return(fhat)
}

## default auxiliary function to convert ks object to tidy format
## x=untidy ks object or tidy ks object
compute.tidy <- function(x, y, d, tidy)
{
    if (missing(tidy)) tidy <- inherits(x, "tbl")
    if (tidy) x <- getElement(dplyr::pull(x, "ks"),1)
    d <- as.integer(d)
    if (d==1)
        gg <- data.frame(x=x$eval.points, estimate=x$estimate)
    else if (d==2)
        gg <- data.frame(expand.grid(x=x$eval.points[[1]], y=x$eval.points[[2]]), estimate=as.vector(x$estimate))
    else if (d>2)
        stop("Data with d>2 not yet implemented.")

    ## keep copy of ks object in first row of ks column
    ## otherwise the data dimension
    gg$ks <- list(d)
    if (missing(y)) gg$ks[[1]] <- x else gg$ks[[1]] <- y
    gg <- dplyr::as_tibble(gg)

    return(gg)
}

## default function to compute tidy versions of ks::k* functions
## data = data frame
tidy_ks <- function(data, fun_ks, rename=TRUE, label, ...)
{
    if (is.vector(data)) data <- data.frame(data)

    ## compute grouped data frame of kernel estimate
    args_ks <- list(...)
    vars <- setdiff(names(data), dplyr::groups(data))

    d <- ncol(dplyr::select(dplyr::ungroup(data), dplyr::all_of(vars)))
    gg <- dplyr::group_modify(.data=data, .f=~dplyr::tibble(ks=list(compute.ks(.x, args_ks=args_ks, fun_ks=fun_ks))))
    gg <- dplyr::group_modify(.data=gg, .f=~compute.tidy(.x, d=d))

    ## add label for type of kernel estimate
    if (missing(label)) label <- fun_ks
    gg <- dplyr::mutate(gg, tks=fun_ks, label=label, .after="ks")

    ## rename gg$x, gg$y to data variable names
    if (rename) gg <- rename_ks(data=data, gg=gg, d=d)

    ## move grouping variable to last column
    gg <- move_group_vars(gg, y=data)

    ## create "tidy_ks" class
    gg <- as_tidy_ks(gg)

    return(gg)
}

## tidy version of ks::kde
## data = data frame
tidy_kde <- function(data, ...)
{
    ## compute KDE
    gg <- tidy_ks(data=data, fun_ks="kde", label="Density", ...)

    return(gg)
}

## tidy version of ks::kdcde
## data = data frame
tidy_kdcde <- function(data, ...)
{
    ## compute deconvolved KDE
    gg <- tidy_ks(data=data, fun_ks="kdcde", label="Deconv density", ...)

    return(gg)
}

## tidy version of ks::kcde
## data = data frame
tidy_kcde <- function(data,  ...)
{
    ## compute KCDE
    gg <- tidy_ks(data=data, fun_ks="kcde", label="Distribution", ...)

    return(gg)
}

## tidy version of ks::kcopula
## data = data frame
tidy_kcopula <- function(data, ...)
{
    ## compute KCDE
    gg <- tidy_ks(data=data, fun_ks="kcopula", label="Copula", ...)

    return(gg)
}

## tidy version of ks::kde.boundary
## data = data frame
tidy_kde_boundary <- function(data,  ...)
{
    ## compute boundary corrected KDE
    gg <- tidy_ks(data=data, fun_ks="kde.boundary", label="Bound density", ...)

    return(gg)
}

## tidy version of ks::kde.balloon
## data = data frame
tidy_kde_balloon <- function(data, ...)
{
    ## compute balloon variable KDE
    gg <- tidy_ks(data=data, fun_ks="kde.balloon", label="Balloon density", ...)

    return(gg)
}

## tidy version of ks::kde.sp
## data = data frame
tidy_kde_sp <- function(data, ...)
{
    ## compute sample point variable KDE
    gg <- tidy_ks(data=data, fun_ks="kde.sp", label="SP density", ...)

    return(gg)
}

## tidy version of ks::kde.truncate
## data = data frame
tidy_kde_truncate <- function(data, boundary, ...) { .tidy_kde_truncate(data, boundary, ...) }

.tidy_kde_truncate <- function(data, boundary, rename=TRUE, ...)
{
    if (is.vector(data)) data <- data.frame(data)

    ## compute truncated KDE
    compute.ks.local <- function(y, args_ks, boundary)
    {
        x <- as.matrix(dplyr::select(dplyr::ungroup(y), dplyr::all_of(vars)))
        boundary <- as.matrix(boundary)
        if (length(args_ks)==0) args_ks <- list(x=x)
        else args_ks <- c(list(x=x), args_ks)
        fhat <- do.call("kde", args_ks)
        fhat <- do.call("kde.truncate", list(fhat=fhat, boundary=boundary))
        if (is.matrix(fhat[["x"]])) 
            if (ncol(fhat[["x"]])==1) fhat[["x"]] <- as.vector(fhat[["x"]])
        fhat[["group"]] <- dplyr::group_vars(y)
        return(fhat)
    }

    ## compute grouped data frame of kernel estimate
    args_ks <- list(...)
    vars <- setdiff(names(data), dplyr::groups(data))
    d <- ncol(dplyr::select(dplyr::ungroup(data), dplyr::all_of(vars)))
    gg <- dplyr::group_modify(.data=data, .f=~dplyr::tibble(ks=list(compute.ks.local(.x, args_ks=args_ks, boundary=boundary))))
    gg <- dplyr::group_modify(.data=gg, .f=~compute.tidy(.x, d=d))
    if (rename) gg <- rename_ks(data=data, gg=gg, d=d)
    gg <- dplyr::mutate(gg, tks="kde.truncate", label="Trunc density", .before=dplyr::last_col())
    gg <- move_group_vars(gg, y=data)
    gg <- as_tidy_ks(gg)

    return(gg)
}

## tidy version of ks::kda
## data = data frame
tidy_kda <- function(data, ...) { .tidy_kda(data, ...) }

.tidy_kda <- function(data, rename=TRUE, ...)
{
    if (is.vector(data)) data <- data.frame(data)
    if (length(dplyr::group_vars(data))==0)
        stop("Requires input data to be grouped")
    else if (!all(is.factor(dplyr::pull(data, dplyr::group_vars(data)))))
        stop("Requires grouping variable to be a factor")

    ## compute KDA
    compute.ks.local <- function(y, args_ks)
    {
        x <- as.matrix(dplyr::select(dplyr::ungroup(y), dplyr::all_of(vars)))
        x.group <- dplyr::pull(y, dplyr::group_vars(y))
        if (length(args_ks)==0) args_ks <- list(x=x, x.group=x.group)
        else args_ks <- c(list(x=x, x.group=x.group), args_ks)

        fhat <- do.call("kda", args_ks)
        return(fhat)
    }

    ## convert to tidy
    compute.tidy.local <- function(y, d, rename)
    {
        j <- match(dplyr::pull(y, dplyr::group_vars(gg1)), levels(dplyr::pull(y, dplyr::group_vars(gg1))))
        x <- dplyr::pull(y, "ks")[[1]]
        xj <- x

        ## replace KDE with weighted KDE
        pp <- xj$prior.prob[j]
        xj$estimate <- x$estimate[[j]]*pp
        xj$H <- x$H[[j]]
        fhat.tidy <- dplyr::mutate(compute.tidy(x=xj, y=x, d=d, tidy=FALSE), prior.prob=pp, rn=dplyr::row_number())
        return(fhat.tidy)
    }

    ## compute grouped data frame of kernel estimate
    args_ks <- list(...)
    vars <- setdiff(names(data), dplyr::groups(data))
    d <- ncol(dplyr::select(dplyr::ungroup(data), dplyr::all_of(vars)))
    gg1 <- dplyr::summarise(data, .groups="keep")
    gg1 <- dplyr::mutate(gg1, ks=list(compute.ks.local(y=data, args_ks=args_ks)))
    gg <- dplyr::group_modify(gg1, .f=~compute.tidy.local(.x,d=d), .keep=TRUE)
    gg <- dplyr::rename(gg, prior_prob=dplyr::all_of("prior.prob"))

    ## create tidy version of partition (x_group)
    eval.points.group <- dplyr::summarise(dplyr::group_by(gg, .data$rn), label=levels(dplyr::pull(gg1,dplyr::group_vars(gg1)))[which.max(.data$estimate)], .groups="keep")
    gg <- dplyr::left_join(gg, eval.points.group, by="rn")
    gg <- dplyr::select(gg, -dplyr::one_of("rn"))
    ggv <- dplyr::group_vars(gg)[1]
    gg <- dplyr::mutate(gg, label=ifelse(.data$label==!!sym(ggv), .data$label, NA))

    if (rename) gg <- rename_ks(data=data, gg=gg, d=d)
    gg <- dplyr::relocate(gg, "prior_prob", .after="estimate")
    gg <- dplyr::mutate(gg, tks="kda", .before="label")
    gv <- dplyr::group_vars(gg1)[1]
    gg$label <- factor(gg$label, levels=levels(getElement(gg1,gv)))
    gg <- dplyr::relocate(gg, dplyr::group_vars(gg), .after=dplyr::last_col())
    gg <- as_tidy_ks(gg)

    return(gg)
}

## tidy version of ks::kdde
## data = data frame
tidy_kdde <- function(data, deriv_order=1, ...) { .tidy_kdde(data, deriv_order=deriv_order, ...) }

.tidy_kdde <- function(data, deriv_order=1, rename=TRUE, ...)
{
    if (is.vector(data)) data <- data.frame(data)

    ## compute KDDE
    compute.ks.local <- function(y, args_ks)
    {
        x <- as.matrix(dplyr::ungroup(y))
        if (length(args_ks)==0) args_ks <- list(x=x, deriv.order=deriv_order)
        else args_ks <- c(list(x=x, deriv.order=deriv_order), args_ks)
        fhat <- do.call("kdde", args_ks)
        if (is.matrix(fhat[["x"]])) 
            if (ncol(fhat[["x"]])==1) fhat[["x"]] <- as.vector(fhat[["x"]])
        return(fhat)
    }

    ## convert to tidy
    compute.tidy.local <- function(y, d)
    {
        x <- dplyr::pull(y, "ks")[[1]]
        deriv.order <- as.integer(x$deriv.order)
        deriv.ind <- x$deriv.ind
        if (d==1)
        {
            x$deriv.ind.scalar <- 1
            fhat.tidy <- dplyr::mutate(compute.tidy(x=x, d=d, tidy=FALSE), deriv_order=deriv.order, deriv_ind=1L, deriv_group=deriv.order)
        }
        else if (d==2)
        {
            for (j in 1:nrow(deriv.ind))
            {
                xj <- x
                x$deriv.ind.scalar <- j
                deriv.indj <- paste0("(",paste0(deriv.ind[j,], collapse=","),")")
                xj$estimate <- x$estimate[[j]]
                if (j==1) fhat.tidy <- dplyr::mutate(compute.tidy(xj, y=x, d=d, tidy=FALSE), deriv_order=deriv.order, deriv_ind=j, deriv_group=deriv.indj)
                else fhat.tidy <- rbind(fhat.tidy, dplyr::mutate(compute.tidy(xj, y=x, d=d, tidy=FALSE), deriv_order=deriv.order, deriv_ind=j, deriv_group=deriv.indj))
            }
        }

        return(fhat.tidy)
    }

    ## compute grouped data frame of kernel estimate
    args_ks <- list(...)
    vars <- setdiff(names(data), dplyr::groups(data))
    d <- ncol(dplyr::select(dplyr::ungroup(data), dplyr::all_of(vars)))
    gg <- dplyr::group_modify(data, ~dplyr::tibble(ks=list(compute.ks.local(y=.x, args_ks=args_ks))))
    gg <- dplyr::group_modify(gg, ~compute.tidy.local(y=.x, d=d))
    if (rename) gg <- rename_ks(data=data, gg=gg, d=d)
    gg$deriv_group <- factor(paste("deriv", gg$deriv_group))
    gg <- dplyr::mutate(gg, tks="kdde", label="Density deriv", .before=dplyr::last_col())
    gg <- move_group_vars(gg, data)
    gg <- dplyr::relocate(gg, "deriv_group", .after=dplyr::last_col())
    gg <- dplyr::relocate(gg, "deriv_order", "deriv_ind", .after="estimate")
    gg <- dplyr::group_by(gg, .data$deriv_group, .add=TRUE)
    gg$deriv_group <- factor(gg$deriv_group , levels=rev(levels(gg$deriv_group)))
    gg <- as_tidy_ks(gg)

    return(gg)
}

## tidy version of ks::kde_local_test
## data1, data2 = data frames
tidy_kde_local_test <- function(data1, data2, labels, ...) 
{ 
    tklt <- .tidy_kde_local_test(data1=data1, data2=data2, rename=TRUE, ...) 
    if (missing(labels)) 
    {
        mc <- as.character(match.call())[2:3]
        labels <- gsub("f1", mc[1], levels(tklt$label))
        labels <- gsub("f2", mc[2], labels)
    }
    else labels <- c(paste0(labels[1],"<",labels[2]), NA, paste0(labels[1],">",labels[2]))
    levels(tklt$label) <- labels

    count <- dplyr::summarise(tklt, count=dplyr::n())
    count <- head(c(1, cumsum(count$count)+1), n=-1)
    for (i in count)
    {
        fhat <- tklt$ks[[i]]
        fhat$label <- levels(tklt$label)
        tklt$ks[[i]] <- fhat
    }

    return(tklt)
}

.tidy_kde_local_test <- function(data1, data2, rename=TRUE, ...)
{
    if (is.vector(data1)) data1 <- data.frame(data1)
    if (is.vector(data2)) data2 <- data.frame(data2)

    ## compute KDDE
    compute.ks.local <- function(y, args_ks)
    {
        y1 <- dplyr::filter(y, .data$.group=="data1")
        y2 <- dplyr::filter(y, .data$.group=="data2")
        y1 <- dplyr::select(y1, -dplyr::all_of(".group"))
        y2 <- dplyr::select(y2, -dplyr::all_of(".group"))
        x1 <- as.matrix(dplyr::select(dplyr::ungroup(y1), dplyr::all_of(vars)))
        x2 <- as.matrix(dplyr::select(dplyr::ungroup(y2), dplyr::all_of(vars)))
        if (ncol(x1)==1) x1 <- as.vector(x1)
        if (ncol(x2)==1) x2 <- as.vector(x2)
        if (length(args_ks)==0) args_ks <- list(x1=x1, x2=x2)
        else args_ks <- c(list(x1=x1, x2=x2), args_ks)
        fhat <- do.call("kde.local.test", args_ks)
        fhat$names <- names(y1)
        return(fhat)
    }

    ## convert to tidy
    compute.tidy.local <- function(y, d)
    {
        x <- dplyr::pull(y, "ks")[[1]]
        fhat.diff <- x$fhat1
        fhat.diff$estimate <- x$fhat.diff
        fhat.diff.lab <- x$fhat1
        fhat.diff.lab$estimate <- x$fhat.diff.pos$estimate - x$fhat.diff.neg$estimate
        fhat.tidy <- compute.tidy(fhat.diff, d=d, tidy=FALSE)
        diff.lab <- compute.tidy(fhat.diff.lab, d=d, tidy=FALSE)$estimate
        diff.lab <- factor(diff.lab, levels=c(-1,0,1), labels=c("f1<f2", NA, "f1>f2"))
        fhat.tidy <- dplyr::mutate(fhat.tidy, label=diff.lab)
        
        fhat.tidy[1,]$ks <- list(x)
        return(fhat.tidy)
    }

    ## compute grouped data frame of kernel estimate
    args_ks <- list(...)
    vars <- setdiff(names(data1), dplyr::groups(data1))
    d <- ncol(dplyr::select(dplyr::ungroup(data1), dplyr::all_of(vars)))
    ddata <- dplyr::bind_rows(dplyr::mutate(data1, .group="data1"), dplyr::mutate(data2, .group="data2"))
    gg <- dplyr::group_modify(ddata, ~dplyr::tibble(ks=list(compute.ks.local(y=.x, args_ks=args_ks))))
    gg <- dplyr::group_modify(gg, ~compute.tidy.local(.x, d=d))
    gg <- dplyr::mutate(gg, tks="kde.loctest", .after="ks")
    gg <- dplyr::relocate(gg, "ks", .after="estimate")
    if (rename) gg <- rename_ks(data=data1, gg=gg, d=d)
    gg <- move_group_vars(gg, data1)
    gg <- as_tidy_ks(gg)

    return(gg)
}

## tidy version of ks::kroc
## data1, data2 = data frames
tidy_kroc <- function(data1, data2, ...)
{
    if (is.vector(data1)) data1 <- data.frame(data1)
    if (is.vector(data2)) data2 <- data.frame(data2)

    ## compute KDDE
    compute.ks.local <- function(y1, y2, args_ks)
    {
        x1 <- as.matrix(dplyr::select(dplyr::ungroup(y1), dplyr::all_of(vars)))
        x2 <- as.matrix(dplyr::select(dplyr::ungroup(y2), dplyr::all_of(vars)))
        if (length(args_ks)==0) args_ks <- list(x1=x1, x2=x2)
        else args_ks <- c(list(x1=x1, x2=x2), args_ks)
        fhat <- do.call("kroc", args_ks)
        return(fhat)
    }

    ## convert to tidy
    compute.tidy.local <- function(y, d=1)
    {
        x <- dplyr::pull(y, "ks")[[1]]
        fhat.tidy <- compute.tidy(x,d=d)
        fhat.tidy[1,]$ks <- list(x)

        return(fhat.tidy)
    }

    ## compute grouped data frame of kernel estimate
    args_ks <- list(...)
    vars <- setdiff(names(data1), dplyr::groups(data1))
    gg <- dplyr::group_modify(data1, ~dplyr::tibble(ks=list(compute.ks.local(y1=.x, y2=data2, args_ks=args_ks))))
    gg <- dplyr::group_modify(gg, ~compute.tidy(.x, d=1L))
    gg <- dplyr::rename(gg, fpr=dplyr::all_of("x"))
    gg <- dplyr::mutate(gg, tks="kroc", label="ROC", .before=dplyr::last_col())
    gg <- move_group_vars(gg, data1)
    gg <- as_tidy_ks(gg)

    return(gg)
}

## tidy version of ks::kfs
## data = data frame
tidy_kfs <- function(data, ...) 
{ 
    labels <- as.character(match.call())[2]
    .tidy_kfs(data, labels=labels, ...) 
}

.tidy_kfs <- function(data, rename=TRUE, labels, ...)
{
    ## convert to tidy
    compute.tidy.local <- function(y, d)
    {
        x <- dplyr::pull(y, "ks")[[1]]
        fhat2 <- x
        if (d==1)
            fhat2$estimate <- ks::kde(x=x$x, eval.points=x$eval.points)$estimate
        else
            fhat2$estimate <- ks::kde(x=x$x, eval.points=expand.grid(x$eval.points))$estimate
        class(fhat2) <- "kde"
        fhat.tidy <- compute.tidy(fhat2, d=d, tidy=FALSE)
        signif.group <- factor(x$estimate, levels=c(0,1), labels=c(NA, labels)) ##factor(x$estimate, levels=c(0,1), labels=c(NA,"Signif curv"))
        fhat.tidy <- dplyr::mutate(fhat.tidy, label=signif.group)
        x$cont <- 0.5; names(x$cont) <- "50%"
        x$label <- labels #"Signif curv"
        fhat.tidy[1,]$ks <- list(x)
        return(fhat.tidy)
    }

    ## compute grouped data frame of kernel estimate
    args_ks <- list(...)
    vars <- setdiff(names(data), dplyr::groups(data))
    d <- ncol(dplyr::select(dplyr::ungroup(data), dplyr::all_of(vars)))
    gg <- dplyr::group_modify(data, ~dplyr::tibble(ks=list(compute.ks(.x, fun_ks="kfs", args_ks=args_ks))))
    gg <- dplyr::group_modify(gg, ~compute.tidy.local(.x, d=d))
    if (rename) gg <- rename_ks(data=data, gg=gg, d=d)
    gg <- dplyr::relocate(gg, "ks", .after="estimate")
    gg <- dplyr::mutate(gg, tks="kfs", .after="ks")
    gg <- move_group_vars(gg, data)
    gg <- as_tidy_ks(gg)

    return(gg)
}

## tidy version of ks::kdr
## data = data frame
tidy_kdr <- function(data, dTolerance, ...) 
{ 
    labels <- as.character(match.call())[2]
    .tidy_kdr(data, dTolerance=dTolerance, labels=labels, ...) 
}

.tidy_kdr <- function(data, rename=TRUE, dTolerance, labels, ...)
{
    if (is.vector(data)) data <- data.frame(data)

    compute.tidy.local <- function(y, d)
    { 
        x <- dplyr::pull(y, "ks")[[1]]
        end.points <- data.frame(x$end.points)
        names(end.points)[1:d] <- c("x","y","z")[1:d]
        fhat.tidy <- dplyr::as_tibble(end.points)
        fhat.tidy$estimate <- 0
        fhat.tidy$ks <- list(d)
        fhat.tidy$ks[[1]] <- x
        return(fhat.tidy)
    }

    ## compute grouped data frame of kernel estimate
    args_ks <- list(...)
    vars <- setdiff(names(data), dplyr::groups(data))
    d <- ncol(dplyr::select(dplyr::ungroup(data), dplyr::all_of(vars)))
    gg <- dplyr::group_modify(.data=data, .f=~dplyr::tibble(ks=list(compute.ks(.x, args_ks=args_ks, fun_ks="kdr"))))
    gg <- dplyr::group_modify(.data=gg, .f=~compute.tidy.local(.x, d=d))
    gv <- dplyr::group_vars(data) 
    gg <- dplyr::group_by(gg, dplyr::across("segment"), .add=TRUE) 

    ## reduce density of points on KDR
    if (missing(dTolerance))
    {
        ## median distance of consecutive point in segmentsgg
        gg3 <- dplyr::filter(gg, dplyr::n()>4)
        gg3 <- dplyr::group_modify(gg3, .f=~sf::st_as_sf(.x, coords=c("x","y")))
        gg3 <- sf::st_sf(gg3)
        gg3 <- dplyr::group_modify(gg3, .f=~dplyr::tibble(geom1=head(.x$geometry, n=-1), geom2=tail(.x$geometry, n=-1)))
        gg3 <- dplyr::summarise(gg3, dist=median(as.numeric(sf::st_distance(.data$geom1, .data$geom2, by_element=TRUE))))
        dTolerance <- median(gg3$dist)
        dTolerance <- min(10^(round(log10(mean(dTolerance))-1,0)), 10)
    }

    gg2 <- dplyr::group_modify(gg, ~data.frame(.x[1,], sf::st_sf(geometry=sf::st_sfc(sf::st_linestring(cbind(.x$x,.x$y))))))
    gg2 <- dplyr::select(gg2, -c("x","y"))
    gg2 <- sf::st_sf(gg2)
    gg2 <- gg2[sf::st_length(gg2)>0,] ## remove singleton segments
    gg2 <- sf::st_simplify(gg2, dTolerance=dTolerance)
    gg2 <- st_add_coordinates(gg2, rename=FALSE)
    gg2 <- dplyr::relocate(gg2, "X", "Y", .before=1)
    gg2 <- dplyr::rename(gg2, x="X", y="Y")
    gg <- dplyr::mutate(gg2, ks=ifelse(dplyr::row_number()>1, d, .data$ks))
    if (rename) gg <- rename_ks(data=data, gg=gg, d=d)
    gg <- dplyr::mutate(gg, estimate=factor(.data$estimate), tks="kdr", label=factor(labels), .before=dplyr::last_col())
    gg <- dplyr::group_by(gg, dplyr::across("segment"), .add=TRUE)
    gg <- move_group_vars(gg, data)
    gg <- dplyr::relocate(gg, "segment", .after="label") 
    sfname <- attr(gg2, "sf_column")
    gg <- dplyr::relocate(gg, dplyr::all_of(sfname), .after=dplyr::last_col())
    gg <- as_tidy_ks(gg)
    
    return(gg)
}

## tidy version of ks::kdr.segment
## data = tidy KDR object
tidy_kdr_segment <- function(data, dTolerance, coords=1:2, ...)
{
    gv <- dplyr::group_vars(data) 
    gv1 <- setdiff(gv, "segment")
    
    ## compute KDR segments
    compute.ks.local <- function(y, args_ks)
    {
        x <- untidy_ks(y)
        if (length(args_ks)==0) args_ks <- list(x=x)
        else args_ks <- c(list(x=x), args_ks)
        fhat <- do.call("kdr.segment", args_ks)

        return(fhat)
    }
    
    compute.tidy.local <- function(y, d=1)
    {
        x <- dplyr::pull(y, "ks")[[1]]
        fhat.tidy <- dplyr::tibble(x$end.points)
        fhat.tidy$estimate <- NA
        fhat.tidy$ks <- list(d)
        fhat.tidy$ks[[1]] <- x

        return(fhat.tidy)
    }

    if (length(gv1)>0) gg <- dplyr::group_by(data, dplyr::across(dplyr::all_of(gv1))) else gg <- dplyr::ungroup(data)
    gg <- dplyr::slice_head(gg, n=1)
    args_ks <- list(...)
    gg <- dplyr::group_modify(.data=gg, .f=~dplyr::tibble(ks=list(compute.ks.local(.x, args_ks=args_ks))))
    gg <- dplyr::group_modify(.data=gg, .f=~compute.tidy.local(.x, d=dim_ks(data)))
    gg <- dplyr::group_by(gg, dplyr::across(dplyr::all_of(gv)))

    if (any(names(data) %in% "dTolerance")) dTolerance <- head(data$dTolerance,n=1)
    if (missing(dTolerance))
    {
        ## median distance of consecutive point in segmentsgg
        gg3 <- dplyr::filter(gg, dplyr::n()>4)
        gg3 <- dplyr::group_modify(gg3, .f=~sf::st_as_sf(.x, coords=c("x","y")))
        gg3 <- sf::st_sf(gg3)
        gg3 <- dplyr::group_modify(gg3, .f=~dplyr::tibble(geom1=head(.x$geometry, n=-1), geom2=tail(.x$geometry, n=-1)))
        gg3 <- dplyr::summarise(gg3, dist=median(as.numeric(sf::st_distance(.data$geom1, .data$geom2, by_element=TRUE))))
        dTolerance <- median(gg3$dist)
        dTolerance <- min(10^(round(log10(mean(dTolerance))-1,0)), 10)
    }
    
    gg <- dplyr::relocate(gg, "segment", .after="estimate")
    gg$segment <- factor(gg$segment)   
    gg <- rename_ks(data=data, gg=gg, d=dim_ks(data))
    gg <- dplyr::mutate(gg, tks="kdr", label="Density ridge", .before=dplyr::last_col())
    gg <- move_group_vars(gg, data)
    gg <- as_tidy_ks(gg)

    return(gg)
}

## tidy version of ks::kms
## data = data frame
tidy_kms <- function(data, ...) { .tidy_kms(data, ...) }

.tidy_kms <- function(data, rename=TRUE, ...)
{
    if (is.vector(data)) data <- data.frame(data)

    compute.tidy.local <- function(y, d)
    {
        x <- dplyr::pull(y, "ks")[[1]]
        fhat.tidy <- data.frame(cbind(x$x, x$label))
        names(fhat.tidy)[1:d] <- c("x","y","z")[1:d]
        names(fhat.tidy)[d+1] <- "estimate"

        fhat.tidy <- dplyr::as_tibble(fhat.tidy)
        fhat.tidy <- dplyr::mutate(fhat.tidy, estimate=factor(.data$estimate), label=.data$estimate)
        fhat.tidy$ks <- list(d)
        fhat.tidy$ks[[1]] <- x
        return(fhat.tidy)
    }

    ## compute grouped data frame of kernel estimate
    args_ks <- list(...)
    vars <- setdiff(names(data), dplyr::groups(data))
   	d <- ncol(dplyr::select(dplyr::ungroup(data), dplyr::all_of(vars)))
    gg <- dplyr::group_modify(.data=data, .f=~dplyr::tibble(ks=list(compute.ks(.x, args_ks=args_ks, fun_ks="kms"))))
    gg <- dplyr::group_modify(.data=gg, .f=~compute.tidy.local(.x, d=d))
    if (rename) gg <- rename_ks(data=data, gg=gg, d=d)
    gg <- dplyr::mutate(gg, tks="kms", label=.data$estimate, .after="ks")
    gg <- dplyr::relocate(gg, "label", .after=dplyr::last_col())
    gg <- move_group_vars(gg, data)
    gg <- as_tidy_ks(gg)

    return(gg)
}

## tidy version of kernel support estimate
## data = tidy kde object
tidy_ksupp <- function(data, cont=95, convex_hull=TRUE, ...) { .tidy_ksupp(data, cont=cont, convex_hull=convex_hull, ...) }

.tidy_ksupp <- function(data, cont=95, convex_hull=TRUE, rename=TRUE, ...)
{
    if (is.vector(data)) data <- data.frame(data)

    compute.ks.local <- function(y, args_ks, cont, convex_hull)
    {
        x <- dplyr::pull(y, "ks")[[1]]
        if (length(args_ks)==0) args_ks <- list(fhat=x, cont=cont, convex.hull=convex_hull)
        else args_ks <- c(list(fhat=x, cont=cont, convex.hull=convex_hull), args_ks)
        fhat <- do.call("ksupp", args_ks)
        names(fhat) <- names(data2)[1:2]
        return(fhat)
    }

    compute.tidy.local <- function(y, d)
    {
        x <- dplyr::pull(y, "ks")[[1]]
        fhat.tidy <- dplyr::as_tibble(data.frame(x=x[,1], y=x[,2], estimate=0))
        fhat.tidy$ks <- list(d)
        fhat.tidy$ks[[1]] <- x
        return(fhat.tidy)
    }
    
    ## compute grouped data frame of kernel estimate
    args_ks <- list(...)
    d <- ncol(get_data_ks(data[1,]))
    data2 <- dplyr::select_if(dplyr::ungroup(data), ~!is.factor(.))

    gg <- dplyr::group_modify(data, ~dplyr::tibble(ks=list(compute.ks.local(y=.x, cont=cont, convex_hull=convex_hull, args_ks=args_ks))))
    gg <- dplyr::group_modify(gg, ~compute.tidy.local(.x, d=d))

    gg <- dplyr::mutate(gg, estimate=factor(.data$estimate), tks="ksupp", label=paste0(cont,"%"), .before=dplyr::last_col())
    if (rename) gg <- rename_ks(data=data2, gg=gg, d=d)
    gg <- move_group_vars(gg, data)
    gg <- as_tidy_ks(gg)

    return(gg)
}

## tidy version of thinned quiver plot from first density derivative estimate
## data = tidy kdde object
tidy_kquiver <- function(data, thin=5, transf=1/4, neg.grad=FALSE) { .tidy_kquiver(data, thin=thin, transf=transf, neg.grad=neg.grad, rename=TRUE) }

.tidy_kquiver <- function(data, thin=5, transf=1/4, neg.grad=FALSE, rename=TRUE)
{
    ## compute quiver
    compute.tidy.local <- function(y)
    {
        fhat <- untidy_ks(y)
        while (!(class(fhat) %in% "kdde")) { fhat <- fhat[[1]] }
        if (fhat$deriv.order!=1) stop("quiver plots are defined only for 2D data with deriv_order=1.")

        ## create tidy version suitable for geom_quiver
        ev <- fhat$eval.points
        est <- fhat$estimate
        if (transf != 0) {
            est[[1]] <- sign(est[[1]]) * abs(est[[1]])^transf
            est[[2]] <- sign(est[[2]]) * abs(est[[2]])^transf
        }
        thin1.ind <- seq(1, length(ev[[1]]), by=thin)
        thin2.ind <- seq(1, length(ev[[2]]), by=thin)
        fx <- est[[1]][thin1.ind, thin2.ind]
        fy <- est[[2]][thin1.ind, thin2.ind]
        if (neg.grad) {
            fx <- -fx
            fy <- -fy
        }
        xy <- expand.grid(x=ev[[1]][thin1.ind], y=ev[[2]][thin2.ind])
        xyuv <- data.frame(xy, u=as.vector(fx), v=as.vector(fy))
        xyuv$ks <- list(2L)
        xyuv$ks[[1]] <- fhat
        return(xyuv)
    }

    ## compute grouped data frame of quiver estimate
    ## coalesce groups for each derivative into 1 group
    if (dplyr::is_grouped_df(data)) data1 <- data
    else data1 <- dplyr::group_by(data, .data$deriv_group)
    data1$.group <- dplyr::group_indices(data1)
    gg1 <- dplyr::filter(data1, .data$.group%%2 == 1)
    gg2 <- dplyr::slice_tail(gg1)
    gg1 <- dplyr::group_modify(gg1, ~compute.tidy.local(.x), .keep=TRUE)
    jv <- "deriv_group"
    if (dplyr::is_grouped_df(data)) jv <- dplyr::group_vars(data)
    data2 <- dplyr::select_if(dplyr::ungroup(data), ~!is.factor(.))
    if (rename) gg1 <- rename_ks(data=data2, gg=gg1, d=2)
    if (rename)
        gg <- dplyr::left_join(gg1, dplyr::select(gg2, -dplyr::one_of(names(data2)[1:2], "ks")), by=jv)
    else
        gg <- dplyr::left_join(gg1, dplyr::select(gg2, -dplyr::one_of("x","y", "ks")), by="group")
    if (length(jv)>1)
    {   
        jv <- setdiff(jv, "deriv_group")

        ##gg <- dplyr::group_by(gg, !!sym(jv))
        gg <- dplyr::group_by(gg, dplyr::across(dplyr::all_of(jv)))
    }
    else gg <- dplyr::ungroup(gg)
    gg <- dplyr::relocate(gg, dplyr::all_of(jv), .before=dplyr::last_col())
    gg <- dplyr::relocate(gg, "estimate", .before="u")
    gg <- dplyr::relocate(gg, "ks", .before="tks")
    gg <- dplyr::mutate(gg, tks="kquiver", label="Density quiver", .after="ks")
    gg <- dplyr::select(gg, -dplyr::one_of(c("deriv_ind","deriv_order","deriv_group",".group")))
    gg <- as_tidy_ks(gg)

    return(gg)
}

## tidy version of ks:::kcurv
## data = tidy kdde object
tidy_kcurv <- function(data, ...) { .tidy_kcurv(data, ...) }

.tidy_kcurv <- function(data, rename=TRUE, ...)
{
    ## compute curvature
    compute.ks.local <- function(y)
    {
        fhat <- untidy_ks(y)
        while (!(class(fhat) %in% "kdde")) { fhat <- fhat[[1]] }
        if (fhat$deriv.order!=2) stop("Curvature plots are defined only for 2D data with deriv_order=2")
        fhat <- ks::kcurv(fhat=fhat)
        return(fhat)
    }

    ## compute grouped data frame of kernel estimate
    data <- dplyr::filter(data, .data$deriv_ind %in% 1)
    d <- 2
    gg <- dplyr::group_modify(.data=data, .f=~dplyr::tibble(ks=list(compute.ks.local(.x))))
    gg <- dplyr::group_modify(.data=gg, .f=~compute.tidy(.x, d=d))

    if (dplyr::is_grouped_df(data)) jv <- dplyr::group_vars(data)
    if (length(jv)>1)
    {   
        jv2 <- setdiff(jv, "deriv_group")
        gg <- dplyr::group_by(gg, dplyr::across(dplyr::all_of(jv2)))
    }
    else gg <- dplyr::ungroup(gg)
    gv <- head(names(dplyr::select_if(data, ~is.factor(.))),1)
    data2 <- dplyr::select_if(dplyr::ungroup(data), ~!is.factor(.))
    if (rename) gg <- rename_ks(data=data2, gg=gg, d=d)
    gg <- dplyr::mutate(gg, tks="kcurv", label="Density curv")
    if (length(jv)>1) gg <- dplyr::relocate(gg, dplyr::all_of(jv), .after=dplyr::last_col())
    gg <- dplyr::select(gg, -dplyr::all_of("deriv_group"))
    gg <- as_tidy_ks(gg)

    return(gg)
}

## tidy version of ks:::histde
tidy_histde <- function(data, ...) { .tidy_histde(data, ...) }

.tidy_histde <- function(data, rename=TRUE, ...)
{
    if (is.vector(data)) data <- data.frame(data)

    ## convert to tidy
    compute.tidy.local <- function(y, d)
    {
        x <- dplyr::pull(y, "ks")[[1]]
        x$eval.points <- tail(x$eval.points, n=-1)
        fhat.tidy <- dplyr::mutate(compute.tidy(x=x, d=d, tidy=FALSE))
    
        return(fhat.tidy)
    }

    ## compute grouped data frame of kernel estimate
    args_ks <- list(...)
    vars <- setdiff(names(data), dplyr::groups(data))
    d <- ncol(dplyr::select(dplyr::ungroup(data), dplyr::all_of(vars)))
    gg <- dplyr::group_modify(data, ~dplyr::tibble(ks=list(compute.ks(y=.x, args_ks=args_ks, fun_ks="histde"))))
    gg <- dplyr::group_modify(gg, ~compute.tidy.local(y=.x, d=d))
    if (rename) gg <- rename_ks(data=data, gg=gg, d=d)
    gg <- dplyr::mutate(gg, tks="histde", label="Hist", .before=dplyr::last_col())
    gg <- move_group_vars(gg, data)
    gg <- as_tidy_ks(gg)

    return(gg)
}


## tidy version of ks:::as.kde
## data = estimation grid + list of evaluation points
tidy_as_kde <- function(data, density, ...) 
{  
    is_tidy <- !any(class(data) %in% "list")
   .tidy_as_kde(data=data, density=density, is_tidy=is_tidy, ...) 
}

.tidy_as_kde <- function(data, density, rename=TRUE, is_tidy, ...)
{
    ## compute grouped data frame of kernel estimate
    args_ks <- list(...)
   
    compute.ks.local <- function(y, args_ks)
    {
        if (length(args_ks)==0) args_ks <- list(x=y, density=density)
        else args_ks <- c(list(x=y, density=density), args_ks)
        fhat <- do.call("as.kde", args_ks)
        if (is.matrix(fhat[["x"]])) 
            if (ncol(fhat[["x"]])==1) fhat[["x"]] <- as.vector(fhat[["x"]])
        
        return(fhat)
    }

    ## convert to tidy
    compute.tidy.local <- function(y, d)
    {
        x <- dplyr::pull(y, "ks")[[1]]
        deriv.order <- as.integer(x$deriv.order)
        deriv.ind <- x$deriv.ind
        if (d==1)
        {
            x$deriv.ind.scalar <- 1
            fhat.tidy <- dplyr::mutate(compute.tidy(x=x, d=d, tidy=FALSE), deriv_order=deriv.order, deriv_ind=1L, deriv_group=deriv.order)
        }
        else if (d==2)
        {
            for (j in 1:nrow(deriv.ind))
            {
                xj <- x
                x$deriv.ind.scalar <- j
                deriv.indj <- paste0("(",paste0(deriv.ind[j,], collapse=","),")")
                xj$estimate <- x$estimate[[j]]
                if (j==1) fhat.tidy <- dplyr::mutate(compute.tidy(xj, y=x, d=d, tidy=FALSE), deriv_order=deriv.order, deriv_ind=j, deriv_group=deriv.indj)
                else fhat.tidy <- rbind(fhat.tidy, dplyr::mutate(compute.tidy(xj, y=x, d=d, tidy=FALSE), deriv_order=deriv.order, deriv_ind=j, deriv_group=deriv.indj))
            }
        }

        return(fhat.tidy)
    }

    if (is_tidy)
    {
        d <- ncol(data)-1
        data <- as.data.frame(data)
    }
    else 
    {
        if (!is.list(data$eval.points)) d <- 1 else d <- length(data$eval.points)
        eval.points <- expand.grid(data$eval.points)
        if (all(is.null(names(data$eval.points)))) colnames(eval.points) <- c("x","y","z")[1:d]
        data <- data.frame(eval.points, estimate=as.vector(data$estimate))
    }
    ## reduce #s.f. to ensure exactly replicated values in data[,1:d] 
    data[,1:d] <- signif(data[,1:d],10)
    if (missing(density)) 
    {
        if (all(data$estimate>=0)) density <- TRUE
        else if (all(data$estimate<0)) density <- FALSE
        else density <- (abs(mean(data$estimate[data$estimate<0]))/mean(data$estimate[data$estimate>=0]) < 1e-6) 
    }
    
    gg <- dplyr::tibble(ks=list(compute.ks.local(data, args_ks=args_ks)))
    
    ## add label for type of kernel estimate
    if (density) 
    {
        gg <- compute.tidy(gg, d=d)
        gg <- dplyr::mutate(gg, tks="kde", label="Density", .after="ks")
    }
    else
    {
        gg <- compute.tidy.local(gg, d=d)
        gg <- dplyr::mutate(gg, tks="kdde", label="Density", .after="ks")
        gg <- dplyr::group_by(gg, .data$deriv_group, .add=TRUE)
    }
    
    ## rename gg$x, gg$y to data variable names
    if (rename) gg <- rename_ks(data=data, gg=gg, d=d)
    
    ## move grouping variable to last column
    gg <- move_group_vars(gg, y=data)

    ## create "tidy_ks" class
    gg <- as_tidy_ks(gg)

    return(gg)
}

## data = complete sf polygon grid 
## or incomplete point tidy matrix
## convert data to tidy matrix of eval points + 
## density (to be suitable input for ks::as.kde or eks::tidy_as_kde)
tidy_intergrid <- function(data, attrib, cellsize, verbose=FALSE)
{        
    ## spatial data
    if (inherits(data, "sf"))
    {    
        if (missing(attrib)) attrib <- 1
        if (is.numeric(attrib)) attrib <- names(data)[attrib]
        ## add grid cell indices if missing
        if (!all(names(data) %in% c("cell_id", "cell_id1", "cell_id2")))
            data <- st_add_index(data)

        ## replace polygon by centroids 
        #xgrid <- suppressWarnings(sf::st_centroid(data))
        #xgrid <- dplyr::distinct(xgrid, .data$geometry, .keep_all=TRUE)
        xgrid <- st_centroid2(data)
        xgrid <- st_add_coordinates(xgrid) 
        xgrid <- dplyr::arrange(xgrid, .data$cell_id2, .data$cell_id1)
        rownames(xgrid) <- NULL
        
        ## create tidy data frame with lon, lat + attrib
        eval.points <- sf::st_drop_geometry(xgrid[,c("lon", "lat")])
        eval.points <- signif(eval.points, 10)
        estimate <- xgrid[[attrib]]
        estimate[is.na(estimate)] <- 0
        estimate <- data.frame(eval.points, estimate) 
    }
    else ## non-spatial data
    {
        d <- ncol(data)-1
        coords <- 1:d 
        coord.names <- colnames(data)[coords]
        attrib <- d+1
        if (is.numeric(attrib)) attrib <- names(data)[attrib]

        datasf <- sf::st_as_sf(data, coords=coords) 
        data2 <- sf::st_drop_geometry(st_add_coordinates(datasf)) 
        data2.lon <- sort(unique(signif(data2$lon,10)))
        data2.lat <- sort(unique(signif(data2$lat,10)))
        if (missing(cellsize))
        {
            dlon <- signif(diff(data2.lon),4)
            dlat <- signif(diff(data2.lat),4)
            dlon <- dlon[dlon>0]
            dlat <- dlat[dlat>0]
            cellsize <- c(as.numeric(names(which.max(table(dlon)))),  as.numeric(names(which.max(table(dlat)))))
        } 
        data2.lon <- seq(min(data2.lon), max(data2.lon), by=cellsize[1]) 
        data2.lat <- seq(min(data2.lat), max(data2.lat), by=cellsize[2])
        
        xgrid <- expand.grid(lon=data2.lon, lat=data2.lat)
        xgrid <- dplyr::arrange(xgrid, .data$lat, .data$lon)
        xgrid <- sf::st_as_sf(xgrid, coords=1:2)
        rownames(xgrid) <- NULL
        xgrid <- dplyr::distinct(xgrid, .data$geometry, .keep_all=TRUE)
        xgrid <- st_add_coordinates(xgrid)

        ## join attributes from data onto xgrid
        xgrid <- dplyr::mutate(xgrid, point_id=1:nrow(xgrid), .before=1)
        xgrid <- sf::st_join(xgrid, sf::st_buffer(datasf, dist=min(cellsize)/4))
        rownames(xgrid) <- NULL
        
        ## create tidy data frame with lon, lat + attrib
        eval.points <- sf::st_drop_geometry(xgrid[,c( "lon", "lat")])
        eval.points <- signif(eval.points, 10)
        estimate <- xgrid[[attrib]]
        estimate[is.na(estimate)] <- 0
        estimate <- data.frame(eval.points, estimate) 
        colnames(estimate)[colnames(estimate) %in% c("lon","lat")] <- coord.names
    }

    if (verbose) cat(paste0("Original grid     sum(", attrib, ") = ", sum(data[[attrib]], na.rm=TRUE)), paste0("\nInterpolated grid sum(", attrib, ") = ", sum(estimate$estimate)), "\n") 
    return(estimate)
}

#####################################################################
## Auxiliary functions for tidy_*/st_* functions
#####################################################################

## create more informative legend labels
## contlabel = contour percentage (without %)
## contperc = contour percentage (with %)
## estimate/level = contour level
## contregion = contour level (with >= or <=)
create_label <- function(x, digits, add_contperc, is_kdde, is_filled_contour=TRUE)
{
    est.ord <- order(x$estimate)
    
    ## create contregion label
    if (is_kdde)
    {
        if (is_filled_contour)
        {
            x$contregion_low <- paste0(ifelse(x$estimate>=0, ">", "<"), round_signif(x$estimate, digits=digits)) 
            x$contregion_high <- paste0(ifelse(x$estimate_high>=0, ">", "<"), round_signif(x$estimate_high, digits=digits)) 
            x$contregion <- ifelse(x$estimate_high<0 & x$estimate<0, x$contregion_high, x$contregion_low)
            x$contregion[x$estimate_high>=0 & x$estimate<0] <- gsub("<",">", x$contregion)[x$estimate_high>=0 & x$estimate<0]
        }
        else 
            x$contregion <- paste0(ifelse(x$estimate>=0, ">", "<"), round_signif(x$estimate, digits=digits))
    }
    else 
        x$contregion <- paste0(">", round_signif(x$estimate, digits=digits))

    x$contregion <- gsub("-", "\u2013", gsub(">", "\u2265 ", gsub("<", "\u2264 ", x$contregion)))
    x$contlabel <- factor(x$contlabel, levels=unique(x$contlabel))
    x$contregion <- factor(x$contregion, levels=unique(x$contregion))
   
    ## create contline label
    if (is_kdde & is_filled_contour)
    {
        x$contline <- ifelse(sign(x$estimate_high)==1 & sign(x$estimate)==-1, NA, 1L)
        x$contline <- factor(x$contline)
    }

    ## create cont percentage label  
    if (add_contperc) 
    {   
        x$contperc <- x$contlabel
        x$contperc <- factor(x$contperc, levels=unique(x$contperc[est.ord]))
        
        levels(x$contperc) <- paste0(levels(x$contperc), "%")
    }
    else 
    {
        x$contperc <- NA  
        x$contperc <- factor(x$contperc)
    }
    
    return(x)
}

## mimics stat_contour_filled_ks output for x = tidy_k* output
add_contour_breaks <- function(x, breaks, digits, ...)
{
    if (missing(digits)) digits <- 4
    if (missing(breaks)) breaks <- contour_breaks(data=x, ...)$breaks
    if (min(x$estimate)>=0) breaks2 <- sort(unique(c(0, breaks, max(x$estimate))))
    else breaks2 <- sort(unique(c(breaks, max(x$estimate))))

    xnames <- colnames(x)
    x$.id <- 1:nrow(x)
    x.ord <- order(x$estimate)

    xlab <- x[x.ord,]
    xlab$estimate <- cut(x[x.ord,]$estimate, breaks2, labels=head(round_signif(breaks2, digits=digits), n=-1), digits=digits, include.lowest=TRUE)
    xlab$estimate_high <- cut(x[x.ord,]$estimate, breaks2, labels=tail(round_signif(breaks2,  digits=digits), n=-1), digits=digits, include.lowest=TRUE)
    xlab$estimate_high <- unfactor(xlab$estimate_high)
    xlab$contlabel <- xlab$estimate <- unfactor(xlab$estimate)
    xlab <- create_label(xlab, digits=digits, add_contperc=FALSE, is_kdde=TRUE, is_filled_contour=TRUE)
    xlab <- xlab[order(xlab$.id),]
    xlab$estimate <- ordered(xlab$estimate)
    xlab <- xlab[c(xnames, "contregion")]
    xlab <- dplyr::relocate(xlab, "contregion", .after="estimate")
   
    return(xlab)
}


## move group_vars in x to last column
## group_vars can be those in y  
move_group_vars <- function(x, y)
{
    if (missing(y)) y <- x
    if (dplyr::is_grouped_df(y))
    {
        gvy <- dplyr::group_vars(y)
        x <- dplyr::group_by(x, dplyr::across(dplyr::all_of(gvy)))
        x <- dplyr::relocate(x, dplyr::all_of(gvy), .after=dplyr::last_col())
    }

    return(x)
}

round_signif <- function(x, digits=4, format=FALSE) 
{
    x <- sapply(x, function(.) ifelse(.>=10^(digits-1), round(.), signif(.,digits))) 
    if (format) x <- format(x, nsmall=digits)

    return(x)
}

unfactor <- function(f) { as.numeric(levels(f))[f] }

## object class 
## type="tks" includes "kde", "kdde", "kda" ks object classes etc.
## type="ks" includes above + "kqde", "kqdde" quasi ks object classes
object_class <- function(x, type="tks")
{
    type <- match.arg(type, c("ks", "tks"))
    if (type=="ks") oc <- ifelse(inherits(x, "sf_ks"), class(x$tidy_ks$ks[[1]]), class(x$ks[[1]]))
    else if (type=="tks") oc <- ifelse(inherits(x, "sf_ks"), x$tidy_ks$tks[1], x$tks[1])

    return(oc)
}

## default aesthetics
## object = ggplot or tidy ks object
aes_ks <- function(object)
{
    if (inherits(object, "ggplot")) object <- object$data
    d <- dim_ks(object)
    if (d==1) aesd <- aes(y=!!sym("estimate"), weight=!!sym("ks"))
    else if (d==2)
    {
        if (all(c("u","v") %in% names(object))) aesd <- aes(u=!!sym("u"), v=!!sym("v"))
        else aesd <- aes(z=!!sym("estimate"), weight=!!sym("ks"))
    }

    return(aesd)
}

## default labels
## object = ggplot or tidy ks object
labs_ks <- function(object, x, y, ...)
{
    if (inherits(object, "ggplot")) object <- object$data
    d <- dim_ks(object)
    oct <- object_class(object, type="tks")

    if (inherits(object, "sf_ks"))
    {
        object <- object$tidy_ks[1,]
        if (missing(x)) x <- "Longitude"
        if (missing(y)) y <- "Latitude"
    }
    else
    {
        if (oct %in% "ksupp") { xnames <- names(untidy_ks(object)) }
        else if (oct %in% "kdr") 
        { 
            gv <- dplyr::group_vars(object)
            gv <- setdiff(gv, "segment")
            if (length(gv)>0) object <- dplyr::group_by(object, dplyr::across(dplyr::all_of(gv)))
            else object <- dplyr::ungroup(object)
            xnames <- untidy_ks(object, "names") 
        }
        else xnames <- untidy_ks(object, "names")
        while (is.list(xnames)) { xnames <- xnames[[1]] }

        if (is.null(xnames)) xnames <- c("x", "y")
        if (missing(x)) x <- xnames[1]
        if (missing(y)) y <- xnames[2]
    }

    if (d==1)
    {
        if (oct %in% c("kde", "kfs", "kde.boundary")) y <- "Density function"
        else if (oct %in% "kda") y <- "Weighted density"
        else if (oct %in% "kdde") y <- "Density derivative"
        else if (oct %in% "kcde") y <- "Distribution function"
        else if (oct %in% "kde.loctest") y <- expression("Density difference "*f1-f2)
    }
    if (oct %in% "kroc")
    {
        x <- expression("False positive rate"~~group("(", list(bar(specificity)), ")"))
        y <- "True positive rate (sensitivity)"
    }

    labsd <- ggplot2::labs(x=x, y=y, ...)

    return(labsd)
}

## default guides parameters
guides_ks <- function(object, kda_part=TRUE)
{
    if (inherits(object, "ggplot")) object <- object$data
    d <- dim_ks(object)
    if (inherits(object, "sf_ks")) object <- object$tidy_ks[1,]
    oct <- object_class(object, type="tks")

    if (d==1)
    {
        titlec <- NULL
        if (oct %in% "kda") { titlef <- "Classif label"; titlec <- "Classif label"; rev <- FALSE }
        else if (oct %in% "kde.loctest") { titlec <- "Signif difference" }
        else if (oct %in% "kfs") { titlec <- "" }
        if (!is.null(titlec)) gu <- ggplot2::guides(colour=ggplot2::guide_legend(title=titlec))
        else gu <- ggplot2::guides()
    }
    else if (d==2)
    {
        titlef <- NULL; titlec <- NULL; rev <- TRUE
        if (oct %in% c("kde", "kdcde", "kde.boundary", "kde.truncate", "kde.sp", "kde.balloon")) titlef <- "Density"
        else if (oct %in% "kcurv") titlef <- "Density curvature"
        else if (oct %in% "kda")
        {
            if (kda_part) { titlef <- "Classif label"; titlec <- "Density"; rev <- FALSE }
            else titlef <- "Density"
        }
        else if (oct %in% "kdde") { titlef <- ifelse(object_class(object, type="ks") %in% "kqdde", "Density", "Density derivative") }
        else if (oct %in% c("kcde", "kcopula")) titlef <- "Distribution"
        else if (oct %in% "kde.loctest") { titlef <- "Signif difference" }
        else if (oct %in% "kms") { titlef <- "Cluster"; rev <- FALSE }
        #else if (oct %in% c("kdr", "kfs")) { titlef <- ""; titlec <- ""; rev <- FALSE }
        else if (oct %in% "kfs") { titlef <- "Signif curv" }
        else if (oct %in% "kdr") { titlef <- "Density ridge" }
        else if (oct %in% "ksupp") titlef <- "Support\nconvex hull" 
        titlec <- titlef

        if ((titlef=="" & titlec=="")) gu <- ggplot2::guides(fill=ggplot2::guide_legend(title=NULL, reverse=rev), colour=ggplot2::guide_legend(title=NULL, reverse=rev))
        else if (!is.null(titlef) & !is.null(titlec)) gu <- ggplot2::guides(fill=ggplot2::guide_legend(title=titlef, reverse=rev), colour=ggplot2::guide_legend(title=titlec, reverse=rev))
        else if (!is.null(titlef) & is.null(titlec)) gu <- ggplot2::guides(fill=ggplot2::guide_legend(title=titlef, reverse=rev))
        else if (is.null(titlef) & !is.null(titlec)) gu <- ggplot2::guides(colour=ggplot2::guide_legend(title=titlec, reverse=rev))
        else gu <- ggplot2::guides()

        if (oct %in% "kdde") gu <- c(ggplot2::guides(linetype="none"), gu)
    }

    return(gu)
}

## default colour scale
scale_colour_ks <- function(object)
{
    if (inherits(object, "ggplot")) object <- object$data
    oc <- object_class(object, type="ks")
    oct <- object_class(object, "tks")
       
    if (inherits(object, "sf_ks")) { object <- object$tidy_ks[1, ]; sfobject <- TRUE }
    else { sfobject <- FALSE }

    if (oc %in% c("kcde", "kcopula")) sc <- ggplot2::scale_colour_viridis_d()
    else if (oc %in% "kda") sc <- colorspace::scale_colour_discrete_qualitative(palette="Dark2", na.translate=FALSE) 
    else if (oc %in% "kda") sc <- colorspace::scale_colour_discrete_sequential()
    else if (oc %in% c("kdde", "kqdde")) sc <- scale_colour_discrete_diverging_breaks()
    else if (oc %in% c("kde", "kqde")) 
    {
        if (oct %in% "kcurv") sc <- colorspace::scale_colour_discrete_sequential(h1=30, c1=360, c2=30) 
        else sc <- colorspace::scale_colour_discrete_sequential(palette="Heat")
    }
    else if (oc %in% "kde.loctest") sc <- colorspace::scale_colour_discrete_qualitative(palette="Dark2", rev=TRUE)
    else if (oc %in% "kdr") sc <- ggplot2::scale_colour_manual(values=6)     
    else if (oc %in% "kfs") sc <- ggplot2::scale_colour_manual(values=7) 
    else if (oc %in% "kms") sc <- colorspace::scale_colour_discrete_qualitative(palette="Set2")  
    else sc <- NULL

    return (sc)
}
scale_color_ks <- scale_colour_ks

scale_colour_discrete_diverging_breaks <- function(..., breaks, extend1)
{
    if (missing(breaks))
        sc <- colorspace::scale_colour_discrete_diverging(...)
    else 
    {
        ## unblanced number of negative + positive contour breaks
        if (inherits(breaks, c("matrix","data.frame","tbl_df"))) breaks <- breaks$breaks
        breaks <- sort(breaks)       
        breaks.rle <- rle(sign(breaks))
        breaks.len <- max(breaks.rle$lengths)
        col.ind.neg <- floor(seq(1, breaks.len, len=length(which(breaks<=0))))
        col.ind.pos <- ceiling(seq(1, breaks.len, len=length(which(breaks>0)))) + breaks.len 
        col.ind <- c(col.ind.neg, col.ind.pos)
        nmax <- max(col.ind)
    
        ## for sf_ks objects, need to add 1 extra colour here 
        if (extend1)
            sc <- colorspace::scale_colour_discrete_diverging(..., nmax=nmax+1, order=c(col.ind, nmax+1))  
        else
            sc <- colorspace::scale_colour_discrete_diverging(..., nmax=nmax, order=col.ind)
    }

    return(sc)
}

## default fill colour scale
scale_fill_ks <- function(object, transp_neutral)
{
    if (inherits(object, "ggplot")) object <- object$data
    if (inherits(object, "sf_ks")) { object <- object$tidy_ks[1,]; sfobject <- TRUE;  } else {sfobject <- FALSE }
    oc <- object_class(object, type="ks")
    oct <- object_class(object, "tks")

    if (oc %in% c("kcde", "kcopula")) sc <- ggplot2::scale_fill_viridis_d()
    else if (oc %in% "kda") sc <- colorspace::scale_fill_discrete_qualitative(palette="Dark2", na.translate=FALSE) 
    else if (oc %in% c("kdde", "kqdde"))
    {
        if (missing(transp_neutral)) transp_neutral <- oc %in% "kdde"
        sc <- scale_fill_discrete_diverging_breaks(transp_neutral=transp_neutral)
    }
    else if (oc %in% c("kde", "kqde")) 
    {
        if (oct %in% "kcurv") sc <- colorspace::scale_fill_discrete_sequential(h1=30, c1=360, c2=30)
        else sc <- colorspace::scale_fill_discrete_sequential(palette="Heat") 
    }
    else if (oc %in% "kde.loctest")
    {
        if (sfobject) sc <- colorspace::scale_fill_discrete_qualitative(palette="Dark2", rev=TRUE, order=1:2, na.translate=FALSE)
        else sc <- colorspace::scale_fill_discrete_qualitative(palette="Dark2", rev=TRUE, order=1:2, na.translate=FALSE) 
    }
    else if (oc %in% "kdr") sc <- ggplot2::scale_fill_manual(values=6) 
    else if (oc %in% "kfs") sc <- ggplot2::scale_fill_manual(values=7)
    else if (oc %in% "kms") sc <- colorspace::scale_fill_discrete_qualitative(palette="Set2")  
    else sc <- NULL

    return (sc)
}

## ggplot scale version of ks::col.diverging for unblanced # negative + positive breaks
scale_fill_discrete_diverging_breaks <- function(..., breaks, transp_neutral, extend1)
{
    if (missing(breaks))
    { 
        sc <- colorspace::scale_fill_discrete_diverging(...)
        if (transp_neutral) sc <- scale_transp_neutral(sc, insert_neutral=TRUE)
    }
    else 
    {   
        ## unblanced number of negative + positive contour breaks
        if (inherits(breaks, c("matrix","data.frame","tbl_df"))) breaks <- breaks$breaks
        breaks <- sort(breaks)
        breaks.rle <- rle(sign(breaks))
        breaks.len <- max(breaks.rle$lengths)
        col.ind.neg <- floor(seq(1, breaks.len, len=length(which(breaks<=0))))
        col.ind.pos <- ceiling(seq(1, breaks.len, len=length(which(breaks>0)))) + breaks.len 
        col.ind <- c(col.ind.neg, col.ind.pos)
        nmax <- max(col.ind)
        
        ## for sf_ks objects, need to add 1 extra colour here 
        if (extend1)
            sc <- colorspace::scale_fill_discrete_diverging(..., nmax=nmax+1, order=c(col.ind, nmax+1))  
        else
            sc <- colorspace::scale_fill_discrete_diverging(..., nmax=nmax, order=col.ind)
        if (transp_neutral) sc <- scale_transp_neutral(sc, insert_neutral=FALSE, ind_neutral=length(col.ind.neg)+extend1)    
    }
    
    return(sc)
}

## make individual colours in colour scale to be transparent 
## insert_neutral=TRUE then add transparent after "ind"th colour
## insert_neutral=FALSE then replace "ind"th colour with transparent 
scale_transp_neutral <- function(x, ind_neutral=NULL, insert_neutral=FALSE)
{  
    Scale <- ggplot2::ggproto("Scale", x, 
        palette = function(self, n)
        { 
            ## palette some times appears to have 1 more colour for NA class
            n <- n-1
            pal <- x$palette(n) 
            if (is.null(ind_neutral)) ind_neutral <- ceiling(stats::median(seq_len(n)))
            if (ind_neutral<1) ind_neutral <- 1
            if (ind_neutral>n) ind_neutral <- n
            if (insert_neutral) pal <- pal[c(1:ind_neutral, ind_neutral:n)] 
            pal[ind_neutral] <- "transparent"
            return(pal)
        }
    )

    return(Scale)
}

scale_fill_remove <- function() scale_remove("fill")
scale_fill_continuous_remove <- function() scale_remove("fill", "continuous")
scale_fill_discrete_remove <- function() scale_remove("fill", "discrete")

scale_color_remove <- scale_colour_remove <- function() scale_remove("colour")
scale_color_continuous_remove <- scale_colour_continuous_remove <- function() scale_remove("colour", "continuous")
scale_color_discrete_remove <- scale_colour_discrete_remove <- function() scale_remove("colour", "discrete")

## asymmetric scale
scale_asymmetric <- function(object, breaks, transp_neutral)
{
    if (missing(object))
    {
        if (missing(transp_neutral)) transp_neutral <- FALSE
        sfobject <- TRUE 
    }
    else
    {
        if (inherits(object, "ggplot")) object <- object$data
        if (inherits(object, "sf_ks")) { object <- object$tidy_ks[1,]; sfobject <- TRUE;  } else { sfobject <- FALSE }
        oct <- object_class(object, type="tks")

        if (oct %in% "kdde")
        {
            if (missing(transp_neutral)) transp_neutral <- object_class(object,type="ks") %in% "kdde"
        }
    }
    if (missing(breaks))
    {
        scc <- scale_colour_discrete_diverging_breaks()
        scf <- scale_fill_discrete_diverging_breaks(transp_neutral=transp_neutral)
    }
    else 
    {
        scc <- scale_colour_discrete_diverging_breaks(breaks=breaks, extend1=sfobject)
        scf <- scale_fill_discrete_diverging_breaks(breaks=breaks, transp_neutral=transp_neutral, extend1=TRUE)
    }
    sc <- list(scale_colour_discrete_remove(), scc, scale_fill_discrete_remove(), scf)
    
    return(sc)
}

## remove scales from ggplot object to avoid multiple scales warning message 
ggplot_add.scale_rem <- function(object, plot, object_name, ...) 
{
    if (length(plot$scales$scales)>0)
    {
        aes.ind <- which(sapply(plot$scales$scales, function(.) .$aesthetics %in% object[1] & inherits(., paste0("Scale", tools::toTitleCase(object[2])))))
        if (length(aes.ind)>0) plot$scales$scales[[aes.ind]] <- NULL
    }

    return(plot)
}

scale_remove <- function(scale1, scale2) 
{ 
    scale2 <- ifelse(missing(scale2), "",  sapply(scale2, function(.) match.arg(., c("continuous","discrete"))))
    rs <- structure(c(ggplot2::standardise_aes_names(scale1), scale2), class="scale_rem") 

    return(rs)
}

#####################################################################
## Compute sf versions of ks functions
#####################################################################

## default auxiliary function to convert ks object to sf object
## x = sf point object or tidy_ks object
st_ks <- function(x, fun_ks, crs, ...)
{
    ## x = tidy_ks object
    if (inherits(x, "tidy_ks"))
    {
        fhat <- x
    }
    ## x = sf object
    else
    {
        ## extract coordinates
        if (missing(crs)) crs <- sf::st_crs(x)
        xcoord <- dplyr::group_modify(.data=x, .f=~as.data.frame(sf::st_coordinates(.x)))
        
        ## compute tidy kernel estimate via tidy_k* function
        fhat <- do.call(fun_ks, args=list(data=xcoord, rename=FALSE, ...))
    }
    fhat.tidy <- fhat
    gv <- dplyr::group_vars(x)

    ## compute slice_head version of kernel estimate for geometry conversion
    fhat <- dplyr::slice_head(fhat)
    fhat <- dplyr::select(fhat, -dplyr::one_of(c("x","y","estimate")))
    fhat <- as_tidy_ks(fhat)

    ## convert ks grid to required geometry
    fhat.geom <- st_evalpoints(x=fhat)
    sfname <- attr(fhat.geom, "sf_column")
    if (length(gv)>0) fhat.geom <- dplyr::relocate(fhat.geom, dplyr::all_of(gv), .before=!!sfname)
    fhat.geom <- sf::st_sf(fhat.geom, crs=crs)
    
    ## convert contour polygons to multipolygon geometry
    cont.seq <- seq(5,95,by=5)
    cont.geom <- st_contourline(x=fhat, cont=cont.seq)
    if (length(gv)>0) cont.geom <- dplyr::relocate(cont.geom, dplyr::all_of(gv), .before=!!sfname)
    cont.geom <- dplyr::relocate(cont.geom, "estimate", .before=1)
    cont.geom <- st_create_label(cont.geom)
    cont.geom <- sf::st_sf(cont.geom, crs=crs)

    ## if original data is not tibble
    if (!inherits(x, "tbl_df")) 
    {
        fhat.geom <- sf::st_sf(as.data.frame(fhat.geom))
        cont.geom <- sf::st_sf(as.data.frame(cont.geom))
    }

	fhat.sf <- list(tidy_ks=fhat, grid=fhat.geom, sf=cont.geom)
	class(fhat.sf) <- "sf_ks"

    return(fhat.sf)
}

## sf version of ks::kde
## x = sf point geometry
st_kde <- function(x, ...)
{
    ## compute KDE
    fhat.sf <- st_ks(x=x, fun_ks="tidy_kde", ...)

    return(fhat.sf)
}

## sf version of ks::kdcde
## x = sf point geometry
st_kdcde <- function(x, ...)
{
    ## compute deconvolved KDE
    fhat.sf <- st_ks(x=x, fun_ks="tidy_kdcde", ...)

    return(fhat.sf)
}

## sf version of ks::kcde
## data = sf point geometry
st_kcde <- function(x, ...)
{
	## compute kernel CDE
	xcoord <- dplyr::group_modify(.data=x, .f=~as.data.frame(sf::st_coordinates(.x)))
	fhat <- tidy_kcde(xcoord, rename=FALSE, ...)
	fhat <- dplyr::slice_head(fhat)
    fhat <- dplyr::select(fhat, -dplyr::one_of(c("x","y","estimate")))

	## convert ks grid to required geometry
    gv <- dplyr::group_vars(x)
	fhat.geom <- st_evalpoints(x=fhat)
    sfname <- attr(fhat.geom, "sf_column")
    if (length(gv)>0) fhat.geom <- dplyr::relocate(fhat.geom, dplyr::all_of(gv), .before=!!sfname)
    crs <- sf::st_crs(x)
    fhat.geom <- sf::st_sf(fhat.geom, crs=crs)
    
    ## convert contour polygons to multipolygon geometry
    cont.seq <- seq(5, 95, by=5)
    cont.geom <- st_contourline(x=fhat, cont=cont.seq)
    if (length(gv)>0) cont.geom <- dplyr::relocate(cont.geom, dplyr::all_of(gv), .before=!!sfname)
    cont.geom <- dplyr::relocate(cont.geom, "estimate", .before=1)
    cont.geom <- st_create_label(cont.geom)
    cont.geom <- sf::st_sf(cont.geom, crs=crs)

    ## if original data is not tibble
    if (!inherits(x, "tbl_df")) 
    {
        fhat.geom <- sf::st_sf(as.data.frame(fhat.geom))
        cont.geom <- sf::st_sf(as.data.frame(cont.geom))
    }
    fhat.sf <- list(tidy_ks=fhat, grid=fhat.geom, sf=cont.geom)
    class(fhat.sf) <- "sf_ks"

    return(fhat.sf)
}

## sf version of ks::kde.boundary
## x = sf point geometry
st_kde_boundary <- function(x, ...)
{
    ## compute boundary corrected KDE
    fhat.sf <- st_ks(x=x, fun_ks="tidy_kde_boundary", ...)

    return(fhat.sf)
}

## sf version of ks::kde.balloon
## x = sf point geometry
st_kde_balloon <- function(x, ...)
{
    ## compute balloon variable KDE
    fhat.sf <- st_ks(x=x, fun_ks="tidy_kde_balloon", ...)

    return(fhat.sf)
}

## sf version of ks::kde.sp
## x = sf point geometry
st_kde_sp <- function(x, ...)
{
    ## compute sample point variable KDE
    fhat.sf <- st_ks(x=x, fun_ks="tidy_kde_sp", ...)

    return(fhat.sf)
}

## sf version of ks::kde.truncate
## x = sf point geometry
## boundary = sf point/polygon geometry
st_kde_truncate <- function(x, boundary, ...)
{
	## compute boundary kernel density estimate
    xcoord <- dplyr::group_modify(.data=x, .f=~as.data.frame(sf::st_coordinates(.x)))
    fhat <- .tidy_kde_truncate(data=xcoord, boundary=sf::st_coordinates(boundary), rename=FALSE, ...)
    #fhat.tidy <- fhat
    fhat <- dplyr::slice_head(fhat)
    fhat <- dplyr::select(fhat, -dplyr::one_of(c("x","y","estimate")))

    gv <- dplyr::group_vars(fhat)
    fhat.geom <- st_evalpoints(x=fhat)
    sfname <- attr(fhat.geom, "sf_column")
    if (length(gv)>0) fhat.geom <- dplyr::relocate(fhat.geom, dplyr::all_of(gv), .before=!!sfname)
    crs <- sf::st_crs(x)
    fhat.geom <- sf::st_sf(fhat.geom, crs=crs)
    
    ## convert contour polygons to multipolygon geometry
    cont.seq <- seq(5, 95, by=5)
    cont.geom <- st_contourline(x=fhat, cont=cont.seq)
    if (length(gv)>0) cont.geom <- dplyr::relocate(cont.geom, dplyr::all_of(gv), .before=!!sfname)
    cont.geom <- dplyr::relocate(cont.geom, "estimate", .before=1)
    cont.geom <- st_create_label(cont.geom)
    cont.geom <- sf::st_sf(cont.geom, crs=crs)

    ## if original data is not tibble
    if (!inherits(x, "tbl_df"))
    {
        fhat.geom <- sf::st_sf(as.data.frame(fhat.geom))
        cont.geom <- sf::st_sf(as.data.frame(cont.geom))
    }
    
	fhat.sf <- list(tidy_ks=fhat, grid=fhat.geom, sf=cont.geom)
	class(fhat.sf) <- "sf_ks"

    return(fhat.sf)
}

## sf version of ks::kda
## x = sf point geometry
st_kda <- function(x, ...)
{
    ## compute kernel discriminant analysis
	xcoord <- dplyr::group_modify(.data=x, .f=~as.data.frame(sf::st_coordinates(.x)))
	fhat <- .tidy_kda(xcoord, rename=FALSE, ...)
	fhat.tidy <- fhat
    fhat <- dplyr::slice_head(fhat)
    fhat <- dplyr::select(fhat, -dplyr::one_of(c("x","y","prior_prob","estimate")))
    fhat <- dplyr::mutate(fhat, label="Classif")
    gv <- dplyr::group_vars(fhat)

    ## convert ks partition grid to required geometry
    fhat.geom <- st_evalpoints(x=fhat[1,])
    fhat.geom <- dplyr::mutate(fhat.geom, label=.data$estimate, .after="estimate")
    #if (length(gv)==1)names(fhat.geom)[names(fhat.geom) %in% "label"] <- gv
    levels(fhat.geom$estimate) <- 1:length(levels(fhat.geom$estimate))
    fhat.geom$estimate <- as.integer(levels(fhat.geom$estimate))[fhat.geom$estimate]
    gv <- dplyr::group_vars(fhat.geom)
    fhat.geom <- dplyr::select(dplyr::ungroup(fhat.geom), -dplyr::all_of(gv))
    crs <- sf::st_crs(x)
    fhat.geom <- sf::st_sf(fhat.geom, crs=crs)
    sfname <- attr(fhat.geom, "sf_column")

    ## convert contours to multipolygon geometry
	fhat <- dplyr::bind_cols(fhat, rn=1:nrow(fhat))
    cont.seq <- seq(5, 95, by=5)
	cont.geom <- dplyr::group_modify(.data=fhat, .f=~dplyr::tibble(geometry=list(st_contourline_kda(x=.x, cont=cont.seq))))
	cont.geom <- dplyr::group_modify(.data=cont.geom, .f=~data.frame(.x$geometry))
	fhat <- dplyr::select(fhat, -dplyr::one_of("rn"))
	if (length(gv)>0) cont.geom <- dplyr::relocate(cont.geom, dplyr::all_of(gv), .before=!!sfname)
    cont.geom <- dplyr::relocate(cont.geom, "estimate", .before=1)
    cont.geom <- st_create_label(cont.geom)
	cont.geom <- sf::st_sf(cont.geom, crs=crs)

	fhat.sf <- list(tidy_ks=fhat, grid=fhat.geom, sf=cont.geom)
	class(fhat.sf) <- "sf_ks"

    return(fhat.sf)
}

## sf version of ks::kdde
## x = sf point geometry
st_kdde <- function(x, deriv_order=1, ...)
{
	## compute kernel density derivatives estimate
	xcoord <- dplyr::group_modify(.data=x, .f=~as.data.frame(sf::st_coordinates(.x)))
	fhat <- .tidy_kdde(xcoord, deriv_order=deriv_order, rename=FALSE, ...)
    fhat <- dplyr::slice_head(fhat)
    gv <- dplyr::group_vars(fhat)
    gv2 <- c("deriv_ind", gv)
    fhat <- dplyr::arrange(fhat, .data$deriv_ind)
    fhat <- dplyr::select(fhat, -dplyr::one_of(c("x","y","estimate")))
    fhat <- as_tidy_ks(fhat)

	## convert ks grid to required geometry
	crs <- sf::st_crs(x) 
    fhatg <- dplyr::group_by(fhat, dplyr::across({{gv2}}))
    
    fhat.geom <- st_evalpoints(x=fhatg)
    sfname <- attr(fhat.geom, "sf_column")   
    fhat.geom <- dplyr::mutate(fhat.geom, estimate=get(paste0("estimate.", dplyr::first(.data$deriv_ind))), .before=!!sfname)
    fhat.geom <- dplyr::select(fhat.geom, -dplyr::starts_with("estimate."))
    fhat.geom <- dplyr::group_by(fhat.geom, dplyr::across(dplyr::all_of(gv)))
    fhat.geom <- dplyr::relocate(fhat.geom, dplyr::all_of(gv), .before=!!sfname)
    fhat.geom <- dplyr::relocate(fhat.geom, "deriv_ind", .after="estimate")
    fhat.geom <- sf::st_sf(fhat.geom, crs=crs)
  
    ## convert contours to multipolygon geometry
    cont.seq <- seq(5, 95, by=5)
	cont.geom <- dplyr::group_modify(.data=fhat, .f=~dplyr::tibble(geometry=list(st_contourline_kdde(x=untidy_ks(.x), which_deriv_ind=.x$deriv_ind, cont=cont.seq))))
	cont.geom <- dplyr::group_modify(.data=cont.geom, .f=~data.frame(.x$geometry))
	cont.geom <- dplyr::left_join(cont.geom, dplyr::select(fhat, dplyr::all_of(gv2)), by=gv)
    cont.geom <- dplyr::relocate(cont.geom, dplyr::all_of(gv2), .before=!!sfname) 
    cont.geom <- dplyr::relocate(cont.geom, "deriv_ind", .after="estimate")
    cont.geom <- dplyr::relocate(cont.geom, "estimate", .before=1) 
    cont.geom <- dplyr::arrange(cont.geom, .data$deriv_ind)
    cont.geom <- st_create_label(cont.geom, is_kdde=TRUE)
    cont.geom <- sf::st_sf(cont.geom, crs=sf::st_crs(x))

    fhat <- dplyr::select(fhat, -dplyr::one_of(c("deriv_order")))
    fhat <- as_tidy_ks(fhat)
	fhat.sf <- list(tidy_ks=fhat, grid=fhat.geom, sf=cont.geom)
	class(fhat.sf) <- "sf_ks"

	return(fhat.sf)
}

## sf version of ks::kde_local_test
## x = sf point geometry
st_kde_local_test <- function(x1, x2, labels, ...)
{
    ## compute kernel local test
	xcoord1 <- dplyr::group_modify(.data=x1, .f=~as.data.frame(sf::st_coordinates(.x)))
	xcoord2 <- dplyr::group_modify(.data=x2, .f=~as.data.frame(sf::st_coordinates(.x)))

    fhat <- .tidy_kde_local_test(data1=xcoord1, data2=xcoord2, rename=FALSE, ...)
    fhat <- dplyr::slice_head(fhat)
    fhat <- dplyr::select(fhat, -dplyr::one_of(c("x","y","estimate")))
    if (missing(labels)) 
    {
        mc <- as.character(match.call())[2:3]
        labels <- gsub("f1", mc[1], levels(fhat$label))
        labels <- gsub("f2", mc[2], labels)
    }
    else labels <- c(paste0(labels[1],"<",labels[2]), NA, paste0(labels[1],">",labels[2]))
    fhat <- dplyr::mutate(fhat, label="Density diff")
    gv <- dplyr::group_vars(fhat)

    ## convert ks grid to required geometry
    fhat.geom <- st_evalpoints(x=fhat)
    sfname <- attr(fhat.geom, "sf_column")
    if (length(gv)>0) fhat.geom <- dplyr::relocate(fhat.geom, dplyr::all_of(gv), .before=!!sfname)
    crs <- sf::st_crs(x1)
    fhat.geom <- sf::st_sf(fhat.geom, crs=crs)
    
	## convert contour polygons to multipolygon geometry
	abs.cont <- c(-0.5, 0.5); names(abs.cont) <- c("50%", "50%")
    for (i in 1:nrow(fhat))
    {
        fhat$ks[[i]]$eval.points <- fhat$ks[[i]]$fhat1$eval.points
        fhat$ks[[i]]$estimate <- fhat$ks[[i]]$fhat.diff.pos$estimate - fhat$ks[[i]]$fhat.diff.neg$estimate
    }
    cont.geom <- dplyr::bind_rows(st_contourline(x=fhat, abs.cont=abs.cont[1]), st_contourline(x=fhat, abs.cont=abs.cont[2]))
    cont.geom <- dplyr::mutate(cont.geom, contlabel=factor(ifelse(.data$estimate<0, paste0("-", 100-as.numeric(as.character(.data$contlabel))), as.character(.data$contlabel))))
    cont.geom <- dplyr::mutate(cont.geom, label=.data$contlabel, .after="estimate")
    levels(cont.geom$label) <- as.vector(na.omit(labels))    
    cont.geom <- dplyr::arrange(cont.geom, sign(.data$estimate), abs(.data$estimate))
    if (length(gv)>0) 
    {   
        cont.geom <- dplyr::relocate(cont.geom, dplyr::all_of(gv), .before=!!sfname)
        cont.geom <- dplyr::arrange(cont.geom, dplyr::across(dplyr::all_of(gv)))
    }
    cont.geom <- dplyr::relocate(cont.geom, "estimate", .before=1)
    cont.geom <- st_create_label(cont.geom, is_kdde=FALSE) 
    cont.geom <- sf::st_sf(cont.geom, crs=crs)

    ## if original data is not tibble
    if (!inherits(x1, "tbl_df"))
    {
        fhat.geom <- sf::st_sf(as.data.frame(fhat.geom))
        cont.geom <- sf::st_sf(as.data.frame(cont.geom))
    }

	fhat.sf <- list(tidy_ks=fhat, grid=fhat.geom, sf=cont.geom)
	class(fhat.sf) <- "sf_ks"

	return(fhat.sf)
}

## sf version of ks::kfs
## x = sf point geometry
st_kfs <- function(x, ...)
{
    ## compute kernel feature significance
	xcoord <- dplyr::group_modify(.data=x, .f=~as.data.frame(sf::st_coordinates(.x)))
    labels <- as.character(match.call())[2]
	fhat <- .tidy_kfs(xcoord, rename=FALSE, labels=labels, ...)
	fhat.tidy <- fhat
    fhat <- dplyr::slice_head(fhat)
    fhat <- dplyr::select(fhat, -dplyr::one_of(c("x","y","estimate")))
    fhat <- dplyr::mutate(fhat, label="Signif curv", .after="tks")

    ## convert ks grid to required geometry
    gv <- dplyr::group_vars(x)
    fhat.geom <- st_evalpoints(x=dplyr::select(fhat, -dplyr::one_of("tks", "label")))
    sfname <- attr(fhat.geom, "sf_column")
    if (length(gv)>0) fhat.geom <- dplyr::relocate(fhat.geom, dplyr::all_of(gv), .before=!!sfname)
    crs <- sf::st_crs(x)
    fhat.geom <- sf::st_sf(fhat.geom, crs=crs)
   
	## convert contour polygons to multipolygon geometry
	abs.cont <- 0.5; names(abs.cont) <- "50%"
    cont.geom <- st_contourline(x=fhat, abs.cont=abs.cont)

    cont.geom <- dplyr::relocate(cont.geom, "estimate", "contlabel", .before=1)
    if (length(gv)>0) cont.geom <- dplyr::relocate(cont.geom, dplyr::all_of(gv), .before=!!sfname)
    cont.geom <- dplyr::mutate(cont.geom, label=!!labels, .after="estimate")
    cont.geom <- dplyr::relocate(cont.geom, "estimate", .before=1)
    cont.geom <- st_create_label(cont.geom)
    cont.geom$contregion <- factor(cont.geom$label)
    cont.geom <- sf::st_sf(cont.geom, crs=sf::st_crs(x))

    ## if original data is not tibble
    if (!inherits(x, "tbl_df")) 
    {
        fhat.geom <- sf::st_sf(as.data.frame(fhat.geom))
        cont.geom <- sf::st_sf(as.data.frame(cont.geom))
    }
    
	fhat.sf <- list(tidy_ks=fhat, grid=fhat.geom, sf=cont.geom)
	class(fhat.sf) <- "sf_ks"

	return(fhat.sf)
}

## sf version of ks::kdr
## x = sf point geometry
st_kdr <- function(x, dTolerance=100, ...)
{
    ## compute kernel density ridge
	xcoord <- dplyr::group_modify(.data=x, .f=~as.data.frame(sf::st_coordinates(.x)))
    labels <- as.character(match.call())[2]
	fhat <- .tidy_kdr(xcoord, rename=FALSE, dTolerance=dTolerance, labels=labels, ...)
    if (inherits(fhat, "sf")) fhat <- sf::st_drop_geometry(fhat)
    
    ## convert density ridge to linestring 
    fhat.sf <- sf::st_as_sf(fhat, coords=c("x","y"), crs=sf::st_crs(x))
    fhat.sf <- dplyr::summarise(fhat.sf, do_union=FALSE, label=head(.data$label,n=1), .groups="keep")
    fhat.sf <- sf::st_cast(fhat.sf, to="LINESTRING")
    fhat.sf <- dplyr::mutate(fhat.sf, contlabel=50L, estimate=0, .before="label")
    fhat.sf <- st_create_label(fhat.sf)
    fhat.sf$contregion <- factor(fhat.sf$label)
    fhat.sf <- dplyr::relocate(fhat.sf, "segment", .after="label")
    fhat.sf <- sf::st_as_sf(fhat.sf)

    gv <- dplyr::group_vars(x)
    if (length(gv)>0) fhat <- dplyr::group_by(fhat, dplyr::across(dplyr::all_of(gv))) else fhat <- dplyr::ungroup(fhat)
    fhat <- dplyr::slice_head(fhat)
    fhat <- dplyr::select(fhat, -dplyr::one_of(c("x","y","segment", "estimate")))
    
    fhat.sf <- list(tidy_ks=fhat, grid=NULL, sf=fhat.sf)
    class(fhat.sf) <- "sf_ks"

    return(fhat.sf) 
}

## sf version of ks::kms
## x = sf point geometry
st_kms <- function(x, ...)
{
    ## compute kernel mean shift
	xcoord <- dplyr::group_modify(.data=x, .f=~as.data.frame(sf::st_coordinates(.x)))
	fhat <- .tidy_kms(xcoord, rename=FALSE, ...)
    fhat <- dplyr::slice_head(fhat)
    fhat <- dplyr::select(fhat, -dplyr::one_of(c("x","y","estimate")))
    gv <- dplyr::group_vars(x)

    ## compute cluster labels
    fhat1 <- dplyr::group_modify(.data=fhat, .f=~dplyr::as_tibble(untidy_ks(.x, "x")))
    fhat2 <- dplyr::group_modify(.data=fhat, .f=~dplyr::rename(dplyr::as_tibble(untidy_ks(.x, "label")), label=dplyr::all_of("value")))
    fhat1$label <- as.integer(fhat2$label)

    fhat.sf <- dplyr::mutate(fhat1, contlabel=50, contlabel=factor(.data$contlabel), estimate=.data$label, .before="label")
    fhat.sf <- dplyr::mutate(fhat.sf, estimate=factor(.data$estimate), label=factor(.data$label))
    fhat.sf <- sf::st_as_sf(fhat.sf, coords=c("X","Y"), crs=sf::st_crs(x))
    fhat.sf <- dplyr::relocate(fhat.sf, "estimate", .before=1)
    fhat.sf <- st_create_label(fhat.sf)
    sfname <- attr(fhat.sf, "sf_column")
    if (length(gv)>0) 
    {
        fhat.sf <- dplyr::relocate(fhat.sf, dplyr::all_of(gv), .before=!!sfname)
        fhat.sf <- sf::st_sf(fhat.sf)
    }

    ## if original data is not tibble
    if (!inherits(x, "tbl_df")) fhat.sf <- sf::st_sf(as.data.frame(fhat.sf))

    fhat.sf <- list(tidy_ks=fhat, grid=NULL, sf=fhat.sf)
    class(fhat.sf) <- "sf_ks"

    return(fhat.sf)
}

## sf version of kernel support estimate
## x= sf_ks object (output from st_kde)
st_ksupp <- function(x, cont=95, convex_hull=TRUE, ...)
{
    ## compute kernel support estimate
    fhat <- x$tidy_ks 
    fhat.tidy <- .tidy_ksupp(fhat, cont=cont, rename=FALSE, convex_hull=convex_hull, ...)
    fhat.tidy <- dplyr::slice_head(fhat.tidy)
    fhat.tidy <- dplyr::select(fhat.tidy, -dplyr::one_of(c("x","y","estimate")))
    
    ## convert ks grid to required geometry
    gv <- dplyr::group_vars(x$tidy_ks)
    fhat.geom <- st_evalpoints(x=fhat)
    sfname <- attr(fhat.geom, "sf_column")
    if (length(gv)>0) fhat.geom <- dplyr::relocate(fhat.geom, dplyr::all_of(gv), .before=!!sfname)
    crs <- sf::st_crs(x)
    fhat.geom <- sf::st_sf(fhat.geom, crs=sf::st_crs(x$sf))

    ## convert kernel support to polygon
    gv <- dplyr::group_vars(x$tidy_ks)
    fhat.sf <- dplyr::group_modify(.data=fhat.tidy, .f=~dplyr::tibble(geometry=list(sf::st_linestring(as.matrix(untidy_ks(.x))))))
    fhat.sf <- sf::st_sf(fhat.sf)
    fhat.sf <- sf::st_cast(fhat.sf, to="POLYGON")

    sfname <- attr(fhat.sf, "sf_column")
    contlev <- min(contourLevels(fhat,cont=cont)$breaks)
    fhat.sf <- dplyr::mutate(fhat.sf, estimate=contlev, contlabel=factor(cont), .before=!!sfname)
    fhat.sf <- dplyr::relocate(fhat.sf, "contlabel", "estimate", .before=1)
    fhat.sf <- dplyr::mutate(fhat.sf, label="Support", .before=!!sfname)
    fhat.sf <- dplyr::relocate(fhat.sf, "estimate", .before=1)
    if (length(gv)>0) fhat.sf <- dplyr::relocate(fhat.sf, dplyr::all_of(gv), .before=!!sfname)
    fhat.sf <- st_create_label(fhat.sf)
    fhat.sf$contregion <- factor(round_signif(fhat.sf$estimate))
    levels(fhat.sf$contregion) <- paste("\u2265", levels(fhat.sf$contregion))
    fhat.sf <- sf::st_sf(fhat.sf)

    ## if original data is not tibble
    if (!dplyr::is_grouped_df(x$sf)) fhat.sf <- data.frame(fhat.sf)
    fhat.sf <- sf::st_sf(fhat.sf, crs=sf::st_crs(x$sf))

    fhat.sf <- list(tidy_ks=fhat.tidy, grid=fhat.geom, sf=fhat.sf)
	class(fhat.sf) <- "sf_ks"

    return(fhat.sf)
}

## sf version of ks::kcurv
## x = sf_ks object (output from st_kdde)
st_kcurv <- function(x, ...)
{
    ## compute kernel curvature estimate
    fhat <- dplyr::slice_head(x$tidy_ks)
    fhat <- tidy_kcurv(fhat, rename=FALSE)
    fhat <- dplyr::slice_head(fhat)
    fhat <- dplyr::select(fhat, -dplyr::one_of(c("x","y","estimate")))
    gv <- setdiff(dplyr::group_vars(x$tidy_ks), "deriv_group")

    ## convert ks grid to required geometry
    fhat.geom <- st_evalpoints(x=fhat)
    sfname <- attr(fhat.geom, "sf_column")
    if (length(gv)>0) fhat.geom <- dplyr::relocate(fhat.geom, dplyr::all_of(gv), .before=!!sfname)
    crs <- sf::st_crs(x$sf)
    fhat.geom <- sf::st_sf(fhat.geom, crs=crs)

	## convert contours to multipolygon geometry
	cont.geom <- dplyr::group_modify(.data=fhat, .f=~dplyr::tibble(geometry=list(.st_contourline(x=untidy_ks(.x)))))
	cont.geom <- dplyr::group_modify(.data=cont.geom, .f=~data.frame(.x$geometry))
    cont.geom <- dplyr::filter(cont.geom, .data$contlabel %in% seq(5,95,by=5))
    #cont.geom <- dplyr::bind_cols(label=fhat$label[1], cont.geom)
    #cont.geom <- dplyr::relocate(cont.geom, "contlabel", "estimate", .before=1)
    if (length(gv)>0) cont.geom <- dplyr::relocate(cont.geom, dplyr::all_of(gv), .before=!!sfname)
    cont.geom <- dplyr::relocate(cont.geom, "estimate", .before=1)
    cont.geom <- st_create_label(cont.geom)
    cont.geom <- sf::st_sf(cont.geom, crs=crs)

    ## if original data is not tibble
    if (length(gv)==0) 
    {
        fhat.geom <- sf::st_sf(as.data.frame(fhat.geom))
        cont.geom <- sf::st_sf(as.data.frame(cont.geom))
    }
	fhat.sf <- list(tidy_ks=fhat, grid=fhat.geom, sf=cont.geom)
	class(fhat.sf) <- "sf_ks"

    return(fhat.sf)
}

## spatial version of ks::kroc
## x1, x2 = sf point geometry
st_kroc <- function(x1, x2, ...)
{
    ## compute kernel ROC 
    xcoord1 <- dplyr::group_modify(.data=x1, .f=~as.data.frame(sf::st_coordinates(.x)))
    xcoord2 <- dplyr::group_modify(.data=x2, .f=~as.data.frame(sf::st_coordinates(.x)))
    fhat <- tidy_kroc(data1=xcoord1, data2=xcoord2, ...)

    return(fhat)
}

## spatial version of ks::kquiver
## x = output from st_kdde(, deriv_order=1)
st_kquiver <- function(x, thin=5, transf=1/4, neg.grad=FALSE, scale=1)
{ 
    ## compute kernel quiver estimate
    jv <- dplyr::group_vars(x$tidy_ks)
    t1 <- dplyr::group_modify(x$tidy_ks, .f=~data.frame(expand.grid(untidy_ks(.x)$eval.points), estimate1=as.vector(untidy_ks(.x)$estimate[[1]]), estimate2=as.vector(untidy_ks(.x)$estimate[[2]])))
    t1 <- dplyr::left_join(t1, x$tidy_ks, by=jv)
    t1 <-dplyr::mutate(t1, rn=dplyr::row_number(), ks=ifelse(.data$rn==1, .data$ks, list(2L)))
    t1 <- dplyr:: mutate(t1, deriv_order=1L, deriv_ind=ifelse(.data$deriv_group=="deriv (1,0)", 1, 2), estimate=ifelse(.data$deriv_ind==1, .data$estimate1, .data$estimate2))
   
    if (any(names(t1) %in% "X")) t1 <- dplyr::rename(t1, x=dplyr::all_of("X"))
    if (any(names(t1) %in% "Y")) t1 <- dplyr::rename(t1, y=dplyr::all_of("Y"))   
    if (any(names(t1) %in% "Var1")) t1 <- dplyr::rename(t1, x=dplyr::all_of("Var1"))
    if (any(names(t1) %in% "Var2")) t1 <- dplyr::rename(t1, y=dplyr::all_of("Var2"))   
    t1 <- dplyr::select(t1, dplyr::all_of(c("x", "y", "estimate", "deriv_order", "deriv_ind", "ks", "tks", "label", jv)))
    t2 <- tidy_kquiver(t1, thin=thin, transf=transf, neg.grad=neg.grad)

    ## rescale quiver arrows 
    dx <- max(diff(t2$x))/2; dy <- max(diff(t2$y))/2
    t3 <- dplyr::filter(t2, abs(.data$u)>0 & abs(.data$v)>0)   
    t3 <- dplyr::mutate(t3, u=(.data$u-min(.data$u))/(max(.data$u)-min(.data$u))*2*dx - dx, v=(.data$v-min(.data$v))/(max(.data$v)-min(.data$v))*2*dy - dy, xend=.data$x+.data$u, yend=.data$y+.data$v)   
    t4 <- dplyr::select(dplyr::mutate(sf::st_as_sf(t3, coords=c("x","y")), .group=dplyr::row_number()), -dplyr::one_of(c("xend", "yend")))          
    t5 <- dplyr::select(dplyr::mutate(sf::st_as_sf(t3, coords=c("xend","yend")), .group=dplyr::row_number()), -dplyr::one_of(c("x", "y"))) 
    t6 <- dplyr::group_by(dplyr::bind_rows(t4,t5), .data$.group)
    t6 <- dplyr::summarise(t6, do_union=FALSE)
    t6 <- sf::st_cast(t6, to="LINESTRING")        
    t6 <- dplyr::left_join(t6, sf::st_drop_geometry(t4), by=".group") 
    t6 <- sf::st_set_crs(t6, sf::st_crs(x$sf))
    t6 <- dplyr::bind_cols(t6, x=t3$x, y=t3$y, xend=t3$xend, yend=t3$yend)
    t6$len <- unit(scale*sqrt(t3$u^2+t3$v^2)/10^ceiling(log10(max(sqrt(t3$u^2+t3$v^2)))), "npc")
    t6 <- dplyr::rename(t6, lon=dplyr::all_of("x"), lat=dplyr::all_of("y"), lon_end=dplyr::all_of("xend"), lat_end=dplyr::all_of("yend"))
    sfname <- attr(t6, "sf_column")
    if (length(jv)>=1) jv <- setdiff(jv, "deriv_group")
    if (length(jv)>=1) t6 <- dplyr::select(t6, dplyr::one_of(c("lon","lat","lon_end","lat_end","len","u","v",jv, sfname))) 
    else t6 <- dplyr::select(t6, dplyr::one_of(c("lon","lat","lon_end","lat_end","len","u","v",sfname))) 
    
    fhat <- dplyr::slice_head(t2, n=1)
    if (length(jv)==0) fhat <- dplyr::ungroup(fhat) 
    else 
    { 
        t6 <- dplyr::group_by(t6, dplyr::across(dplyr::all_of(jv)))
        t6 <- dplyr::arrange(t6, dplyr::across(dplyr::all_of(jv)), .data$lat, .data$lon) 
    } 
    fhat <- dplyr::select(fhat, -dplyr::one_of(c("x","y","u","v","estimate")))
    fhat.sf <- list(tidy_ks=fhat, grid=NULL, sf=t6)
    class(fhat.sf) <- "sf_ks"

    return(fhat.sf)
}

## spatial version of ks::as.kde
## x = complete sf polygon grid
st_as_kde <- function(x, attrib=1, density, cont, ...)
{
    if (is.numeric(attrib)) attrib <- names(x)[attrib]
    x2 <- tidy_intergrid(x, attrib=attrib) 
    crs <- sf::st_crs(x)

    if (missing(density)) 
    {
        if (all(x2$estimate>=0)) density <- TRUE
        else if (all(x2$estimate<0)) density <- FALSE
        else density <- (abs(mean(x2$estimate[x2$estimate<0]))/mean(x2$estimate[x2$estimate>=0]) < 1e-6) 
    }

    ## convert to tidy KDE
    ## fhat is based on ks::kde so is evaluated at vertices of M1 x M2 estimation grid
    fhat <- .tidy_as_kde(data=x2, is_tidy=TRUE, density=density, rename=FALSE, ...)
    fhat <- dplyr::slice_head(fhat)
    fhat <- dplyr::select(fhat, -dplyr::one_of(c("x","y","estimate")))
    if (!density)
    {
        fhat <- dplyr::select(fhat, -dplyr::one_of(c("deriv_order")))
        fhat <- as_tidy_ks(fhat)
    }
    
    ## convert ks grid to required geometry
    ## fhat.geom is evaluated at (M1-1) x (M2-1) centroids of M1 x M2 estimation grid
    ## i.e. fhat.geom is (M1-1) x (M2-1) rectangle polygon grid
    fhat.geom <- dplyr::transmute(x, estimate=.data[[attrib]])
    if (!density)
    {
        fhat.geom <- dplyr::mutate(fhat.geom, deriv_order=1L, deriv_ind=1L, deriv_group="(1)", .after="estimate")
        fhat.geom <- dplyr::group_by(fhat.geom, .data$deriv_group, .add=TRUE)
    }
    
    ## convert contour polygons to multipolygon geometry
    if (missing(cont)) cont <- seq(5, 95, by=5)
    if (density) 
    {
        cont.geom <- st_contourline(x=fhat, cont=cont)
        cont.geom <- sf::st_sf(cont.geom, crs=crs)
        cont.geom <- suppressWarnings(sf::st_crop(cont.geom, sf::st_bbox(x)))
        cont.geom <- st_create_label(cont.geom)
    }
    else 
    {
        fhat <- dplyr::group_by(fhat, .data$deriv_group, .add=TRUE)
        gv <- dplyr::group_vars(fhat)
        gv2 <- c("deriv_ind", gv)
        crs <- sf::st_crs(x) 
        
        ## convert contours to multipolygon geometry
        cont.geom <- dplyr::group_modify(.data=fhat, .f=~dplyr::tibble(geometry=list(st_contourline_kdde(x=untidy_ks(.x), which_deriv_ind=.x$deriv_ind, cont=cont))))
        cont.geom <- dplyr::group_modify(.data=cont.geom, .f=~data.frame(.x$geometry))
        cont.geom <- dplyr::left_join(cont.geom, dplyr::select(fhat, dplyr::all_of(gv2)), by=gv)
        cont.geom <- dplyr::arrange(cont.geom, .data$deriv_ind)
        cont.geom <- sf::st_sf(cont.geom, crs=crs)
        sfname <- attr(cont.geom, "sf_column")
        cont.geom <- dplyr::relocate(cont.geom, "deriv_ind", .after="estimate") 
        cont.geom <- dplyr::relocate(cont.geom, dplyr::all_of(gv2), .before=!!sfname) 
        cont.geom <- sf::st_sf(cont.geom, crs=crs)
        cont.geom <- suppressWarnings(sf::st_crop(cont.geom, sf::st_bbox(x)))
        cont.geom <- st_create_label(cont.geom)

        #fhat.sf <- list(tidy_ks=fhat, grid=fhat.geom, sf=cont.geom)
        #class(fhat.sf) <- "sf_ks"
    }

    ## if original data is not tibble
    if (!inherits(x, "tbl_df")) 
    {
        fhat.geom <- sf::st_sf(as.data.frame(fhat.geom))
        cont.geom <- sf::st_sf(as.data.frame(cont.geom))
    }
    fhat.sf <- list(tidy_ks=fhat, grid=fhat.geom, sf=cont.geom)
    class(fhat.sf) <- "sf_ks"

    return(fhat.sf)
}


## extract or compute new contours
## x = ks_sf object
st_get_contour <- function(x, cont=c(25,50,75), breaks, disjoint=TRUE, digits)
{
    ## some defaults
    crop <- TRUE      ## crop to bounding box of x$grid
    cont100 <- TRUE   ## compute 100% contour = bounding box \ union {all other contour regions}
    oct <- object_class(x, type="tks")

    if (missing(digits)) digits <- ifelse(oct %in% c("kcde", "kcopula", "kde.loctest", "kfs", "kdr", "kms"), 2, 4)
    fhat <- dplyr::slice_head(x$tidy_ks)
    missing_breaks <- missing(breaks)

    if (oct %in% c("kdde", "kde.loctest")) cont <- unique(c(cont, -cont))
    if (oct %in% c("kms", "kfs"))
    {
        xc <- dplyr::filter(x$sf, .data$contlabel %in% cont)
    }
    else if (oct %in% c("kdr", "kms", "kquiver", "ksupp", "kde.loctest"))
    {
        xc <- x$sf
    }
    else if (oct %in% "kdde")
    { 
        ## absolute contour levels 
        if (!missing(breaks))
        {
            xc <- dplyr::group_modify(.data=fhat, .f=~dplyr::tibble(geometry=list(st_contourline_kdde(x=untidy_ks(.x), which_deriv_ind=.x$deriv_ind, abs.cont=breaks, digits=digits)), deriv_ind=.x$deriv_ind))
            xc <- dplyr::group_modify(.data=xc, .f=~data.frame(deriv_ind=.x$deriv_ind, .x$geometry))
            xc <- dplyr::arrange(xc, .data$deriv_group, .data$contlabel)
            xc <- sf::st_as_sf(xc)
        }
        ## add non-integer contour levels
        else
        {
            xc <- dplyr::filter(x$sf, .data$contlabel %in% cont)
            xc <- dplyr::arrange(xc, .data$deriv_group, .data$estimate)
            cont2 <- cont[!(cont %in% x$sf$contlabel)]
            cont2 <- cont2[cont2>=0 & cont2<=100]
            cont2 <- cont2[round(cont2,0)!=cont2]
            sfname <- attr(xc, "sf_column")
            if (length(cont2)>0)
            {   
                cont.geom <- dplyr::group_modify(.data=fhat, .f=~dplyr::tibble(geometry=list(st_contourline_kdde(x=untidy_ks(.x), which_deriv_ind=.x$deriv_ind, cont=cont2, digits=digits)), deriv_ind=.x$deriv_ind))
                cont.geom <- lapply(cont.geom$deriv_ind, function(.) cbind(dplyr::select(cont.geom[.,], -!!sfname), cont.geom[[sfname]][[.]]))
                cont.geom <- do.call(rbind, cont.geom)
                cont.geom <- sf::st_sf(cont.geom, crs=sf::st_crs(x$sf))
                #xc <- rbind(xc, cont.geom)
                xc <- rbind(xc, st_create_label(cont.geom, is_kdde=TRUE))
            }
            cont.pos <- cont[cont>=0]
            xc$contlabel <- factor(xc$contlabel, levels=c(-sort(cont.pos), sort(cont.pos, decreasing=TRUE)))
        }
         
        sfname <- attr(xc, "sf_column")
        if (is.factor(xc$estimate)) xc$estimate <- unfactor(xc$estimate)

        xc <- dplyr::mutate(xc, pos=sign(.data$estimate), .before=!!sfname)
        xc <- dplyr::arrange(xc, .data$deriv_ind, .data$contlabel)
        xc <- dplyr::group_by(xc, .data$deriv_group, .data$pos, .add=TRUE)
        xc <- dplyr::group_modify(xc, .f=~dplyr::arrange(.x, .data$estimate))
        gv <- dplyr::group_vars(xc)
        gv <- setdiff(gv, "pos")
        xc <- dplyr::group_by(xc, dplyr::across({{gv}}))
        xc <- dplyr::select(xc, -"pos")
    } 
    else 
    {
        ## absolute contour levels 
        if (!missing(breaks))
        {
            if (!inherits(breaks, c("matrix","data.frame","tbl_df"))) breaks <- dplyr::as_tibble(data.frame(breaks=breaks))
            if (oct %in% "kda") stop("Explicit contour breaks not supported for kda objects")
            xc <- st_contourline(x=fhat, abs.cont=breaks$breaks, digits=digits)
            xc <- sf::st_sf(xc, crs=sf::st_crs(x$sf))
            xc <- dplyr::arrange(xc, .data$contlabel)
        }
        ## add other contour levels
        else
        {
            xc <- dplyr::filter(x$sf, .data$contlabel %in% cont)
            cont2 <- cont[!(cont %in% x$sf$contlabel)]
            cont2 <- cont2[cont2>=0 & cont2<=100]
            #cont2 <- cont2[round(cont2,0)!=cont2]
            if (length(cont2)>0)
            {   
                cont.geom <- st_contourline(x=fhat, cont=cont2, digits=digits)
                cont.geom <- sf::st_sf(cont.geom, crs=sf::st_crs(x$sf))
                xc <- rbind(xc, st_create_label(cont.geom))
                xc <- dplyr::arrange(xc, .data$contlabel)
                xc <- dplyr::arrange(xc, sign(.data$estimate), abs(.data$estimate))
            }
        }
    }
    xc <- sf::st_as_sf(xc)
    sf::st_crs(xc) <- sf::st_crs(x$grid)
    gv <- dplyr::group_vars(xc)
    
    ## crop contours to bounding box
    if (crop & !is.null(x$grid)) 
    {
        xbbox <- sf::st_as_sfc(sf::st_bbox(x$grid))
        xc <- suppressWarnings(dplyr::group_modify(xc, .f=~sf::st_crop(.x, xbbox))) 
        xc <- sf::st_as_sf(xc)
    }

    ## add 100% contour for KDDE 
    if (oct %in% "kdde" & cont100) xc <- st_add_contour100(xc, xbbox=xbbox, missing_breaks=missing_breaks) 
    
    ## create disjoint contours
    xc <- dplyr::group_by(xc, dplyr::across({{gv}}))
    if (!(oct %in% c("kdr", "kms")) & disjoint) xc <- st_contour_disjoint(xc) 

    ## add more informative labels
    xc <- st_create_label(xc, digits=digits, missing_breaks=missing_breaks, is_kdde=oct %in% "kdde")
    if (oct %in% "ksupp") xc$contregion <- x$sf$contregion
    
    ## factor for $estimate has fixed number of signif digits displayed 
    ## slightly different from $contlabel
    xc$estimate <- round_signif(xc$estimate, digits=digits)
    xc$estimate <- factor(xc$estimate, levels=unique(xc$estimate))

    ## if original data is not tibble
    if (!inherits(x$sf, "tbl_df"))  xc <- sf::st_sf(data.frame(xc))
    xc <- sf::st_sf(xc, crs=sf::st_crs(x$sf))

    return(xc)
}

## add 100% contour
st_add_contour100 <- function(x, xbbox, missing_breaks)
{
    if (missing(xbbox)) xbbox <- sf::st_as_sfc(sf::st_bbox(x))
    xc <- dplyr::group_modify(x, .f=~.st_add_contour100(.x, xbbox=xbbox, missing_breaks=missing_breaks))
    xc <- dplyr::select(xc, names(x))
    xc <- sf::st_sf(xc)

    return(xc)
}

.st_add_contour100 <- function(x, xbbox, missing_breaks)
{
    if (any(x$estimate==0)) xc <- x
    else
    {
        xc100.geom <- sf::st_union(sf::st_geometry(x))
        xc100.geom <- sf::st_difference(xbbox, xc100.geom)
        xc100 <- x[1,]
        sf::st_geometry(xc100) <-  sf::st_geometry(xc100.geom) 
        if (missing_breaks) xc100$contlabel <- "100" else xc100$contlabel <- "0"
        xc100$estimate <- 0;
        xc <- rbind(x[x$estimate<0,], xc100, x[x$estimate>0,])
        xc$contlabel <- factor(xc$contlabel)
    }    
    return(xc)
}

## x = contour regions as multipolygons (e.g. $sf from sf_ks objects)
st_create_label <- function(x, digits=4, is_kdde=FALSE, missing_breaks=TRUE)
{      
    ## create labels for contour regions    
    ## contlabel is deprecated but kept backwards compatibility < 1.1.0
    gv <- dplyr::group_vars(x)
    if (is.factor(x$estimate)) x$estimate <- unfactor(x$estimate)
     
    ## create contregion label
    if (any(names(x) %in% "label")) { xc <- x; xc$contregion <- factor(xc$label) }
    else 
    {
        if (is_kdde)
            xc <- dplyr::mutate(x, contregion=paste0(ifelse(.data$estimate>0, ">", "<"), round_signif(.data$estimate, digits=digits)), .after="estimate")
        else
            xc <- dplyr::mutate(x, contregion=paste0(">", round_signif(.data$estimate, digits=digits)), .after="estimate")
        xc$contregion[xc$contlabel %in% c("0","100")] <- gsub("<", ">", xc$contregion[which(xc$contlabel %in% c("0","100"))-1])
        xc <- dplyr::arrange(xc, .data$estimate, .by.group=TRUE)
        xc$contregion <- gsub("-", "\u2013", gsub(">", "\u2265 ", gsub("<", "\u2264 ", xc$contregion)))
    }
    if (is.factor(xc$contlabel)) 
        xc$contlabel <- factor(droplevels(xc$contlabel), levels=unique(xc$contlabel))
    else xc$contlabel <- factor(xc$contlabel, levels=unique(xc$contlabel))
    xc$contregion <- factor(xc$contregion, levels=unique(xc$contregion))

    ## create contline label
    if (is_kdde)
    {
        xc <- dplyr::mutate(xc, contline=ifelse(.data$estimate==0, 0L, 1L), .after="contregion")
        xc$contline <- factor(xc$contline)
    }

    ## create cont percentage label  
    if (missing_breaks) 
    {   
        xc$contperc <- xc$contlabel
        levels(xc$contperc) <- paste0(levels(xc$contperc), "%")
    }
    else 
    {
        xc$contperc <- NA  
        xc$contperc <- factor(xc$contperc)
    }
    
    sfname <- attr(xc, "sf_column")
    xc <- dplyr::relocate(xc, "contlabel", "contperc", .after="estimate")
    xc <- dplyr::relocate(xc, dplyr::starts_with("cont"), .after="estimate")
    xc <- dplyr::relocate(xc, dplyr::all_of(gv), .before=!!sfname)
    if (any(names(xc) %in% "deriv_ind") & any(names(xc) %in% "deriv_group"))  xc <- dplyr::relocate(xc, "deriv_ind", .before="deriv_group")
    xc <- dplyr::group_by(xc, dplyr::across({{gv}}))
    xc <- dplyr::arrange(xc, dplyr::across({{gv}}), .data$contlabel)
    xc <- sf::st_as_sf(xc)
     
    return(xc)
}

## interpolate sf polygon grid and fill in missing values on x 
st_intergrid <- function(x, attrib, cellsize, verbose=FALSE)
{
    sf::st_geometry(x) <- "geometry"
    sfname <- attr(x, "sf_column")
    if (missing(attrib)) attrib <- 1
    if (is.numeric(attrib)) attrib <- names(x)[attrib] 
    xcrs <- sf::st_crs(x)
    if (sf::st_is_longlat(x)) sf::st_crs(x) <- NA
    
    ## default grid cell size - based on grid cell with mean area 
    if (missing(cellsize))
    {
        stop("Missing cellsize not supported.")
        # xarea <- as.numeric(sf::st_area(x)) #geos::geos_area(x)
        # xarea.mn <- mean(xarea, na.rm=TRUE)
        # xseg1 <- st_rectangle_segments(x[which.min(abs(xarea.mn-xarea)),])
        # cellsize.default <- as.numeric(sf::st_length(x))[1:2] #geos::geos_length(xseg1)[1:2]
        # cellsize <- signif(cellsize.default,3)
    }
    else if (!missing(cellsize)) { if (length(cellsize)<2) cellsize <- rep(cellsize, times=2)[1:2] }
  
    ## create bounding box which fits exactly ncell gridcells of dimension cellsize
    xbbox <- sf::st_as_sfc(sf::st_bbox(x))
    xbbox.seg <- st_bbox_segments(x)
    ncell <- round(as.numeric(sf::st_length(xbbox.seg))[1:2]/cellsize) #round(geos::geos_length(xbbox.seg)[1:2]/cellsize)
    xgrid.bbox <- sf::st_bbox(x)
    xgrid.bbox[3:4] <- xgrid.bbox[1:2] + ncell*cellsize
    xgrid.bbox <- sf::st_as_sfc(xgrid.bbox)
    sf::st_crs(xgrid.bbox) <- sf::st_crs(x)
    
    ## for grid cells with varying areas, subdivide grid cells before interpolation  
    xarea <- as.numeric(sf::st_area(x)) #geos::geos_area(x)
    if (verbose) cat("Start st_make_grid\n")    
    nsubdiv <- ifelse(all(abs(xarea - mean(xarea)) <= 0.1*mean(xarea)), 1, 2) 
    xgrid.subdiv <- sf::st_sf(geometry=sf::st_make_grid(xgrid.bbox, cellsize=cellsize/nsubdiv, crs=sf::st_crs(x), square=TRUE))
    sf::st_geometry(xgrid.subdiv) <- sfname
    xgrid.subdiv <- st_add_index(xgrid.subdiv)
    if (verbose) cat("End st_make_grid\n") 

    ## check if x is already complete 
    xcent <- st_centroid2(x)
    xcent.ind <- sf::st_covered_by(xcent, xgrid.subdiv)
    xcent.ind <- sort(unlist(xcent.ind[lengths(xcent.ind)==1]))
    xcent <- xcent[xcent.ind,]
    if (identical(xcent.ind, seq_len(nrow(x))))
    {
        xgrid.subdiv[[attrib]] <- xcent[[attrib]]
        xgrid.subdiv <- sf::st_sf(xgrid.subdiv)
        xgrid.subdiv <- dplyr::relocate(xgrid.subdiv, dplyr::all_of(attrib), .before=1)
        xgrid <- dplyr::arrange(xgrid.subdiv, .data$cell_id, .data[[attrib]])
    }
    else
    {
        ## assign those sub-divided cells which cover >=2 different original 
        ## grid cells in xgrid to single value that is closest to weighted mean
        ## search for more efficient operation than st_intersection?
        if (verbose) cat("Start st_intersection\n") 
        xgrid.subdiv.int <- suppressWarnings(sf::st_intersection(x, xgrid.subdiv))
        xgrid.subdiv.int <- xgrid.subdiv.int[sf::st_is(xgrid.subdiv.int, "POLYGON"),]
        xgrid.subdiv.int <- dplyr::mutate(xgrid.subdiv.int, area=as.numeric(sf::st_area(xgrid.subdiv.int)), .before=!!sfname)
        #xgrid.subdiv.int <- dplyr::mutate(xgrid.subdiv.int, area=geos::geos_area(xgrid.subdiv.int), .before=!!sfname)
        ## remove small intersections <= 0.01 * grid cell area
        ## as these may unduly affect the attrib weighted mean
        ## if these intersections have attrib value=0
        xgrid.subdiv.int <- dplyr::filter(xgrid.subdiv.int, .data$area>prod(!!cellsize)*1e-2)
        xgrid.subdiv.int <- dplyr::arrange(xgrid.subdiv.int, dplyr::desc(.data$area*.data[[attrib]]))
        xgrid.subdiv.int <- dplyr::group_by(xgrid.subdiv.int, .data$cell_id)
        xgrid.subdiv.int <- dplyr::summarise(sf::st_drop_geometry(xgrid.subdiv.int), attribm=weighted.mean(.data[[attrib]], .data$area, na.rm=TRUE), .attrib=.data[[attrib]][which.min(abs(.data$attribm-.data[[attrib]]))[1]])
        xgrid.subdiv.int <- dplyr::rename_with(xgrid.subdiv.int, function(.) attrib, ".attrib")
        xgrid.subdiv.int$attribm <- NULL

        ## fill in other grid cells that are not intersected 
        xgrid.subdiv <- dplyr::right_join(xgrid.subdiv.int, xgrid.subdiv, by=c("cell_id"))
        xgrid.subdiv <- sf::st_sf(xgrid.subdiv)
        xgrid.subdiv <- dplyr::relocate(xgrid.subdiv, dplyr::all_of(attrib), .before=1)
        xgrid.subdiv <- dplyr::arrange(xgrid.subdiv, .data$cell_id, .data[[attrib]])
        if (verbose) cat("End st_intersection\n") 

        ## combine subdivided grid cells into original grid cells
        if (verbose) cat("Start st_union_grid\n")
        xgrid <- st_union_grid(xgrid.subdiv, attrib=attrib, n=nsubdiv, .f=mean)
        xgrid <- st_add_index(xgrid)
        if (verbose) cat("End st_union_grid\n")
    }

    ## keep track of NA attribute values
    # xgrid[["attrib_na"]] <- xgrid[[attrib]]
    xgrid[[attrib]][is.na(xgrid[[attrib]])] <- 0
    # xgrid <- dplyr::relocate(xgrid, "attrib_na", .after={{attrib}})
    # xgrid <- dplyr::rename_with(xgrid, function(.) paste0(attrib,"_na"), "attrib_na")
    if (is.na(sf::st_crs(xgrid))) sf::st_crs(xgrid) <- xcrs

    if (verbose) cat(paste0("Original grid     sum(", attrib, ") = ", sum(x[[attrib]], na.rm=TRUE)), paste0("\nInterpolated grid sum(", attrib, ") = ", sum(xgrid[[attrib]])), "\n") 

    return(xgrid)
}

## convert complete raster to sf polygon grid
st_raster_as_grid <- function(x)
{
    dims <- dim(x)-1
    xsf <- sf::st_as_sf(x)
    xsf <- cbind(expand.grid(cell_id1=1:dims[1], cell_id2=dims[2]:1), xsf)
    xsf <- dplyr::as_tibble(xsf)
    xsf <- dplyr::arrange(xsf, .data$cell_id2, .data$cell_id1)
    xsf <- sf::st_sf(xsf) 
    sfname <- attr(xsf, "sf_column")
    xsf <- dplyr::relocate(xsf, "cell_id1", "cell_id2", .before=!!sfname)
    
    return(xsf)
}

## prepare kde grid for raster conversion
## x = ks_sf object (output from st_kde/st_as_kde) 
st_add_contour_label <- function(x, cont=c(25,50,75))
{
    if (inherits(x, "sf_ks")) xgrid <- sf::st_drop_geometry(st_add_index(x$grid))
    else xgrid <- sf::st_drop_geometry(st_add_index(x))
    raster.dim <- as.vector(apply(xgrid[,c("cell_id1", "cell_id2")], 2, max))
    breaks <- contourLevels(x, cont=cont)
    contlev <- sort(c(min(x$grid$estimate)-0.1*abs(max(x$grid$estimate)), breaks$breaks))
    breaks$cont <- sort(c(0,head(breaks$cont,n=-1)))
    xcontlabel <- cut(x$grid$estimate, contlev, right=TRUE, labels=breaks$cont)
    xcontperc <- xcontlabel 
    levels(xcontperc) <- paste0(levels(xcontperc), "%")
    xg <- dplyr::mutate(x$grid, contlabel=!!xcontlabel, contperc=!!xcontperc,  nx=raster.dim[1], ny=raster.dim[2], .after="estimate")
    
    return(xg)
}

## add indices, starting from SW corner, to rectangular grid
## that it already sorted from SW corner to NE corner (e.g. output from st_make_grid)
st_add_index <- function(x)
{
    xcrs <- sf::st_crs(x)
    if (!is.na(sf::st_crs(x))) { if (sf::st_is_longlat(x)) sf::st_crs(x) <- NA }
      
    sfname <- attr(x, "sf_column")
    xbbox <- st_bbox_segments(x)
    h1 <- xbbox[1]; v1 <- xbbox[2]  
    #xh <- geos::geos_intersects(x, h1)
    #xv <- geos::geos_intersects(x, v1)
    xh <- as.vector(sf::st_intersects(sf::st_geometry(x), sf::st_geometry(h1), sparse=FALSE))
    xv <- as.vector(sf::st_intersects(sf::st_geometry(x), sf::st_geometry(v1), sparse=FALSE))
    xhv <- expand.grid(cell_id2=1:sum(xh), cell_id1=1:sum(xv))
    xgrid <- dplyr::mutate(x, cell_id=1:nrow(x), cell_id1=!!xhv[,1], cell_id2=!!xhv[,2], .before=!!sfname)
    
    return(xgrid) 
}

## add indices, starting from SW corner, to rectangular grid
## for any grid
st_add_index_unsort <- function(x)
{
    xcrs <- sf::st_crs(x)
    if (!is.na(sf::st_crs(x)))
    {
        if (sf::st_is_longlat(x)) sf::st_crs(x) <- NA
    } 
    ## centroids of first horizontal row and first vertical col
    xbbox <- st_bbox_segments(x)
    h1 <- xbbox[1]; v1 <- xbbox[2]

    xh.cent <- sf::st_filter(x, h1) 
    xv.cent <- sf::st_filter(x, v1)    
    xh.cent <- suppressWarnings(sf::st_centroid(xh.cent))
    xv.cent <- suppressWarnings(sf::st_centroid(xv.cent))

    sfname <- attr(x, "sf_column")
    xcent <- suppressWarnings(sf::st_centroid(x))
    ## compute complete aligned grid with row,col indices
    xgrid <- dplyr::mutate(x, cell_id1=sf::st_nearest_feature(xcent, xh.cent), cell_id2=sf::st_nearest_feature(xcent, xv.cent), .before=!!sfname)
    xgrid$cell_id1 <- factor(xgrid$cell_id1) 
    levels(xgrid$cell_id1) <- 1:(length(levels(xgrid$cell_id1)))
    xgrid$cell_id1 <- unfactor(xgrid$cell_id1)
    xgrid$cell_id2 <- factor(xgrid$cell_id2) 
    levels(xgrid$cell_id2) <- 1:(length(levels(xgrid$cell_id2)))
    xgrid$cell_id2 <- unfactor(xgrid$cell_id2)
    xgrid <- dplyr::arrange(xgrid, .data$cell_id2, .data$cell_id1)
    xgrid <- dplyr::mutate(xgrid, cell_id=1:nrow( xgrid), .before="cell_id1")
    
    if (is.na(sf::st_crs(xgrid))) sf::st_crs(xgrid) <- xcrs

    return(xgrid)
}

median_closest <- function(x) { md <- median(x, na.rm=TRUE); ifelse(is.na(md), NA, x[which.min(abs(md-x))]) }
mean_closest <- function(x) { mn <- mean(x, na.rm=TRUE); ifelse(is.na(mn), NA, x[which.min(abs(mn-x))]) }

## aggregate grid by combining n x n grid cells  
st_union_grid <- function(x, attrib, n=1, .f)
{
    if (missing(.f)) .f <- function(x) (mean(x, na.rm=TRUE))
    xu <- dplyr::mutate(x, celln_id1=(.data$cell_id1-1)%/%n+1, celln_id2=(.data$cell_id2-1)%/%n+1, .after="cell_id2")
    xu <- dplyr::group_by(xu, .data$celln_id1, .data$celln_id2)
    xu2 <- dplyr::filter(xu, dplyr::n()>1)
    xu2.attrib <- dplyr::summarise(xu2, .attrib=.f(.data[[attrib]]), do_union=FALSE, .groups="drop")
    xu2.attrib <- sf::st_convex_hull(xu2.attrib)
    xu2.attrib <- dplyr::rename(xu2.attrib, cell_id1="celln_id1", cell_id2="celln_id2")
    xu2.attrib <- dplyr::rename_with(xu2.attrib, function(.) attrib, ".attrib")
    xu2.attrib <- dplyr::relocate(xu2.attrib, {{attrib}}, .before=1)
    
    xu.attrib <- rbind(xu2.attrib, dplyr::select(dplyr::filter(xu, dplyr::n()<=1), names(xu2.attrib)))
    xu.attrib <- dplyr::arrange(xu.attrib, .data$cell_id2, .data$cell_id1)
    xu.attrib <- dplyr::mutate(xu.attrib, cell_id=1:nrow(xu.attrib), .before="cell_id1")
    
    return(xu.attrib)
}

## x= st_kde object
st_maximum_attrib <- function(x, dist=1000, nshape=1, shape="circle", attrib_min, cont_start=0.25, verbose=FALSE)
{
    ## some defaults
    sfname <- attr(x$sf, "sf_column")
    attrib <- "estimate" 
    x$grid <- st_add_index(x$grid) 
    xorig <- x
    #cellsize <- geos::geos_length(st_bbox_segments(xorig$grid))[1:2]/c(max(xorig$grid$cell_id1), max(xorig$grid$cell_id2))
    cellsize <- as.numeric(sf::st_length(st_bbox_segments(xorig$grid)))[1:2]/c(max(xorig$grid$cell_id1), max(xorig$grid$cell_id2))
    cellsize.ind <- ceiling(dist/cellsize)

    ## sum is aggregate statistic function
    .f <- sum
    shape <- match.arg(shape, c("circle", "rect"))
    # if (verbose & all(as.numeric(sf::st_length(st_rectangle_segments(x$grid[1,]))) >dist)) warning("dist is smaller than grid cell size so results may not be accurate", immediate.=TRUE)
    if (missing(attrib_min)) { missing_attrib_min <- TRUE; attrib_min <- min(x$grid[[attrib]]) } else { missing_attrib_min <- FALSE; nshape <- 100 }
    if (nshape<1) nshape <- 1

    xattrib_max <- attrib_min
    xattrib_max_prev <- 10*xattrib_max 
    xc.list <- NULL
    i <- 1
    while ((abs(xattrib_max-xattrib_max_prev)/xattrib_max >= 0.1 | xattrib_max >= attrib_min) & (i <= nshape))
    {
        ## copy cell indices from xorig to modified x
        x$grid <- dplyr::mutate(x$grid, cell_id=xorig$grid$cell_id, cell_id1=xorig$grid$cell_id1, cell_id2=xorig$grid$cell_id2, .before=!!sfname)
        xc <- .st_maximum_attrib(x=x, dist=dist, nshape=1, .f=.f, cont_start=cont_start, cont_step=0.25, cellsize.ind=cellsize.ind, align_grid=TRUE)
        xc <- dplyr::arrange(xc, dplyr::desc(.data[[attrib]]))
        ## keep only mutually exclusive circles
        if (nrow(xc)>1) xc <- xc[sort(unique(sapply(sf::st_intersects(xc), min))),]
        #if (shape=="square") xc <- sf::st_minimum_rotated_rectangle(xc)
        xc <- sf::st_sf(xc)
        xattrib_max_prev <- xattrib_max

        ## update sum of attrib
        xc[[attrib]] <- sapply(st_filter_approx(x=xorig, y=xc, ind=cellsize.ind), function(.) do.call(.f, list(.[[attrib]])))
        xattrib_max <- max(xc[[attrib]])
        xc.list <- c(xc.list, list(xc))  
        if (verbose) cat("i =", i, ", max", attrib, "=", xattrib_max, "\n") 
        
        ## update intersection of (x$grid, xc) attrib = 0
        ## approx intersection based on cell indices is faster than sf::st_intersection           
        for (j in 1:nrow(xc))
            x$grid[x$grid$cell_id1 %in% (xc$cell_id1[j] + (-cellsize.ind[1]:cellsize.ind[1])) & x$grid$cell_id2 %in% (xc$cell_id2[j] + (-cellsize.ind[2]:cellsize.ind[2])),][[attrib]] <- 0
        
        ## update st_as_kde
        if (all(zapsmall(x$grid$estimate)==0)) break
        i <- i+1
        x <- st_as_kde(x$grid, cont=1)   
    }
    xc <- do.call(rbind, xc.list)
    xc <- sf::st_sf(xc)

    if (shape=="rect")
    {
        ## replace circle polygon
        xc.geom <- lapply(1:nrow(xc), function(.) sf::st_as_sfc(sf::st_bbox(xc[.,])))
        sf::st_geometry(xc) <- sf::st_as_sfc(do.call(rbind, xc.geom), crs=sf::st_crs(x$sf))
    }
    else 
    {
        ## recalculate attrib summaries for circle polygons 
        xg <- st_filter_approx(x=xorig, y=xc, ind=cellsize.ind)
        xg <- lapply(1:nrow(xc), function(.) suppressWarnings(sf::st_intersection(xg[[.]], xc[.,])))
        xc[[attrib]] <- sapply(xg, function(.) do.call(.f, list(.[[attrib]])))
    }
    
    xc <- dplyr::arrange(xc, dplyr::desc(.data[[attrib]]))
    if (missing_attrib_min) xc <- xc[1:nshape,]
    else xc <- xc[xc[[attrib]] >= attrib_min,]

    return(xc)
}

## x= sf_ks object with st_add_index applied to x$grid
## y = sf object with st_add_index applied
st_filter_approx <- function(x, y, ind)
{ 
    lapply(1:nrow(y), function(.) 
    { 
        x$grid[x$grid$cell_id1 %in% (y[.,]$cell_id1 + c(-ind[1]:ind[1])) & x$grid$cell_id2 %in% (y[.,]$cell_id2 + c(-ind[2]:ind[2])),]
    })
}

.st_maximum_attrib <- function(x, dist, nshape, .f, cont_start, cont_step, cellsize.ind, align_grid)
{
    sfname <- attr(x$sf, "sf_column")
    attrib <- "estimate" 

    ct <- cont_start
    xc <- st_get_contour(x, cont=ct)
    xc <- sf::st_cast(sf::st_cast(xc, to="MULTIPOLYGON"), to="POLYGON", warn=FALSE)
    xc <- sf::st_minimum_bounding_circle(xc)
    xc$radius_mbc <- sqrt(as.numeric(sf::st_area(xc))/pi)
  
    ## search by increasing contour level 
    while (ct < 100 & sum(xc$radius_mbc<dist)<nshape) 
    {
        xc <- st_get_contour(x, cont=ct)
        xc <- sf::st_cast(sf::st_cast(xc, to="MULTIPOLYGON"), to="POLYGON", warn=FALSE)
        xc <- sf::st_minimum_bounding_circle(xc)
        xc$radius_mbc <- sqrt(as.numeric(sf::st_area(xc))/pi)
        ct <- ct + cont_step
    } 
    xc <- dplyr::relocate(xc, "radius_mbc", .before=!!sfname)
    
    ## reduce grid to speed up st_filter
    if (is.factor(xc[[attrib]])) att <- min(unfactor(xc[[attrib]])) else att <- min(xc[[attrib]])
    xg <- x$grid[x$grid[[attrib]]>0.2*att,]

    ## align centre of circle to grid cell centroid
    xc <- st_centroid2(xc)
    if (align_grid) xc <- st_centroid2(sf::st_filter(xg, xc))
    xc <- sf::st_buffer(xc, dist=dist, nQuadSegs=60)

    ## compute approx attrib summary
    xc[[attrib]] <- sapply(st_filter_approx(x=x, y=xc, ind=cellsize.ind), function(.) do.call(.f, list(.[[attrib]])))
    xc <- dplyr::mutate(xc, radius=!!dist, .before=!!sfname)
    xc <- dplyr::arrange(xc, dplyr::desc(.data[[attrib]]))
    
    return(xc)
}

## ----echo=FALSE, message=FALSE------------------------------------------------
knitr::opts_chunk$set(global.par=TRUE, collapse=TRUE, comment="#>", fig.width=5, fig.height=5, fig.align="center", dpi=96)
options(tibble.print_min=3L, tibble.print_max=3L)

## ----message=FALSE------------------------------------------------------------
library(eks)
library(colorspace)
library(ggplot2)
library(dplyr)
set_theme(theme_bw())

## crabs data 
data(crabs, package="MASS")
crabs2 <- select(crabs, FL, CW)
xlab <- "Frontal lobe size (mm)"
ylab <- "Carapace width (mm)"

## ----message=FALSE------------------------------------------------------------
## KDE contour plot + scatter plot
tkde2 <- tidy_kde(crabs2)
gkde2 <- ggplot(tkde2, aes(x=FL, y=CW)) + labs(x=xlab, y=ylab)
gkde2 + geom_point_ks(colour=8) + geom_contour_ks(colour=1)

## ----message=FALSE------------------------------------------------------------
## geom_density_2d KDE contour plot + scatter plot
mkde2 <- ggplot(crabs2, aes(x=FL, y=CW))
mkde2 + geom_point(colour=8) + geom_density_2d(colour=1, bins=4)

## ----message=FALSE------------------------------------------------------------
## KDE filled contour plot
gkde2 + geom_contour_filled_ks(colour=1) 

## ----message=FALSE------------------------------------------------------------
## KDE continuous colour scale plot
gkde2 + geom_raster(aes(fill=estimate), interpolate=TRUE) + 
   scale_fill_continuous_sequential(palette="Heat")

## ----message=FALSE, fig.width=7-----------------------------------------------
crabs2g <- select(crabs, FL, CW, sp)
crabs2g <- group_by(crabs2g, sp)
tkde2g <- tidy_kde(crabs2g)
gkde2g <- ggplot(tkde2g, aes(x=FL, y=CW, group=sp)) 

## facetted KDE contour plots + scatter plots
gkde2g + geom_point_ks(colour=8) + labs(x=xlab, y=ylab, colour="Species") + 
   scale_colour_manual(values=c(4, 7)) + 
   guides(colour=guide_legend(title="Species")) +
   geom_contour_ks(aes(colour=sp)) + facet_wrap(~sp) 

## ----fig.width=7--------------------------------------------------------------
## facetted KDE filled contour plots with fixed contour levels for all facets
bkde2g <- contour_breaks(tkde2g)
gkde2g + geom_contour_filled_ks(breaks=bkde2g, colour=1) + facet_wrap(~sp)

## ----message=FALSE------------------------------------------------------------
library(sf)

## Grevillea data
data(grevilleasf, package="eks")
grevilleasf <- mutate(grevilleasf, species=factor(species))
paradoxa <- filter(grevilleasf, name %in% "Grevillea paradoxa")
eryngioides <- filter(grevilleasf, name %in% "Grevillea eryngioides")
grevillea_ep <- filter(grevilleasf, name %in% c("Grevillea eryngioides", 
   "Grevillea paradoxa"))
grevillea_ep <- group_by(grevillea_ep, name)
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)

## WA polygon
data(wa, package="eks")
gwa <- geom_sf(data=wa, fill=NA, colour=1)

## -----------------------------------------------------------------------------
## base R scatter plot
plot(st_geometry(wa), xlim=xlim, ylim=ylim)
plot(st_geometry(eryngioides), add=TRUE, col=3, pch=16, cex=0.5)
plot(st_geometry(paradoxa), add=TRUE, col=6, pch=17, cex=0.5)
mapsf::mf_legend(type="symb", val=c("Grevillea eryngioides", "Grevillea paradoxa"), 
   pal=c(3,6), pch=16:17, cex=c(1,1), title="Species", pos="bottomleft")
## geom_sf scatter plot 
ggplot() + gwa + theme_sf() +
   geom_sf(data=grevillea_ep, aes(colour=name, shape=name)) + 
   coord_sf(xlim=xlim, ylim=ylim) + scale_colour_manual(values=c(3, 6)) + 
   guides(colour=guide_legend(title="Species"), shape=guide_legend(title="Species"))

## -----------------------------------------------------------------------------
skde1 <- st_kde(paradoxa)

## base R contour plot
plot(st_geometry(wa), xlim=xlim, ylim=ylim)
plot(st_geometry(paradoxa), add=TRUE, pch=16, col=8, cex=0.5)
plot(skde1, add=TRUE, col=NA, border=1, legend=FALSE)

## -----------------------------------------------------------------------------
## geom_sf contour plot
gs <- ggplot(skde1) + gwa + theme_sf()
gs + geom_sf(data=paradoxa, col=8, size=0.5) + 
   geom_sf(data=st_get_contour(skde1), colour=1, fill=NA) + 
   coord_sf(xlim=xlim, ylim=ylim)

## -----------------------------------------------------------------------------
## R base filled contour plot
plot(st_geometry(wa), xlim=xlim, ylim=ylim)
plot(skde1, add=TRUE)
## geom_sf filled contour
gs + geom_sf(data=st_get_contour(skde1), aes(fill=contperc)) + 
  coord_sf(xlim=xlim, ylim=ylim) 

## ----fig.width=7--------------------------------------------------------------
skde1g <- st_kde(grevillea_ep)

## facetted geom_sf filled contour
gs + geom_sf(data=st_get_contour(skde1g), aes(fill=contperc)) + 
   facet_wrap(~name) + coord_sf(xlim=xlim, ylim=ylim) 
## facetted geom_sf filled contour with fixed contour levels for all facets
bkde1g <- contour_breaks(skde1g, group=TRUE)
gs + geom_sf(data=st_get_contour(skde1g, breaks=bkde1g), aes(fill=estimate)) +
   facet_wrap(~name) + coord_sf(xlim=xlim, ylim=ylim) 

#####################################################################
## Auxiliary functions for st_* functions
#####################################################################

## plot function for sf_ks object
ggplot.sf_ks <- function(data=NULL, mapping=aes(), ..., which_geometry="sf")
{
    if (!is.null(data))
    {
        g <- match.arg(which_geometry, c("sf","grid"))
        data1 <- data[[g]] 
        transp_neutral <- object_class(data, type="ks") %in% "kdde"
    }
    else transp_neutral <- FALSE
    p <- ggplot2::ggplot(data=data1, mapping=mapping, ...) + labs_ks(data) + guides_ks(data)
    p <- p + scale_colour_ks(data) + scale_fill_ks(data, transp_neutral=transp_neutral)

    return(p)
}

plot.sf_ks <- function(x, ...) { plot_sf_ks(x=x, ...) }

plot_sf_ks <- function(x, which_geometry="sf", cont=c(25,50,75), abs_cont=breaks, breaks, which_deriv_ind=1, main="", pal, col, border, lty, transp_neutral, alpha, legend=TRUE, key.pos=NULL, legend.title, pos="bottomleft", digits, ...) 
{
    g <- match.arg(which_geometry, c("sf","grid"))
    y <- x[[g]]
    oct <- head(object_class(x, type="tks"), n=1)
    oc <- object_class(x, type="ks")
    if (missing(digits)) digits <- ifelse(oc %in% c("kcde", "kcopula", "kde.loctest", "kfs", "kdr"), 2, 4)

    if (missing(transp_neutral)) transp_neutral <- object_class(x, type="ks") %in% "kdde"
    if (missing(alpha)) { if (g=="sf") alpha <-1 else alpha <- 0.1 }
    
    if (missing(col))
    {
        if (oct %in% "kdr") col <- ggplot2::alpha(6, alpha=alpha)
        else if (oct %in% "kfs") col <- ggplot2::alpha(7, alpha=alpha)
        else if (oct %in% "ksupp") col <- NA
        else if (oct %in% "kquiver") col <- ggplot2::alpha(1, alpha=alpha)
    
        pal.missing <- FALSE
        if (missing(pal)) 
        {
            pal.missing <- TRUE
            if (oct %in% "kdde") 
            { 
                pal1 <- function(.) colorspace::diverging_hcl(n=., palette="Blue-Red", alpha=alpha)
                if (missing(breaks)) breaks1 <- sort(c(-(1:length(cont)), 1:length(cont)))
                else 
                {
                    if (inherits(breaks, c("matrix", "data.frame", "tbl_df"))) breaks1 <- breaks$breaks else breaks1 <- breaks
                    breaks1 <- sort(breaks1)
                }
                pal <- function(.) { ks::col.diverging(f=pal1, levels=breaks1, display="filled.contour", transp.neutral=transp_neutral) }
                col <- pal()
            }
            else if (oct %in% "kcurv") pal <- function(.) { colorspace::sequential_hcl(n=., h1=30, c1=360, c2=60, alpha=alpha, rev=TRUE) }
            else if (oct %in% "kde.loctest") pal <- function(.) { colorspace::qualitative_hcl(n=., palette="Dark2", alpha=alpha, rev=TRUE) }
            else if (oct %in% "kcde") pal <- function(.) { colorspace::sequential_hcl(n=., rev=FALSE, palette="Viridis", alpha=alpha) }
            else if (oct %in% c("kroc","kms")) pal <- function(.) { colorspace::qualitative_hcl(n=., palette="Set2", alpha=alpha) }
            else if (any(oct %in% "kda")) pal <- function(.) { colorspace::qualitative_hcl(n=., palette="Dark2", alpha=alpha) }
            else pal <- function(.) { colorspace::sequential_hcl(n=., rev=TRUE, palette="Heat", alpha=alpha) }
        }
        if ((oct %in% "kda") & (g=="sf"))
        {
            ng <- length(table(dplyr::group_indices(y)))
            cols <- pal(ng)
            if (missing(breaks)) nc <- length(cont) else nc <- length(breaks)
            col <- rep(cols, each=nc)
        }
    }
    else 
    {
        col <- ggplot2::alpha(col, alpha=alpha)
        pal.missing <- TRUE
    }

    if (g=="sf")
	{   
        if (oct %in% c("kdr", "kms"))
        {
            yd <- dplyr::select(y, dplyr::all_of("label"))
        }
        else if (oct %in% "ksupp")
        {
            yd <- dplyr::select(y, dplyr::all_of("contlabel")) 
        }
        else if (oct %in% "kdde")
        {
            y <- st_get_contour(x, cont=cont, breaks=abs_cont, digits=digits) 
            y <- y[y$deriv_ind == which_deriv_ind,]
            y$estimate <- droplevels(y$estimate)
            yd <- dplyr::select(y, dplyr::all_of("estimate")) 

            if (length(y$estimate)!=length(col)) { breaks1 <- tail(unfactor(y$estimate),n=-1); col <- pal() }
        }
        else
        {
            y <- st_get_contour(x, cont=cont, breaks=abs_cont, digits=digits)
            yd <- dplyr::select(y, dplyr::all_of("estimate"))
        }
        if (missing(lty)) 
        { 
            if (oct %in% "kdde") 
            { lty <- as.numeric(levels(y$contline))[y$contline]; lty[is.na(lty)] <- 0 }
            else lty <- rep(1, nrow(yd))
            if (any(names(yd) %in% "estimate")) lty[y$estimate==0] <- 0 
        }
        
        if (missing(col)) plot(yd, main=main, pal=pal, lty=lty, key.pos=key.pos, ...)
        else 
        { 
            if (oct %in% "kda") 
            {
                if (missing(border)) plot(yd, main=main, col=NA, border=col, lty=lty, key.pos=key.pos, ...)
                else plot(yd, main=main, col=NA, border=border, lty=lty, key.pos=key.pos, ...)
            }
            else 
            {
                if (missing(border)) plot(yd, main=main, col=col, lty=lty, key.pos=key.pos, ...)
                else plot(yd, main=main, col=col, border=border, lty=lty, key.pos=key.pos, ...)
            }
        }
	}
	else if (g=="grid") 
	{
        if (oct %in% "kdde") y <- y[y$deriv_ind == which_deriv_ind,]
        yd <- dplyr::select(y, dplyr::all_of("estimate"))
        if (missing(lty)) lty <- 1
        if (oct %in% "kda")
        {
            if (missing(col)) plot(yd, main=main, pal=pal, lty=lty, border=border, key.pos=key.pos, ...)
            else plot(yd, main=main, col=col, lty=lty, border=border, key.pos=key.pos, ...)
        }
        else 
        {
            if (missing(col)) plot(yd, main=main, pal=pal, lty=lty,, key.pos=key.pos, ...)
            else plot(yd, main=main, col=col, lty=lty, key.pos=key.pos, ...)
        }
    }

    ## add legend
    if (legend)
    {   
        if (!requireNamespace("mapsf", quietly=TRUE)) stop("Install the mapsf package as it is required.", call.=FALSE)       
        forms <- list(...)
        forms <- forms[names(forms) %in% c("pos", "val", "pal", "title", "title_cex", "val_cex", "col_na", "no_data", "no_data_txt", "frame", "border", "bg", "fg", "cex")]
        forms <- forms[!(names(forms) %in% "border")] 
        gu <- guides_ks(dplyr::add_row(dplyr::ungroup(x$tidy_ks), ks=list(2L)))
       
        if (is.list(gu)) gu <- list(guides=unlist(lapply(gu, function(.) unlist(.[["guides"]]))))
        if (missing(legend.title)) gu.title <- gu$guides$fill$params$title
        else gu.title <- legend.title
        
        if (g=="sf")
        {
            if (oct %in% c("kdde", "kdr")) 
            {
                contlabel <- sort(unique(y$contregion), decreasing=TRUE) 
                nc <- length(contlabel)
            }
            else 
            { 
                contlabel <- sort(unique(y$contlabel), decreasing=TRUE)
                if (missing(breaks)) contlabel <- paste0(contlabel,"%")
                nc <- length(contlabel)
            }
        }

        if (g=="grid")
        {
            if (oct %in% "kda")
            {
                gv <- levels(dplyr::pull(sf::st_drop_geometry(y), .data$label))
                do.call(mapsf::mf_legend, args=c(list(type="typo", val=gv, pal=pal(length(gv)), pos=pos, title=gu.title, box_border=NA), forms))
            }
            else
            {
                y2 <- st_get_contour(x, cont=cont, breaks=abs_cont, digits=digits)
                yd2 <- dplyr::select(y2, dplyr::all_of("estimate"))
                gv <- yd2$estimate
                do.call(mapsf::mf_legend, args=c(list(type="choro", val=gv, pal=pal(length(gv)), pos=pos, title=gu.title), forms))
            }
        }
        else if (!is.null(forms$border)) 
        {
            if (oct %in% "kda")
            {
                gv <- levels(dplyr::pull(sf::st_drop_geometry(y), dplyr::all_of(dplyr::group_vars(y))))
                ng <- length(gv)
                do.call(mapsf::mf_legend, args=c(list(type="symb", val=gv, pal=unique(forms$border), pos=pos, title=gu.title, cex=rep(3,ng), pch=rep("-", ng)), forms[!(names(forms) %in% c("cex","pch"))]))
            }
            else if (oct %in% "kde.loctest")
            {
                gv <- levels(dplyr::pull(sf::st_drop_geometry(y), .data$label))
                ind <- order(gv, decreasing=TRUE)
                do.call(mapsf::mf_legend, args=c(list(type="symb", val=gv[ind], pal=rev(border), pos=pos, title=gu.title, cex=3, pch="-"), forms[!(names(forms) %in% c("cex","pch"))]))
            }
            else 
            {
                if (!all(is.na(forms$border)) & pal.missing) 
                {
                    forms <- forms[!(names(forms) %in% c("cex","pch"))]
                    do.call(mapsf::mf_legend, args=c(list(type="symb", val=contlabel, pal=rev(forms$border), pos=pos, title=gu.title, cex=3, pch="-"), forms[!(names(forms) %in% c("cex","pch"))]))
                }
                else 
                    do.call(mapsf::mf_legend, args=c(list(type="typo", val=contlabel, pal=rev(pal(nc)), pos=pos, title=gu.title), forms)) 
            }
        }
        else if (!missing(col))
        {
            ## bug in mapsf::mf_legend that doesn't allow single values in val
            if (oct %in% "kdr") do.call(mapsf::mf_legend, args=c(list(type="symb", val=c(levels(y$label), ""), pal=c(col,NA),  pos=pos, title=gu.title, cex=3, pch="-"), forms[!(names(forms) %in% c("cex","pch"))])) 
            else if (oct %in% "kfs") do.call(mapsf::mf_legend, args=c(list(type="typo", val=c(y$label[1],""), pos=pos, pal=c(col,NA), box_border=c("#333333",NA), title=gu.title), forms))  
            ## mapsf::mf_legend doesn't accept lty argument
            else if (oct %in% "ksupp") do.call(mapsf::mf_legend, args=c(list(type="typo", val=c(contlabel,""), pos=pos, pal=c(col,NA), box_border=c("#333333",NA), title=gu.title), forms))
            else if (oct %in% "kda")
            {
                gv <- levels(dplyr::pull(sf::st_drop_geometry(y), dplyr::all_of(dplyr::group_vars(y))))
                ng <- length(gv)
                do.call(mapsf::mf_legend, args=c(list(type="typo", val=gv, pal=pal(ng), pos=pos, title=gu.title), forms))
            }
            else if (oct %in% "kde.loctest" & all(is.na(col)))
            {
                gv <- levels(dplyr::pull(sf::st_drop_geometry(y), .data$label))
                ind <- order(gv, decreasing=TRUE)
                do.call(mapsf::mf_legend, args=c(list(type="typo", val=gv[ind], pos=pos, pal=rep(NA,length(border)), box_border=rev(border), title=gu.title), forms))
            }
            else 
            {
                if (!missing(breaks) | oct %in% "kdde") { col <- rev(col) }
                do.call(mapsf::mf_legend, args=c(list(type="typo", val=contlabel, pal=col, pos=pos, title=gu.title), forms))
            }
        }
        else if (!missing(pal)) 
        {        
            if (oct %in% "kms") 
            {
                gv <- levels(dplyr::pull(sf::st_drop_geometry(y), .data$label))
                nc <- length(gv)
                formsg <- list(...)
                if (is.null(formsg$pch)) pch <- 1 else pch <- unique(formsg$pch) 
                forms <- forms[!(names(forms) %in% c("cex","pch"))]
                do.call(mapsf::mf_legend, args=c(list(type="symb", val=gv, pal=pal(nc), pos=pos, title=gu.title, cex=rep(1,nc), pch=pch), forms[!(names(forms) %in% c("cex","pch"))]))
            }
            else if (oct %in% "kde.loctest")
            {
                gv <- levels(dplyr::pull(sf::st_drop_geometry(y), .data$label))
                ind <- order(gv, decreasing=TRUE)

                do.call(mapsf::mf_legend, args=c(list(type="typo", val=gv[ind], pal=pal(length(gv))[ind], pos=pos, title=gu.title), forms))
            }
            else 
                do.call(mapsf::mf_legend, args=c(list(type="typo", val=contlabel, pal=rev(pal(nc)), pos=pos, title=gu.title), forms))
        }
    }
}

## probability contour regions
contourLevels.sf_ks <- function(x, ..., cont=c(25,50,75)) { ks::contourLevels(x=x$tidy_ks, ..., cont=cont) }

## convert contour polygons in ks object to multipolygon geometry
## x = ks object
## output sf-geometry is named "geometry"   
.st_contourline <- function(x, cont, abs.cont, edge_zero=TRUE, digits)
{   
    ## flag to use values of abs.cont in contlabel rather than names(abs.cont)
    if (missing(abs.cont)) abs.cont.flag <- FALSE 
    else if (is.null(names(abs.cont))) abs.cont.flag <- TRUE
    else abs.cont.flag <- !all(grepl("[0-9]%", names(abs.cont)))
    
    if (missing(abs.cont) & missing(cont)) abs.cont <- x$cont
    if (!missing(cont)) 
    {
        if (class(x) %in% c("kfs", "kde.loctest")) abs.cont <- abs.cont
        else abs.cont <- contourLevels(x, cont=cont)
    }
    abs.cont <- sort(abs.cont)
   
    ## convert contourLines output to linestring
    cline2ls <- function(y) 
    { 
        sf::st_linestring(as.matrix(data.frame(x=y["x"], y=y["y"]))) 
    }
    
    ## add zero estimate values at edges of estimation grid 
    if (edge_zero)
    {
        x.zero <- x
        delta <- sapply(x$eval.points, function(.) head(diff(.), n=1))

        a <- 1
        x.zero$eval.points[[1]] <- c(min(x.zero$eval.points[[1]]) - delta[1]*(a:1), x.zero$eval.points[[1]], max(x.zero$eval.points[[1]]) + delta[1]*(1:a))
        x.zero$eval.points[[2]] <- c(min(x.zero$eval.points[[2]]) - delta[2]*(a:1), x.zero$eval.points[[2]], max(x.zero$eval.points[[2]]) + delta[2]*(1:a))
        x.zero$estimate <- matrix(0, ncol=ncol(x.zero$estimate)+2*a, nrow=nrow(x.zero$estimate)+2*a)
        x.zero$estimate[(1:nrow(x$estimate))+a, (1:ncol(x$estimate))+a] <- x$estimate
        x <- x.zero
    }

    j <- 0   
    for (i in 1:length(abs.cont))
    {
        if (is.list(x$estimate)) estimate <- x$estimate[[1]] else estimate <- x$estimate
    	cline <- contourLines(x=x$eval.points[[1]], y=x$eval.points[[2]], z=estimate, levels=abs.cont[i])
        cline.len <- lengths(lapply(cline, getElement, "x"))
        cline <- cline[cline.len>=4]
     	cont.polygon <- lapply(cline, cline2ls)
    	cont.polygon <- try(sf::st_polygon(cont.polygon), silent=TRUE)
        if (inherits(cont.polygon, "try-error")) cont.polygon <- sf::st_polygon()
        cont.polygon <- sf::st_sf(geometry=sf::st_geometry(cont.polygon))

        if (abs.cont.flag) 
            cont.polygon <- cbind(contlabel=round_signif(abs.cont[i], digits=digits), estimate=as.numeric(abs.cont[i]), cont.polygon) 
    	else 
            cont.polygon <- cbind(contlabel=as.numeric(gsub("%", "", names(abs.cont[i]))), estimate=as.numeric(abs.cont[i]), cont.polygon) 	
        j <- j + !sf::st_is_empty(cont.polygon)
    	if (j<=1) { cont.polygon.all <- cont.polygon ; j <- j+1 }
        else if (j>1) cont.polygon.all <- rbind(cont.polygon.all, cont.polygon)
    }

    cont.polygon.all <- cont.polygon.all[!sf::st_is_empty(cont.polygon.all),]
    cont.polygon.all <- dplyr::arrange(cont.polygon.all, .data$estimate, .data$contlabel)
    
    if (nrow(cont.polygon.all)>0)
    {
        cont.polygon.all <- sf::st_make_valid(cont.polygon.all, geos_method="valid_linework")
        cont.polygon.all <- dplyr::distinct(cont.polygon.all)
       
        if (abs.cont.flag) 
        {
            cont.polygon.all$contlabel <- factor(cont.polygon.all$contlabel)
        }
        else
        {
            cont.polygon.all$contlabel <- factor(cont.polygon.all$contlabel)
            levels(cont.polygon.all$contlabel) <- as.character(100-as.numeric(levels(cont.polygon.all$contlabel)))
        }
    }
   
	return(cont.polygon.all)
}

## x = tibble with ks column (output from first part of st_ks)
st_contourline <- function(x, cont, abs.cont, edge_zero=TRUE, digits, ...)
{
    ## convert contour polygons to multipolygon geometry
    if (missing(abs.cont)) cont.geom <- dplyr::group_modify(.data=x, .f=~dplyr::tibble(.st_contourline(untidy_ks(.x), cont=cont, edge_zero=edge_zero, digits=digits))) 
    else if (!missing(abs.cont)) cont.geom <- dplyr::group_modify(.data=x, .f=~dplyr::tibble(.st_contourline(untidy_ks(.x), abs.cont=abs.cont, edge_zero=edge_zero, digits=digits))) 
    
    return(cont.geom)
}

## create set of disjoint contour polygons
## x = multipolygon geometry of contour regions
st_contour_disjoint <- function(x)
{
    gv <- dplyr::group_vars(x)
    xc <- dplyr::group_modify(.data=x, .f=~dplyr::tibble(.st_contour_disjoint(.x))) 
    xc <- sf::st_sf(xc)
    xc <- sf::st_cast(xc, to = "MULTIPOLYGON")
    xc <- dplyr::ungroup(xc)
    if (length(gv)>0) xc <- dplyr::group_by(xc, dplyr::across({{gv}}))
    xc$.contlabel <- NULL

    return(xc)
}

.st_contour_disjoint <- function(x)
{
    if (any(as.numeric(as.character(x$contlabel))<0)) oc <- "kdde" else oc <- "kde"
    if (any(oc %in% "kdde"))
    {
        cp <- x
        cp <- dplyr::mutate(cp, estimate=as.numeric(as.character(.data$estimate)), contlabel=as.numeric(as.character(.data$contlabel)))
        if (!any(names(cp) %in% "deriv_ind")) cp <- dplyr::arrange(cp, .data$estimate) 
        else cp <- dplyr::arrange(cp, .data$deriv_ind, .data$estimate)

        cp1 <- dplyr::filter(cp, .data$contlabel<=0)
        cp1 <- sf::st_geometry(cp1)
        cp1 <- st_difference_sequence(cp1, headtail=FALSE)
        cp1 <- sf::st_sfc(cp1)
        cp1 <- c(head(sf::st_geometry(dplyr::filter(cp, .data$contlabel<=0)),n=1),cp1)
        cp1 <- sf::st_set_crs(cp1, sf::st_crs(x))
        
        cp2 <- dplyr::filter(cp, .data$contlabel>0)
        cp2 <- sf::st_geometry(cp2)
        cp2 <- st_difference_sequence(cp2, headtail=TRUE)
        cp2 <- sf::st_sfc(cp2)
        cp2 <- c(cp2, tail(sf::st_geometry(dplyr::filter(cp, .data$contlabel>0)),n=1))
        cp2 <- sf::st_set_crs(cp2, sf::st_crs(x))
        x <- sf::st_set_geometry(x, value=c(cp1,cp2))
    }
    else
    { 
        cp <- sf::st_geometry(x)
        cp <- st_difference_sequence(cp, headtail=TRUE)
        cp <- sf::st_sfc(cp)
        cp <- c(cp, tail(sf::st_geometry(x),n=1))
        cp <- sf::st_set_crs(cp, sf::st_crs(x))
        x <- sf::st_set_geometry(x, value=cp)

    }
    x <- sf::st_make_valid(x, geos_method="valid_linework") 
    
    return(x)
}

## compute st_difference on pairs of adjacent rows in sf object
## x = sf object
st_difference_sequence <- function(x, headtail=TRUE) 
{ 
    if (length(x)>1)
    {
        for (i in 1:(length(x)-1)) 
        {
            xh <- head(x, n=-1)
            xt <- tail(x, n=-1)
            if (headtail) 
            {
                xpt <- sf::st_difference(xh[i], xt[i])
                if (length(xpt)==0) xpt <- sf::st_sfc(sf::st_polygon(), crs=sf::st_crs(x))
                if (i==1) xp <- xpt else xp <- c(xp, xpt)
            }
            else
            {
                xpt <- sf::st_difference(xt[i], xh[i]) 
                if (length(xpt)==0) xpt <- sf::st_sfc(sf::st_polygon(), crs=sf::st_crs(x))
                if (i==1) xp <- xpt else xp <- c(xp, xpt)
            }
        }
    }
    else xp <- x[0,]
    
    return(xp)
}

## x = tibble with ks column (output from first part of st_kdde)
st_contourline_kdde <- function(x, cont, abs.cont, which_deriv_ind=1, digits)
{
    x.temp <- x
    if (missing(abs.cont)) 
    { 
        abs.cont <- contourLevels(x, cont=cont, which.deriv.ind=which_deriv_ind)
        abs.cont.flag <- FALSE
        abs.cont <- sort(c(abs.cont[1,], abs.cont[2,]))
    }
    else 
    {
        if (!is.vector(abs.cont) & !is.null(abs.cont$breaks)) abs.cont <- abs.cont$breaks
        abs.cont <- sort(abs.cont)
        abs.cont.flag <- TRUE
    }
    abs.cont <- abs.cont[!duplicated(abs.cont)]
    x.temp$estimate <- x$estimate[[which_deriv_ind]]

    cont.geom <- .st_contourline(x=x.temp, abs.cont=abs.cont, digits=digits) 

    if (!abs.cont.flag) 
    {
        cont.geom <- dplyr::mutate(cont.geom, contlabel=ifelse(.data$estimate<0, paste0("-", as.character(.data$contlabel)), as.character(.data$contlabel)))
        cont.geom$contlabel <- factor(cont.geom$contlabel, levels=c(-sort(cont), sort(cont, decreasing=TRUE)))
    }
    else cont.geom$contlabel <- factor(cont.geom$contlabel)
    
    return(cont.geom)
}

## x = tibble with ks column (output from first part of st_kda)
st_contourline_kda <- function(x, cont, digits)
{
    xj <- untidy_ks(x)
    ## replace KDE with weighted KDE
    j <- x$rn
    pp <- xj$prior.prob[j]
    abs.cont <- contourLevels(xj, cont=cont)[[j]]*pp
    xj$estimate <- xj$estimate[[j]]*pp
    xj$H <- xj$H[[j]]
    xj$cont <- xj$cont[[j]]
    class(xj) <- "kde"
    
    cont.geom <- .st_contourline(x=xj, abs.cont=abs.cont, digits=digits) 
    cont.geom <- dplyr::mutate(cont.geom, contlabel=factor(ifelse(.data$estimate<0, paste0("-", 100-as.numeric(as.character(.data$contlabel))), as.character(.data$contlabel))))
    cont.geom <- dplyr::arrange(cont.geom, sign(.data$estimate), abs(.data$estimate))
    levels(cont.geom) <- as.character(cont.geom$contlabel)
       
    return(cont.geom)
}

## convert eval.points in ks object to rectangle polygon/point geometry
## x = tibble with ks column (output from first part of st_ks)
st_evalpoints <- function(x)
{
    ## convert ks grid to rectangle polygon geometry
    if (inherits(untidy_ks(x[1,]), "kde.loctest"))
    {
        fhat.geom.neg <- dplyr::group_modify(.data=x, .f=~dplyr::tibble(.st_evalpoints(untidy_ks(.x)$fhat.diff.neg)))
        fhat.geom.pos <- dplyr::group_modify(.data=x, .f=~dplyr::tibble(.st_evalpoints(untidy_ks(.x)$fhat.diff.pos)))
        fhat.geom.neg <- dplyr::mutate(fhat.geom.neg, estimate=as.integer(.data$estimate>=0.5))
        fhat.geom.pos <- dplyr::mutate(fhat.geom.pos, estimate=as.integer(.data$estimate>=0.5))
        fhat.geom <- fhat.geom.pos
        fhat.geom$estimate <- fhat.geom.pos$estimate - fhat.geom.neg$estimate
    }
    else 
        fhat.geom <- dplyr::group_modify(.data=x, .f=~dplyr::tibble(.st_evalpoints(untidy_ks(.x))))
    fhat.geom <- sf::st_as_sf(fhat.geom, crs=sf::st_crs(x))
       
    return(fhat.geom)
}

## convert eval.points in ks object to rectangle polygon geometry
## x = ks object
## ouput sf-geometry column is named "geometry"
.st_evalpoints <- function(x)
{
    ## convert KDE to rectangle polygon geometry
    nx <- lengths(x$eval.points)[1]
    ny <- lengths(x$eval.points)[2]
    ep1 <- x$eval.points[[1]]
    ep2 <- x$eval.points[[2]]
    fhat.point <- sf::st_as_sf(x=expand.grid(x=ep1[c(1,nx)], y=ep2[c(1,ny)]), coords=1:2)
    fhat.geom <- sf::st_make_grid(fhat.point, n=c(nx-1,ny-1))
    fhat.geom.cent <- sf::st_centroid(sf::st_geometry(fhat.geom))
    fhat.geom.cent <- sf::st_coordinates(fhat.geom.cent)
    ## "geometry" is automatically given sf-geometry column when concatenating into data frame
    fhat.geom <- data.frame(estimate=predict(x, x=fhat.geom.cent), fhat.geom)
  
    return(fhat.geom)
}

## remove segments in linestring where distance between two consecutive points > len 
## (metres) useful to neaten appearance of polygons which are split to separate edges of 
## world map after projection  
## x = linestring or multipoint geometry
st_remove_long_segment <- function(x, len=500e3)
{
    x.geom <- dplyr::group_modify(x, .f=~dplyr::as_tibble(sf::st_coordinates(sf::st_geometry(.))))
    gvar <- setdiff(names(x.geom), c("X","Y")) 
    x.geom <- dplyr::group_by(x.geom, dplyr::across({{gvar}}))
    x.geom <- dplyr::mutate(x.geom, L1=cumsum(c(0, sqrt(diff(.data$X)^2 + diff(.data$Y)^2) > len)))
    x.geom <- dplyr::group_by(x.geom, dplyr::across({{gvar}}))
    valid.LS <- names(table(x.geom$L1))[table(x.geom$L1)>2]
    
    if (length(valid.LS)>0) 
    {
        x.geom <- dplyr::filter(x.geom, .data$L1 %in% valid.LS)
        x.geom <- dplyr::summarise(sf::st_as_sf(x.geom, coords=c("X","Y"), crs=sf::st_crs(x)), do_union=FALSE, .groups="drop")
        x.geom <- sf::st_cast(x.geom, to="LINESTRING")
        x.geom <- dplyr::filter(x.geom, as.numeric(sf::st_length(.data$geometry))>0)
        x.geom <- dplyr::group_by(x.geom, dplyr::across({{gvar}}))
        x.geom <- dplyr::summarise(sf::st_sf(x.geom), do_union=FALSE, .groups="drop")
        sfname <- attr(x.geom, "sf_column")
    
        x2 <- dplyr::right_join(sf::st_drop_geometry(x), x.geom, by=dplyr::group_vars(x))
        x2 <- dplyr::relocate(x2, !!sfname, .after=dplyr::last_col())
        x <- sf::st_sf(x2)
    }
    
    return(x)
}

## concatenate sf objects
c_sf <- function(..., labels)
{
    args <- list(...)
    nargs <- length(args)
    if (nargs>=1) 
    {
        sfname <- attr(args[[1]], "sf_column")
        tt <- dplyr::mutate(args[[1]], group=1L)
        if (nargs>=2)
        {
            for (i in 2:nargs) 
                tt <- dplyr::bind_rows(tt, dplyr::mutate(args[[i]], group=i))
        }
        if (!missing(labels)) tt$group <- factor(tt$group, labels=labels)
        tt <- dplyr::relocate(tt, !!sfname, .after=dplyr::last_col())
        tt <- sf::st_sf(tt)
    }
    else tt <- NULL

    return(tt)
}

## concatenate sf_ks objects
#c.sf_ks <- function(..., labels) { c_sf_ks(..., labels=labels) }

c.sf_ks <- function(..., labels)
{
    args <- list(...)
    nargs <- length(args)
    if (nargs>=1) 
    {
        csf <- list()
        csf$tidy_ks <- do.call(c.tidy_ks, lapply(args, getElement, "tidy_ks"))
        csf$grid <- do.call(c_sf, lapply(args, getElement, "grid"))
        csf$sf <- do.call(c_sf, lapply(args, getElement, "sf"))

        if (!missing(labels)) 
        {
            csf$tidy_ks$group <- factor(csf$tidy_ks$group, labels=labels)
            csf$grid$group <- factor(csf$grid$group, labels=labels)
            csf$sf$group <- factor(csf$sf$group, labels=labels)
        }

        csf$tidy_ks <- dplyr::group_by(csf$tidy_ks, .data$group)
        csf$grid <- dplyr::group_by(csf$grid, .data$group)
        csf$sf <- dplyr::group_by(csf$sf, .data$group)
        csf$tidy_ks <- as_tidy_ks(csf$tidy_ks)
    }
    else csf <- NULL
    class(csf) <- "sf_ks"

    return(csf)
}

rbind.sf_ks <- c.sf_ks

## add coordinates as attributes to point geometry
st_add_coordinates <- function(x, as_sf=TRUE, rename=TRUE)
{
    sfname <- attr(x, "sf_column")
    xc <- sf::st_cast(x, to="POINT", warn=FALSE)
    xcoord <- sf::st_coordinates(sf::st_geometry(xc))
    xc$X <- xcoord[,1]
    xc$Y <- xcoord[,2]
    
    if (!as_sf)  xc <- sf::st_drop_geometry(xc)
    else xc <- dplyr::relocate(xc, dplyr::all_of(c("X", "Y")), .before=!!sfname)
    if (inherits(x, "tbl")) xc <- dplyr::as_tibble(xc)
    if (as_sf)  xc <- sf::st_sf(xc)
    if (rename) xc <- dplyr::rename(xc, lon=dplyr::all_of("X"), lat=dplyr::all_of("Y"))

    return(xc)
}

## returns segments of bounding box of x
## h1 = (xmax, ymin) -> (xmin, ymin)
## v1 = (xmin, ymin) -> (xmin, ymax)
## h2 = (xmin, ymax) -> (xmax, ymax)
## v2 = (xmax, ymax) -> (xmax, ymin)
st_bbox_segments <- function(x)
{
    xbbox <- sf::st_bbox(x)
    xcrs <- sf::st_crs(x)

    ## vertical and horizontal lines of bounding box
    h1 <- sf::st_sfc(sf::st_linestring(rbind(xbbox[c(3,2)], xbbox[c(1,2)])), crs=xcrs)
    v1 <- sf::st_sfc(sf::st_linestring(rbind(xbbox[c(1,2)], xbbox[c(1,4)])), crs=xcrs)
    h2 <- sf::st_sfc(sf::st_linestring(rbind(xbbox[c(1,4)], xbbox[c(3,4)])), crs=xcrs)
    v2 <- sf::st_sfc(sf::st_linestring(rbind(xbbox[c(3,4)], xbbox[c(3,2)])), crs=xcrs)

    return(c(h1, v1, h2, v2))
}

## add "%" suffix - deprecated but kept for backwards compatibility 
label_percent <- function(y)
{
    factor(y, labels=paste0(levels(y),"%"))
}

## read GPKG and renames "geom" as "geometry"
st_read_gpkg <- function(..., as_tibble=TRUE)
{
    sf::st_set_geometry(sf::st_read(..., as_tibble=as_tibble), "geometry")
}

## convenience function to write to progress geopackage for diagnostics
write_temp <- function(x, layer, dsn) 
{ 
    if (!inherits(x, "sf")) x <- sf::st_sf(x)
    if (missing(layer))
    {
        if (any(sf::st_is(x, "POINT"))) layer <- "temp3"
        else if (any(sf::st_is(x, "POLYGON"))) layer <- "temp4"
        else layer <- "temp"
    }
    epsg <- sf::st_crs(x)$epsg
    if (missing(dsn)) 
    { 
        if (is.na(epsg)) dsn <- "~/Downloads/flowmap_progress.gpkg"
        else dsn <- paste0("~/Downloads/flowmap_progress_", epsg,".gpkg")
    }
    x <- dplyr::select(x, !dplyr::where(is.list))

    sf::write_sf(x, dsn=dsn, layer=layer)
}

st_diameter <- function(x)
{
    xch <- sf::st_convex_hull(x)
    xch <- sf::st_geometry(xch)
    diam <- sapply(1:length(xch), function (.) 
    {
        y <- sf::st_cast(sf::st_cast(xch[.], to="MULTIPOINT", warn=FALSE), to="POINT", warn=FALSE)
        diam <- max(as.numeric(sf::st_distance(y)))
    })
    
    return(diam)
}

## more efficient centroid than sf::st_centroid?
st_centroid2 <- function(x)
{
    #sf::st_geometry(x) <- sf::st_as_sfc(geos::geos_centroid(x))
    sf::st_geometry(x) <- sf::st_geometry(sf::st_centroid(sf::st_geometry(x)))
    return(x)
}

---
title: "Kernel density estimates for tidy and geospatial data in the eks package"
output: rmarkdown::html_vignette
description: R package for feature significance for multivariate kernel density estimation
date: "`r format(Sys.time(), '%d %B %Y')`"
author: Tarn Duong https://mvstat.net/tduong/
vignette: >
  %\VignetteIndexEntry{Kernel density estimates for tidy and geospatial data in the eks package}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r, echo=FALSE, message=FALSE}
knitr::opts_chunk$set(global.par=TRUE, collapse=TRUE, comment="#>", fig.width=5, fig.height=5, fig.align="center", dpi=96)
options(tibble.print_min=3L, tibble.print_max=3L)
```

## Introduction 
Kernel smoothers are essential tools for data analysis due to their ability to convey complex statistical information with concise graphical visualisations. The most widely used kernel smoother is the kernel density estimator (KDE), though there remain some important gaps in the implementation in ```R``` for specialised data types, most notably for tibbles (tidy data) within the tidyverse, and for simple features (geospatial data) within Geographical Information Systems (GIS) analysis. The `tidy_kde` and `st_kde` functions in the `eks` package fills in these gaps.

## Tidy data example 
The data set we focus on is the ```crabs``` data set from the ```MASS``` package, with the variables ```FL``` frontal lobe size (mm), ```CW``` carapace width (mm) and ```sp``` species (```B``` for blue, ```O``` for orange). 
```{r, message=FALSE}
library(eks)
library(colorspace)
library(ggplot2)
library(dplyr)

## crabs data 
data(crabs, package="MASS")
crabs2 <- select(crabs, FL, CW)
xlab <- "Frontal lobe size (mm)"
ylab <- "Carapace width (mm)"
```

The KDE for tidy data is computed by ```tidy_kde```. From the output, the scatter plot of the data is generated by ```geom_point_ks``` and the contour plot of the KDE by ```geom_contour_ks```. The bimodal structure of the data  distribution, corresponding to the two species, is clearly visible from the KDE plot from ```tidy_kde```. This is due to the optimal choice of the matrix of smoothing parameters. This optimal smoothing matrix is the plug-in bandwidth computed by ```ks::Hpi```, and it is suitable for a wide range of data sets. For further details of the computation of the kernel density estimate and the bandwidth, see ```?ks::kde``` and ```?ks::Hpi```.

```{r}
## KDE contour plot + scatter plot
tkde2 <- tidy_kde(crabs2)
gkde2 <- ggplot(tkde2, aes(x=FL, y=CW)) + labs(x=xlab, y=ylab)
gkde2 + geom_point_ks(colour=8) + geom_contour_ks(colour=1)
```

On the other hand, the default bandwidth and the resulting KDE computed by ```ggplot2::geom_density_2d``` leads to an oversmoothed KDE which does not reveal the data bimodality. 
```{r}
## geom_density_2d KDE contour plot + scatter plot
mkde2 <- ggplot(crabs2, aes(x=FL, y=CW))
mkde2 + geom_point(colour=8) + geom_density_2d(colour=1, bins=4)
```

The default choice of the contour levels in the ```eks``` package is based on probability contours. Probability contours offer an intuitive approach to selecting the contour levels that reveal the pertinent characteristics of the data distribution. See Chacon & Duong (2018, Chapter 2.2). Filled contour plots, generated by ```geom_contour_filled_ks```, can be coloured with an appropriate sequential colour scale. For example, a 25% contour region (red region) is the smallest region that contains 25% of the probability mass defined by the KDE. The 50% contour region consists of the union of the orange region and the red region, and it contains 50% of the data points etc.  Note that the 25%, 50% and 75% contour regions of the ```crabs``` KDE are composed of separate, unconnected contour sub-regions. 

```{r}
## KDE filled contour plot
gkde2 + geom_contour_filled_ks(colour=1) 
```

As an alternative to these discretised contours, the usual ```ggplot2::geom_raster``` generates a plot with a continuous colour scale.

```{r}
## KDE continuous colour scale plot
gkde2 + geom_raster(aes(fill=estimate), interpolate=TRUE) + 
   scale_fill_continuous_sequential(palette="Heat")
```

One of the main advantages of ```ggplot2``` is its ability to handle grouped data: in this case, the KDE plots for each species. The KDE with blue contours is for the ```B``` species, and orange contours for the ```O``` species.  
```{r, fig.width=7}
crabs2g <- select(crabs, FL, CW, sp)
crabs2g <- group_by(crabs2g, sp)
tkde2g <- tidy_kde(crabs2g)
gkde2g <- ggplot(tkde2g, aes(x=FL, y=CW, group=sp)) + labs(x=xlab, y=ylab, colour="Species") + 
   scale_colour_manual(values=c(4, 7)) + 
   guides(colour=guide_legend(title="Species"))

## facetted KDE contour plots + scatter plots
gkde2g + geom_point_ks(colour=8) + 
   geom_contour_ks(aes(colour=sp)) + facet_wrap(~sp) 
```

The probability contour levels computed in ```geom_contour_ks``` and ```geom_contour_filled_ks``` are relative to the grouping variable. So whilst the same probability 25% level is applied to both groups KDE, the height of 25% contour region for the blue species is 0.04143, and for the orange species it is 0.03171. 
For a direct comparison of different KDEs, it is useful to have a set of fixed contour heights for all KDEs . A heuristic solution is implemented in ```contour_breaks```. For the ```crabs``` data, this gives 0.01418, 0.02690, 0.04143. Since the KDE for the ```B``` species exceeds the highest level 0.04143, whereas the ```O``` species KDE doesn't reach this is level, the former KDE is more peaked.  

```{r, fig.width=7}
## facetted KDE filled contour plots with fixed contour levels for all facets
bkde2g <- contour_breaks(tkde2g)
gkde2g + geom_contour_filled_ks(breaks=bkde2g, colour=1) + facet_wrap(~sp)
```

## Geospatial data example
GIS for geospatial data analysis in ```R``` is implemented in the ```sf``` package, and the ```eks``` package builds on this. To illustrate geospatial KDE, we focus on the ```grevilleasf``` data set in the ```eks``` package.  It has 22303 rows, where each row corresponds to an observed grevillea plant in Western Australia. In addition, we utilise ```wa```, the geospatial polygon for Western Australia. Both of these geospatial data sets are in the EPSG:7850 (GDA2020/MGA zone 50) projection. 

```{r, message=FALSE}
library(sf)

## Grevillea data
data(grevilleasf, package="eks")
grevilleasf <- mutate(grevilleasf, species=factor(species))
paradoxa <- filter(grevilleasf, name %in% "Grevillea paradoxa")
eryngioides <- filter(grevilleasf, name %in% "Grevillea eryngioides")
grevillea_ep <- filter(grevilleasf, name %in% c("Grevillea eryngioides", 
   "Grevillea paradoxa"))
grevillea_ep <- group_by(grevillea_ep, name)
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)

## WA polygon
data(wa, package="eks")
gwa <- geom_sf(data=wa, fill=NA, colour=1)
```

Since geospatial data can be visualised with both base ```R``` and ```ggplot2``` graphics engines, we provide code for both: their outputs are similar due to the standardisation of geospatial maps within GIS. Though these plots can't be mixed due to fundamental differences between the graphical rendering in base ```R``` and ```ggplot2```.     

```{r}
## base R scatter plot
plot(st_geometry(wa), xlim=xlim, ylim=ylim)
plot(st_geometry(eryngioides), add=TRUE, col=3, pch=16, cex=0.5)
plot(st_geometry(paradoxa), add=TRUE, col=6, pch=17, cex=0.5)
mapsf::mf_legend(type="symb", val=c("Grevillea eryngioides", "Grevillea paradoxa"), 
   pal=c(3,6), pch=16:17, cex=c(1,1), title="Species", pos="bottomleft")
## geom_sf scatter plot 
theme_set(ggthemes::theme_map())
ggplot() + gwa + 
   geom_sf(data=grevillea_ep, aes(colour=name, shape=name)) + 
   coord_sf(xlim=xlim, ylim=ylim) + scale_colour_manual(values=c(3, 6)) + 
   guides(colour=guide_legend(title="Species"), shape=guide_legend(title="Species"))
```

The KDE for geospatial data is computed by ```st_kde```. The calculations of the KDE, including the bandwidth matrix fo smoothing parameters, are the same as in ```tidy_kde```. Though, unlike for ```tidy_kde``` where the probability contour regions are computed dynamically in ```geom_contour_filled_ks```, the 5% to 95% regions are explicitly computed as multipolygons in ```st_kde``` since this conversion can be computationally heavy to execute for each plot. For display, it is a matter of selecting the appropriate contour regions. The quartile contours 25%, 50%, 75% are selected by default in ```geom_contour_filled_ks``` for tidy data. This is also the case for the base ```R``` plot. 

```{r}
skde1 <- st_kde(paradoxa)

## base R contour plot
plot(st_geometry(wa), xlim=xlim, ylim=ylim)
plot(st_geometry(paradoxa), add=TRUE, pch=16, col=8, cex=0.5)
plot(skde1, add=TRUE, col=NA, border=1, legend=FALSE)
```
On the other hand, we can't replicate exactly the default contour selection in ```geom_sf```, so we first apply the auxiliary function ```st_get_contour``` to the input of ```geom_sf```.  

```{r}
## geom_sf contour plot
gs <- ggplot(skde1) + gwa + ggthemes::theme_map()
gs + geom_sf(data=paradoxa, col=8, size=0.5) + 
   geom_sf(data=st_get_contour(skde1), colour=1, fill=NA) + 
   coord_sf(xlim=xlim, ylim=ylim)
```

To generate a filled contour plot, the only required change is to set the fill aesthetic to ```contperc``` (to produce legend labels with density contour percentages) or ```estimate``` (labels with density estimate levels).   

```{r}
## R base filled contour plot
plot(st_geometry(wa), xlim=xlim, ylim=ylim)
plot(skde1, add=TRUE)
## geom_sf filled contour
gs + geom_sf(data=st_get_contour(skde1), aes(fill=contperc)) + 
  coord_sf(xlim=xlim, ylim=ylim) 
```

Since the output from ```st_kde``` is compatible with ```geom_sf```, then it is easy to generate multiple maps of related geospatial KDEs. For example, KDEs for each Grevillea species, with probability contour levels or with fixed contour levels:

```{r, fig.width=7}
skde1g <- st_kde(grevillea_ep)

## facetted geom_sf filled contour
gs + geom_sf(data=st_get_contour(skde1g), aes(fill=contperc)) + 
   facet_wrap(~name) + coord_sf(xlim=xlim, ylim=ylim) 
## facetted geom_sf filled contour with fixed contour levels for all facets
bkde1g <- contour_breaks(skde1g, group=TRUE)
gs + geom_sf(data=st_get_contour(skde1g, breaks=bkde1g), aes(fill=estimate)) +
   facet_wrap(~name) + coord_sf(xlim=xlim, ylim=ylim) 
```

## References

Chacon, J. E. and Duong, T. (2018). _Multivariate Kernel Smoothing and Its Applications._ Chapman & Hall/CRC Press, Boca Raton.

Duong, T. (2024) Statistical visualisation for tidy and geospatial data in R via kernel smoothing methods in the eks package. _Computational Statistics_ https://doi.org/10.1007/s00180-024-01543-9

---
title: "Kernel density estimates for tidy and geospatial data in the eks package"
output: rmarkdown::html_vignette
description: Kernel density estimates for tidy and geospatial data in the eks package
author: Tarn Duong
vignette: >
  %\VignetteIndexEntry{Kernel density estimates for tidy and geospatial data in the eks package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo=FALSE, message=FALSE}
knitr::opts_chunk$set(global.par=TRUE, collapse=TRUE, comment="#>", fig.width=5, fig.height=5, fig.align="center", dpi=96)
options(tibble.print_min=3L, tibble.print_max=3L)
```

## Introduction 
Kernel smoothers are essential tools for data analysis due to their ability to convey complex statistical information with concise graphical visualisations. The most widely used kernel smoother is the kernel density estimator (KDE), though there remain some important gaps in the implementation in `R` for specialised data types, most notably for tibbles (tidy data) within the tidyverse, and for simple features (geospatial data) within Geographical Information Systems (GIS) analysis. The `tidy_kde()` and `st_kde()` functions in the `eks` package fills in these gaps.

## Tidy data example 
The data set we focus on is the `MASS::crabs` data set, with the variables `FL` frontal lobe size (mm), `CW` carapace width (mm) and `sp` species (`B` for blue, `O` for orange). 
```{r, message=FALSE}
library(eks)
library(colorspace)
library(ggplot2)
library(dplyr)
set_theme(theme_bw())

## crabs data 
data(crabs, package="MASS")
crabs2 <- select(crabs, FL, CW)
xlab <- "Frontal lobe size (mm)"
ylab <- "Carapace width (mm)"
```

The KDE for tidy data is computed by `tidy_kde()`. From the output, the scatter plot of the data is generated by `geom_point_ks()` and the contour plot of the KDE by `geom_contour_ks()`. The bimodal structure of the data  distribution, corresponding to the two species, is clearly visible from the KDE plot from `tidy_kde`. This is due to the optimal choice of the matrix of smoothing parameters. This optimal smoothing matrix is the plug-in bandwidth computed by `ks::Hpi`, and it is suitable for a wide range of data sets. For further details of the computation of the kernel density estimate and the bandwidth, see `?ks::kde` and `?ks::Hpi`.

```{r, message=FALSE}
## KDE contour plot + scatter plot
tkde2 <- tidy_kde(crabs2)
gkde2 <- ggplot(tkde2, aes(x=FL, y=CW)) + labs(x=xlab, y=ylab)
gkde2 + geom_point_ks(colour=8) + geom_contour_ks(colour=1)
```

On the other hand, the default bandwidth and the resulting KDE computed by `ggplot2::geom_density_2d` leads to an oversmoothed KDE which does not reveal the data bimodality. 
```{r, message=FALSE}
## geom_density_2d KDE contour plot + scatter plot
mkde2 <- ggplot(crabs2, aes(x=FL, y=CW))
mkde2 + geom_point(colour=8) + geom_density_2d(colour=1, bins=4)
```

The default choice of the contour levels in the `eks` package is based on probability contours. Probability contours offer an intuitive approach to selecting the contour levels that reveal the pertinent characteristics of the data distribution. See Chacon & Duong (2018, Chapter 2.2). Filled contour plots, generated by `geom_contour_filled_ks()`, can be coloured with an appropriate sequential colour scale. For example, a 25% contour region (red region) is the smallest region that contains 25% of the probability mass defined by the KDE. The 50% contour region consists of the union of the orange region and the red region, and it contains 50% of the data points etc.  Note that the 25%, 50% and 75% contour regions of the `crabs` KDE are composed of separate, unconnected contour sub-regions. 

```{r, message=FALSE}
## KDE filled contour plot
gkde2 + geom_contour_filled_ks(colour=1) 
```

As an alternative to these discretised contours, the usual `ggplot2::geom_raster` generates a plot with a continuous colour scale.

```{r, message=FALSE}
## KDE continuous colour scale plot
gkde2 + geom_raster(aes(fill=estimate), interpolate=TRUE) + 
   scale_fill_continuous_sequential(palette="Heat")
```

One of the main advantages of `ggplot2` is its ability to handle grouped data: in this case, the KDE plots for each species. The KDE with blue contours is for the `B` species, and orange contours for the `O` species.  
```{r, message=FALSE, fig.width=7}
crabs2g <- select(crabs, FL, CW, sp)
crabs2g <- group_by(crabs2g, sp)
tkde2g <- tidy_kde(crabs2g)
gkde2g <- ggplot(tkde2g, aes(x=FL, y=CW, group=sp)) 

## facetted KDE contour plots + scatter plots
gkde2g + geom_point_ks(colour=8) + labs(x=xlab, y=ylab, colour="Species") + 
   scale_colour_manual(values=c(4, 7)) + 
   guides(colour=guide_legend(title="Species")) +
   geom_contour_ks(aes(colour=sp)) + facet_wrap(~sp) 
```

The probability contour levels computed in `geom_contour_ks` and `geom_contour_filled_ks` are relative to the grouping variable. So whilst the same probability 25% level is applied to both groups KDE, the height of 25% contour region for the blue species is 0.04143, and for the orange species it is 0.03171. 
For a direct comparison of different KDEs, it is useful to have a set of fixed contour heights for all KDEs . A heuristic solution is implemented in `contour_breaks()`. For the `crabs` data, this gives 0.01418, 0.02690, 0.04143. Since the KDE for the `B` species exceeds the highest level 0.04143, whereas the `O` species KDE doesn't reach this is level, the former KDE is more peaked.  

```{r, fig.width=7}
## facetted KDE filled contour plots with fixed contour levels for all facets
bkde2g <- contour_breaks(tkde2g)
gkde2g + geom_contour_filled_ks(breaks=bkde2g, colour=1) + facet_wrap(~sp)
```

## Geospatial data example
GIS for geospatial data analysis in `R` is implemented in the `sf` package, and the `eks` package builds on this. To illustrate geospatial KDE, we focus on the `grevilleasf` data set in the `eks` package.  It has 22303 rows, where each row corresponds to an observed grevillea plant in Western Australia. In addition, we utilise `wa`, the geospatial polygon for Western Australia. Both of these geospatial data sets are in the EPSG:7850 (GDA2020/MGA zone 50) projection. 

```{r, message=FALSE}
library(sf)

## Grevillea data
data(grevilleasf, package="eks")
grevilleasf <- mutate(grevilleasf, species=factor(species))
paradoxa <- filter(grevilleasf, name %in% "Grevillea paradoxa")
eryngioides <- filter(grevilleasf, name %in% "Grevillea eryngioides")
grevillea_ep <- filter(grevilleasf, name %in% c("Grevillea eryngioides", 
   "Grevillea paradoxa"))
grevillea_ep <- group_by(grevillea_ep, name)
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)

## WA polygon
data(wa, package="eks")
gwa <- geom_sf(data=wa, fill=NA, colour=1)
```

Since geospatial data can be visualised with both base `R` and `ggplot2` graphics engines, we provide code for both: their outputs are similar due to the standardisation of geospatial maps within GIS. Though these plots can't be mixed due to fundamental differences between the graphical rendering in base `R` and `ggplot2`.     

```{r}
## base R scatter plot
plot(st_geometry(wa), xlim=xlim, ylim=ylim)
plot(st_geometry(eryngioides), add=TRUE, col=3, pch=16, cex=0.5)
plot(st_geometry(paradoxa), add=TRUE, col=6, pch=17, cex=0.5)
mapsf::mf_legend(type="symb", val=c("Grevillea eryngioides", "Grevillea paradoxa"), 
   pal=c(3,6), pch=16:17, cex=c(1,1), title="Species", pos="bottomleft")
## geom_sf scatter plot 
ggplot() + gwa + theme_sf() +
   geom_sf(data=grevillea_ep, aes(colour=name, shape=name)) + 
   coord_sf(xlim=xlim, ylim=ylim) + scale_colour_manual(values=c(3, 6)) + 
   guides(colour=guide_legend(title="Species"), shape=guide_legend(title="Species"))
```

The KDE for geospatial data is computed by `st_kde()`. The calculations of the KDE, including the bandwidth matrix fo smoothing parameters, are the same as in `tidy_kde()`. Though, unlike for `tidy_kde` where the probability contour regions are computed dynamically in `geom_contour_filled_ks()`, the 5% to 95% regions are explicitly computed as multipolygons in `st_kde` since this conversion can be computationally heavy to execute for each plot. For display, it is a matter of selecting the appropriate contour regions. The quartile contours 25%, 50%, 75% are selected by default in `geom_contour_filled_ks` for tidy data. This is also the case for the base `R` plot. 

```{r}
skde1 <- st_kde(paradoxa)

## base R contour plot
plot(st_geometry(wa), xlim=xlim, ylim=ylim)
plot(st_geometry(paradoxa), add=TRUE, pch=16, col=8, cex=0.5)
plot(skde1, add=TRUE, col=NA, border=1, legend=FALSE)
```
On the other hand, we can't replicate exactly the default contour selection in `geom_sf`, so we first apply the auxiliary function `st_get_contour()` to the input of `geom_sf`.  

```{r}
## geom_sf contour plot
gs <- ggplot(skde1) + gwa + theme_sf()
gs + geom_sf(data=paradoxa, col=8, size=0.5) + 
   geom_sf(data=st_get_contour(skde1), colour=1, fill=NA) + 
   coord_sf(xlim=xlim, ylim=ylim)
```

To generate a filled contour plot, the only required change is to set the fill aesthetic to `contperc` (to produce legend labels with density contour percentages) or `estimate` (labels with density estimate levels).   

```{r}
## R base filled contour plot
plot(st_geometry(wa), xlim=xlim, ylim=ylim)
plot(skde1, add=TRUE)
## geom_sf filled contour
gs + geom_sf(data=st_get_contour(skde1), aes(fill=contperc)) + 
  coord_sf(xlim=xlim, ylim=ylim) 
```

Since the output from `st_kde` is compatible with `geom_sf`, then it is easy to generate multiple maps of related geospatial KDEs. For example, KDEs for each Grevillea species, with probability contour levels or with fixed contour levels:

```{r, fig.width=7}
skde1g <- st_kde(grevillea_ep)

## facetted geom_sf filled contour
gs + geom_sf(data=st_get_contour(skde1g), aes(fill=contperc)) + 
   facet_wrap(~name) + coord_sf(xlim=xlim, ylim=ylim) 
## facetted geom_sf filled contour with fixed contour levels for all facets
bkde1g <- contour_breaks(skde1g, group=TRUE)
gs + geom_sf(data=st_get_contour(skde1g, breaks=bkde1g), aes(fill=estimate)) +
   facet_wrap(~name) + coord_sf(xlim=xlim, ylim=ylim) 
```

## References
Chacon, J. E. and Duong, T. (2018) [_Multivariate Kernel Smoothing and Its Applications_](https://mvstat.net/mvksa/) Chapman & Hall/CRC Press, Boca Raton.

Duong, T. (2025) [Statistical visualisation for tidy and geospatial data in R via kernel smoothing methods in the eks package](https://mvstat.net/tduong/research/publications/duong-2025-compstat.pdf) _Computational Statistics_ **40**, 2825--2847.


&nbsp;

-- Generated `r format(Sys.time(), '%d %B %Y')`. 

\name{contour}
\alias{add_contour_breaks}
\alias{contour_breaks}
\alias{contourLevels.tidy_ks}
\alias{contourLevels.sf_ks}
\alias{st_get_contour}
\title{Contour functions for tidy and geospatial kernel estimates}

\description{
    Contour functions for tidy and geospatial kernel estimates.
}

\usage{
\method{contourLevels}{tidy_ks}(x, ..., cont=c(25,50,75))
\method{contourLevels}{sf_ks}(x, ..., cont=c(25,50,75))
contour_breaks(data, cont=c(25,50,75), n=3, group=FALSE, type="density")
st_get_contour(x, cont=c(25,50,75), breaks, disjoint=TRUE, digits)
add_contour_breaks(x, breaks, digits, ...)
}

\arguments{
    \item{x,data}{tidy kernel estimate (output from \code{tidy_k*}) or geospatial kernel estimate (output from \code{st_k*})}
    \item{cont}{vector of contour levels. Default is c(25,50,75). Used only for \code{level="density", "quantile"}.}
    \item{n}{number of contour levels. Default is 3. Used only for \code{level="length", "natural"}.}
    \item{type}{type of contour levels: one of \code{"density", "length", "quantile", "natural"}. Default is "density".}
    \item{group}{flag to compute contour levels per group. Default is FALSE.}
    \item{breaks}{tibble or vector of contour levels (e.g. output from \code{contour_breaks})}
    \item{disjoint}{flag to compute disjoint contours. Default is TRUE.}
    \item{digits}{number of significant digits in output labels. If missing, default is 4.}  
    \item{...}{parameters to \code{contour_breaks} (only for \code{add_contour_breaks})}    
}

\value{
    --The output from \code{contour_breaks} is a tibble of the values of the contour breaks. If \code{group=FALSE} then a single set of contour breaks is returned. If \code{group=TRUE} then a set of contour breaks for each group is returned. 

    --The output from \code{st_get_contour} is an sf object of the contours as multipolygons, with added attributes
    \item{estimate}{density value or factor with \code{digits} s.f.}
    \item{contlabel}{factor of contour probability level}
    \item{contperc}{factor of contour probability level with added \%}
    \item{contregion}{factor of contour probability level with added \eqn{\leq}{<=} or \eqn{\geq}{>=}}
    These are used to render the contours in the base \proglang{R} base graphics, and via the aesthetics in \code{ggplot}. 
}

\details{
    By default, the 5\%, 10\%, ..., 95\% contours are computed for an \code{st_k*} output, though a plot of 19 of them would be too crowded. \code{st_get_contour} selects a subset of these, as specified by \code{cont}. If a contour level in \code{cont} does not already exist or if absolute contour levels are specified in \code{breaks}, then the corresponding contours are computed. If \code{disjoint=TRUE} (default) then the contours are computed as a set of disjoint multipolygons: this allows for plotting without overlapping transparent colours. If \code{disjoint=FALSE} then the contours are overlapping and so their colours alpha-mixed, but they strictly satisfy the probabilistic definition, e.g. a 25\% contour region is the smallest region that contains 25\% of the probability mass defined by the kernel estimate, see \code{\link{geom_contour_ks}}.

    Since these default probability contours are relative contour levels, they are not suitable for producing a contour plot with fixed contour levels across all groups. It may require trial and error to obtain a single set of contour levels which is appropriate for all groups: one possible choice is provided by \code{contour_breaks}. 
}

\seealso{\code{\link{geom_contour_ks}}}

\examples{
library(ggplot2)
theme_set(theme_bw())
data(crabs, package="MASS")
## tidy data 
crabs2 <- dplyr::select(crabs, FL, CW)
t1 <- tidy_kde(crabs2)
ggplot(t1, aes(x=FL, y=CW)) + geom_contour_filled_ks()

## add density contour region labels  
add_contour_breaks(t1, group=TRUE)

\donttest{## geospatial data
crabs2s <- sf::st_as_sf(crabs2, coords=c("FL","CW"))
s1 <- st_kde(crabs2s)
s1_cont <- st_get_contour(s1, cont=seq(10,90, by=20))

## base R plot
vc <- function(.) colorspace::sequential_hcl(palette="viridis", n=.)
plot(s1_cont[,"contperc"], pal=vc)

## ggplot
ggplot() + scale_fill_viridis_d() + 
    geom_sf(data=s1_cont, aes(fill=contperc))}
}

\keyword{smooth}

\name{tidyst_intergrid}
\alias{tidy_intergrid}
\alias{st_intergrid}
\title{Tidy and geospatial interpolated grids}

\description{
    Tidy and geospatial interpolated grids for 2-dimensional gridded data.
}

\usage{
tidy_intergrid(data, attrib, cellsize, verbose=FALSE)
st_intergrid(x, attrib, cellsize, verbose=FALSE)
}

\arguments{
    \item{data,x}{sf object, where attribute is estimate value and incomplete/irregular estimation grid is polygon geometry}
    \item{attrib}{name or position of estimate variable. Default is 1.}
    \item{cellsize}{cell size. If missing then automatically calculated from data.}
    \item{verbose}{flag for verbose output. Default is FALSE.}
}

\value{
    The input gridded data is interpolated to a complete regular rectangular grid. 
}

\details{
   Any missing grid cells are inferred so there are no gaps in the output, and the attribute value to set to 0. For any other grid cells with missing attribute values, the attributate is also set to 0. \code{tidy_overgrid}/\code{st_overgrid} is usually deployed on gridded data from third parties, where geometries are excluded/varying to reduce storage requirements, but \code{tidy_as_kde}/\code{st_as_kde} require complete regular rectangular grids.  
}

\examples{
## geospatial quasi density estimate
library(ggplot2)
data(ales_grid, package="eks")

## incomplete 1 km x 1 km grid
## ind = #individuals in grid cells 
gs <- ggplot() + theme_sf() + 
    colorspace::scale_fill_continuous_sequential(palette="Heat", breaks=seq(0,6000,by=1000))  
gs + geom_sf(data=ales_grid, aes(fill=ind))

## complete regular interpolated 1 km x 1 km grid
ales_sgrid <- st_intergrid(ales_grid, attrib="ind", cellsize=c(1000,1000))
gs + geom_sf(data=ales_sgrid, aes(fill=ind)) 

## geom_sf KDE plot
ales_skde <- st_as_kde(ales_sgrid)
ggplot(ales_skde) + theme_sf() + 
    geom_sf(data=st_get_contour(ales_skde), aes(fill=contperc))
}

\keyword{smooth}

\name{tidyst_kfs}
\alias{tidy_kfs}
\alias{st_kfs}
\title{Tidy and geospatial kernel feature significance}

\description{
  Tidy and geospatial versions of kernel feature significance for 1- and 2-dimensional data.
}

\usage{
tidy_kfs(data, ...)
st_kfs(x, ...)
}

\arguments{
    \item{data}{data frame/tibble of data values}
    \item{x}{sf object with point geometry}
    \item{...}{other parameters in \code{ks::kfs} function}
}
  
\value{
   The output from \code{tidy_kfs} has the same structure as the kernel density estimate from \code{\link{tidy_kde}}, except that all values of \code{estimate} outside of the significant curvature regions are set to zero, and the \code{label} indicates whether the corresponding \code{x,y} point is inside a significant curvature region. 
   
   The output from \code{st_kfs} has a single contour, with \code{contlabel=50}, as a multipolygon which delimits significant curvature regions. 
}
 
\details{
    A significant kernel curvature region consist of all points whose density curvature value is significantly different less than zero (i.e. forms a bump surrounding a local maximum). A Hochberg procedure is employed to control the significance level for multiple significance tests. 

    For details of the computation of the significant kernel curvature regions, see \code{?ks::kfs}. The bandwidth matrix of smoothing parameters is computed as in \code{ks::kdde(deriv_order=2)}.  
}

\examples{
\donttest{## tidy significant curvature regions
library(ggplot2)
data(hsct, package="ks")
hsct <- dplyr::as_tibble(hsct)
hsct <- dplyr::filter(hsct, PE.Ly65Mac1>0 & APC.CD45.2>0)
hsct12 <- dplyr::filter(hsct, subject==12)   
hsct12 <- dplyr::select(hsct12, PE.Ly65Mac1, APC.CD45.2)
t1 <- tidy_kde(hsct12)
t2 <- tidy_kfs(hsct12)
gt <- ggplot(t2, aes(x=PE.Ly65Mac1, y=APC.CD45.2)) + 
    geom_contour_ks(data=t1, colour="grey50", cont=seq(10,90,by=10)) 
gt + geom_contour_filled_ks(aes(fill=after_stat(contregion)), colour=1) 

## geospatial significant curvature regions
data(wa)
data(grevilleasf)
hakeoides <- dplyr::filter(grevilleasf, species=="hakeoides")
s1 <- st_kfs(hakeoides)

## base R plot
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)
plot(wa, xlim=xlim, ylim=ylim)
plot(s1, add=TRUE)

## geom_sf plot
gs <- ggplot(s1) + geom_sf(data=wa, fill=NA) + theme_sf() 
gs + geom_sf(data=st_get_contour(s1), aes(fill=contregion)) +
    coord_sf(xlim=xlim, ylim=ylim)}
}

\keyword{smooth}

\name{tidyst_ksupp}
\alias{tidy_ksupp}
\alias{st_ksupp}
\title{Tidy and geospatial kernel support estimate}

\description{
  Tidy and geospatial versions of a kernel support estimate for 2-dimensional data.
}

\usage{
tidy_ksupp(data, cont=95, convex_hull=TRUE, ...)
st_ksupp(x, cont=95, convex_hull=TRUE, ...)
}

\arguments{
    \item{data}{tidy kernel density estimate (output from \code{\link{tidy_kde}})}
    \item{x}{spatial kernel density estimate (output from \code{\link{st_kde}})}
    \item{cont}{scalar contour level. Default is 95.}
    \item{convex_hull}{flag to compute convex hull of contour region. Default is TRUE.}
    \item{...}{other parameters in \code{ks::ksupp} function}
}

\value{
    The output from \code{*_ksupp} have the same structure as the kernel density estimate from \code{\link[=st_kde]{*_kde}}, except that \code{x,y} indicate the boundary of the density support estimate (if \code{convex_hull=TRUE}) or the grid points inside the density support (if \code{convex_hull=FALSE}), rather than the complete grid points themselves.

    For \code{st_ksupp}, the density support estimate is stored as a (multi)polygon \code{sf} object.
}

\details{
    The kernel support estimate is considered to be the \code{cont}\% probability contour of the kernel density estimate, with an additional convex hull calculation if \code{convex_hull=TRUE}. For details of the computation of the kernel support estimate, see \code{?ks::ksupp}.
}

\examples{
## tidy density support estimate
library(ggplot2)
data(crabs, package="MASS")
crabs2 <- dplyr::select(crabs, FL, CW)
t1 <- tidy_kde(crabs2)
t2 <- tidy_ksupp(t1)
ggplot(t1, aes(x=FL, y=CW)) + 
    geom_contour_filled_ks(cont=c(25,50,75,95), colour="grey50") +
    geom_polygon(data=t2, aes(linetype=label), fill=NA, colour=1) +
    scale_linetype_manual(values="dashed", name="Support\nconvex hull")

## geospatial density support estimate
\donttest{data(wa)
data(grevilleasf)
hakeoides <- dplyr::filter(grevilleasf, species=="hakeoides")
s1 <- st_kde(hakeoides)
s2 <- st_ksupp(s1, cont=97.5)

## base R plot
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)
plot(wa, xlim=xlim, ylim=ylim)
plot(s1, cont=c(25,50,75,97.5), add=TRUE, border="grey50")
plot(s2, add=TRUE, lty=2, pos="bottom")

## geom_sf plot
ggplot(s1) + geom_sf(data=wa, fill=NA) + theme_sf() + 
    scale_linetype_manual(values=c("dashed"), name="Support\nconvex hull") + 
    geom_sf(data=st_get_contour(s1, cont=c(25,50,75,97.5)), aes(fill=contperc)) +
    geom_sf(data=st_get_contour(s2), aes(linetype=contperc), fill=NA, col=1) + 
    coord_sf(xlim=xlim, ylim=ylim)}
}

\keyword{smooth}

\name{tidyst_kde_boundary}
\alias{tidy_kde_boundary}
\alias{tidy_kde_truncate}
\alias{st_kde_boundary}
\alias{st_kde_truncate}
\title{Tidy and geospatial kernel density estimates with boundary and truncated kernels}

\description{
    Tidy and geospatial versions of kernel density estimates with boundary and truncated kernels for 1- and 2-dimensional data. 
}

\usage{
tidy_kde_boundary(data, ...)
tidy_kde_truncate(data, boundary, ...)
st_kde_boundary(x, ...)
st_kde_truncate(x, boundary, ...)
}

\arguments{
    \item{data}{data frame/tibble of data values}
    \item{x}{sf object with point geometry}
    \item{boundary}{data frame/sf point geometry of boundary} 
    \item{...}{other parameters in \code{ks::kde.boundary} function}
}
  
\value{
   The outputs from \code{*_kde_boundary}, \code{*_kde_truncate} have the same structure as the standard kernel density estimate from \code{\link[=st_kde]{*_kde}}. 
}

\details{
    A boundary kernel density estimate is a modification of the standard density estimate for bounded data. There are two main types: beta kernels (\code{boundary.kernel="beta"}) and linear kernels (\code{boundary.kernel="linear"}). For details of the computation of the boundary kernel estimates and of the bandwidth selector procedure, see \code{ks::kde.boundary}. Currently only rectangular boundaries are supported, as defined by \code{xmin} x \code{xmax}. 
    
    A truncated kernel density estimate is a modification of the standard density estimate for bounded data. All the probability mass outside of \code{boundary} is set to zero and re-assigned over the regions inside in the boundary. The boundary can be any polygon. For further details of the computation of the truncated kernel estimate, see \code{ks::kde.truncate}. 

    For details of the computation of the boundary kernel estimates and the truncated kernel density estimates, and of the bandwidth selector procedure, see \code{ks::kde.boundary}, \code{ks::kde.truncate}.
}

\examples{
## tidy boundary density estimates
library(ggplot2)
data(worldbank, package="ks")
worldbank <- dplyr::as_tibble(worldbank)
## percentage data is bounded on [0,100] x [0,100]
wb2 <- na.omit(worldbank[,c("internet", "ag.value")])
xmin <- c(0,0); xmax <- c(100,100)
rectb <- data.frame(x=c(0,100,100,0,0), y=c(0,0,100,100,0))

## standard density estimate
t1 <- tidy_kde(wb2)
## tidy truncated density estimate
t2 <- tidy_kde_truncate(wb2, boundary=rectb)
tt <- rbind(t1, t2, labels=c("Standard KDE", "Truncated KDE"))
tb <- contour_breaks(tt, group=FALSE)

## standard estimate exceeds boundary but not truncated estimate
gr <- geom_polygon(data=rectb, aes(x=x,y=y), inherit.aes=FALSE, fill=NA, 
    colour=1, linetype="dashed")
gt <- ggplot(tt, aes(x=internet, y=ag.value)) 
gt + geom_contour_filled_ks(colour=1, breaks=tb) + gr + facet_wrap(~group)

\donttest{## linear boundary density estimate
## beta boundary density estimate
t3 <- tidy_kde_boundary(wb2, boundary.kernel="beta", xmin=xmin, xmax=xmax)
t4 <- tidy_kde_boundary(wb2, boundary.kernel="linear", xmin=xmin, xmax=xmax)
t5 <- rbind(t1, t2, t3, t4, labels=c("Standard KDE", "Truncated KDE", 
    "Beta bd KDE", "Linear bd KDE"))
tb <- contour_breaks(t5, group=FALSE)
## standard estimate exceeds boundary but not boundary or truncated estimates
gt + geom_contour_filled_ks(data=t5, colour=1, breaks=tb) + gr + facet_wrap(~group)}

## geospatial boundary density estimates
\donttest{data(wa)
data(grevilleasf)
hakeoides <- dplyr::filter(grevilleasf, species=="hakeoides")
hakeoides_bbox <- sf::st_bbox(c(xmin=4e5, xmax=5.7e5, ymin=6.47e6, ymax=7e6), 
    crs=sf::st_crs(hakeoides))
xminb <- hakeoides_bbox[1:2]; xmaxb <- hakeoides_bbox[3:4]
hakeoides_bbox <- sf::st_as_sfc(hakeoides_bbox)
hakeoides_crop <- sf::st_filter(hakeoides, hakeoides_bbox)

s1 <- st_kde(hakeoides_crop)
s2 <- st_kde_boundary(hakeoides_crop, boundary.kernel="beta", 
    xmin=xminb, xmax=xmaxb)
s3 <- st_kde_boundary(hakeoides_crop, boundary.kernel="linear", 
    xmin=xminb, xmax=xmaxb)
## geospatial truncated density estimate    
s4 <- st_kde_truncate(x=hakeoides_crop, boundary=hakeoides_bbox)
s5 <- rbind(s1, s2, s3, s4, labels=c("Standard KDE", "Beta bd KDE", 
    "Linear bd KDE", "Truncated KDE"))

## base R plots
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)
cols <- colorspace::qualitative_hcl(n=4, palette="Set2")

layout(matrix(c(1,3,2,4), ncol=2))
plot(wa, xlim=xlim, ylim=ylim)
plot(hakeoides_bbox, add=TRUE, lty=3, lwd=2)
plot(s1, add=TRUE, border=cols[1], col=NA, legend=FALSE)

plot(wa, xlim=xlim, ylim=ylim)
plot(hakeoides_bbox, add=TRUE, lty=3, lwd=2)
plot(s2, add=TRUE, border=cols[2], col=NA, legend=FALSE)

plot(wa, xlim=xlim, ylim=ylim)
plot(hakeoides_bbox, add=TRUE, lty=3, lwd=2)
plot(s3, add=TRUE, border=cols[3], col=NA, legend=FALSE)
mapsf::mf_legend(type="symb", val=c("Standard KDE","Beta bd KDE", 
    "Linear bd KDE", "Truncated KDE"), pal=cols, cex=rep(3,4), 
    pch=rep("-",4), title="Density",  pos="bottomleft")

plot(wa, xlim=xlim, ylim=ylim)
plot(hakeoides_bbox, add=TRUE, lty=3, lwd=2)
plot(s4, add=TRUE, border=cols[4], col=NA, legend=FALSE)
layout(1)

## geom_sf plots
gs <- ggplot() + geom_sf(data=wa, fill=NA) + theme_sf() + 
    geom_sf(data=hakeoides_bbox, linetype="dotted", fill=NA) +
    colorspace::scale_colour_discrete_qualitative(palette="Set2")
gs + geom_sf(data=st_get_contour(s5), aes(colour=group), fill=NA) + 
    coord_sf(xlim=xlim, ylim=ylim) + 
    guides(colour=guide_legend(title="Density")) + facet_wrap(~group)}
}

\keyword{smooth}

\name{ales_grid}
\docType{data}
\alias{ales_grid}

\title{Gridded census data in Alès, France}

\description{
The \code{ales_grid} data set contains census data of 806 1 km x 1 km grid cells which cover the city of Alès and its surrounding regions in southern France. The grid cell size was chosen to be a compromise between ensuring privacy and providing detailed public statistics from the 2019 census.
}

\usage{data(ales_grid)
}

\format{\code{ales_grid} is an sf object, whose geometry are 1 km x 1 km square polygons in the European-wide standard EPSG:3035 (ETRS89-extended / LAEA Europe) projection. There are 31 attributes collected by the French national statistical agency (Insee), e.g. \code{ind} is the number of individuals. See the URL below for more details.
} 

\source{
Insee (2023) \href{https://www.insee.fr/fr/statistiques/7655464?sommaire=7655515}{Revenus, pauvreté et niveau de vie en 2019 - Données carroyées [In French]}. Accessed on 2025-01-16. 
}

\keyword{datasets}

\name{tidyst_plot}
\alias{ggplot.tidy_ks}
\alias{ggplot.sf_ks}
\alias{plot.sf_ks}
\alias{aes_ks}
\alias{guides_ks}
\alias{labs_ks}
\title{Plots for tidy and geospatial kernel estimates}

\description{
  Plots for tidy and geospatial kernel estimates.
}

\usage{
\method{ggplot}{tidy_ks}(data=NULL, mapping=aes(), ...)
\method{ggplot}{sf_ks}(data=NULL, mapping=aes(), ..., which_geometry="sf")
\method{plot}{sf_ks}(x, ...)
}

\arguments{
  \item{data,x}{object of class \code{tidy_ks} (output from \code{tidy_k*}) or object of class \code{sf_ks} (output from \code{st_k*})}
  \item{mapping}{default list of aesthetic mappings to use for plot.}
  \item{which_geometry}{type of geometry to display: one of \code{"sf","grid"}. Default is \code{"sf"}.}
  \item{...}{other graphics parameters. See below.}
}
  
\value{
  \code{ggplot} plot object is created. Base \R plot is sent to graphics window. 
}

\details{
    For \code{tidy_ks} objects, the \code{ggplot} method adds some default aesthetics based on derived variables in the computed kernel estimate. These are \code{aes(y=estimate, weight=ks)} (1-d) and are \code{aes(z=estimate, weight=ks)} (2-d). These derived variables computed in the tibble output from \code{tidy_k*} are: \code{estimate} is the kernel estimate value and \code{ks} is the untidy version of the kernel estimate, which is required to compute contour levels. The \code{ggplot} method also adds some default labels for the axes and grouping variable, and some default formatting for the legends. These defaults replicate the appearance of the corresponding plots from the \pkg{ks} package.  

    For \code{sf_ks} objects, the \code{ggplot} method is similar to the above method, except no default aesthetics are added. The function header for the \code{plot} method is 
    \preformatted{
    plot(x, which_geometry="sf", cont=c(25,50,75), abs_cont=breaks, breaks, 
      which_deriv_ind=1, pal, col, legend=TRUE, legend.title, digits=4, ...)}
    where
    \describe{  
      \item{\code{which_geometry}}{type of geometry to display: one of \code{c("sf", "grid")}. Default is \code{"sf"}.}
      \item{\code{cont}}{vector of percentages for contour heights}
      \item{\code{abs_cont},\code{breaks}}{vector of values for contour heights}
      \item{\code{which_deriv_ind}}{index for partial derivative for density derivative estimate. Default is 1.}
      \item{\code{pal}}{colour palette function}
      \item{\code{col}}{vector of colours}
      \item{\code{legend}}{flag to add legend. Default is TRUE. The output from \code{mapsf::mf_legend} in base \R is not as robust as the legend output in \code{ggplot2}.}
      \item{\code{legend.title}}{legend title}
      \item{\code{digits}}{number of significant digits in legend key. Default is 4.}
      \item{...}{other graphics parameters in the \code{plot} method for \code{sf} objects or for \code{mapsf::mf_legend}.}
      }    
}

\seealso{\code{\link[=st_kde]{tidy_kde,st_kde}}}

\keyword{hplot}

\name{tidyst_kquiver}
\alias{tidy_kquiver}
\alias{st_kquiver}
\title{Tidy and geospatial kernel density quiver estimate}

\description{
  Tidy and geospatial versions of a kernel density quiver estimate for 2-dimensional data.
}

\usage{
tidy_kquiver(data, thin=5, transf=1/4, neg.grad=FALSE)
st_kquiver(x, thin=5, transf=1/4, neg.grad=FALSE, scale=1)
}

\arguments{
  \item{data}{tidy kernel density gradient estimate (output from \code{\link{tidy_kdde}(deriv_order=1)})}
  \item{x}{geospatial kernel density gradient estimate (output from \code{\link{st_kdde}(deriv_order=1)})}
  \item{thin}{number to thin out estimation grid. Default is 5.}
  \item{transf}{power index in transformation. Default is 1/4.} 
  \item{neg.grad}{flag to compute arrows in negative gradient direction. Default is FALSE.}
  \item{scale}{scale factor to normalise length of arrows. Default is 1.}
}
  
\value{
    The output from \code{tidy_kquiver} has the same structure as the input kernel density gradient estimate, with the added columns \code{u,v} for the density gradient value in the \eqn{x}{x}-, \eqn{y}{y}-axis. This structure is compatible with the \code{ggquiver::geom_quiver} layer function for quiver plots. 

    Since \code{ggquiver::geom_quiver} is not compatible with \code{geom_sf} layers, the output from \code{st_kquiver} has added columns \code{lon}, \code{lat}, \code{lon_end}, \code{lat_end}, \code{len} which are required in \code{geom_segment}. 
}
 
\details{
    A kernel quiver estimate is a modification of the standard kernel density gradient estimate in \code{\link[=st_kdde]{*_kdde}} where the density derivatives are not given in the separate groups as indexed in \code{deriv_group}, but as extra columns \code{u} (for \code{deriv_group=(1,0)}) and \code{v} (for \code{deriv_group=(0,1)}). 

    The bandwidth matrix of smoothing parameters is computed as in \code{ks::kdde(deriv_order=1)}.
}

\examples{
## tidy kernel quiver estimate
\donttest{library(ggplot2)
data(crabs, package="MASS")
crabs2 <- dplyr::select(crabs, FL, CW)
t1 <- tidy_kde(crabs2)
t2 <- tidy_kdde(crabs2, deriv_order=1)
t3 <- tidy_kquiver(t2, thin=5) 
gt <- ggplot(t1, aes(x=FL, y=CW)) 
gt + geom_contour_filled_ks(alpha=0.5) + 
    ggquiver::geom_quiver(data=t3, aes(u=u, v=v), colour=1)

## geospatial kernel `quiver' estimate
data(wa)
data(grevilleasf)
hakeoides <- dplyr::filter(grevilleasf, species=="hakeoides")
s1 <- st_kde(hakeoides)
s2 <- st_kdde(hakeoides, deriv_order=1)
s3 <- st_kquiver(s2, thin=5)

## base R plot
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)
plot(wa, xlim=xlim, ylim=ylim)
plot(s1, add=TRUE, alpha=0.5, border="grey50")
plot(s3$tidy_ks$ks[[1]], thin=5, add=TRUE, display="quiver")

## geom_sf plot 
gs <- ggplot(s1) + geom_sf(data=wa, fill=NA) + theme_sf()
gs + geom_sf(data=st_get_contour(s1), aes(fill=contperc), alpha=0.5) +
   ggquiver::geom_quiver(data=s3$sf, aes(x=lon, y=lat, u=u, v=v), colour=1) + 
    coord_sf(xlim=xlim, ylim=ylim)}
}

\keyword{smooth}

\name{geom_contour_ks}
\alias{geom_contour_ks}
\alias{stat_contour_ks}
\alias{geom_contour_filled_ks}
\alias{stat_contour_filled_ks}
\alias{GeomContourKs}
\alias{StatContourKs}
\alias{GeomContourFilledKs}
\alias{StatContourFilledKs}
\title{Contour and filled contour plots for tidy kernel estimates}

\description{
    Contour and filled contour plots for tidy kernel estimates for 2-dimensional data.
}

\usage{
geom_contour_ks(mapping=NULL, data=NULL, stat="contour_ks",
    position="identity", ..., cont=c(25,50,75), contperc=TRUE, 
    breaks=NULL, digits=NULL, show.legend=NA, inherit.aes=TRUE)
stat_contour_ks(mapping=NULL, data=NULL, geom="contour_ks",
    position="identity", ..., show.legend=NA, inherit.aes=TRUE)
geom_contour_filled_ks(mapping=NULL, data=NULL, stat="contour_filled_ks",
    position="identity", ..., cont=c(25,50,75), contperc=TRUE, 
    breaks=NULL, transp_neutral=NULL, digits=NULL, show.legend=NA, 
    inherit.aes=TRUE)
stat_contour_filled_ks(mapping=NULL, data=NULL, geom="contour_filled_ks",
    position="identity", ..., show.legend=NA, inherit.aes=TRUE)
}

\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{aes()}. If specified and \code{inherit.aes = TRUE} (the
default), it is combined with the default mapping at the top level of the
plot. You must supply \code{mapping} if there is no plot mapping.}

\item{data}{The data to be displayed in this layer. There are three
options:

If \code{NULL}, the default, the data is inherited from the plot
data as specified in the call to \code{ggplot()}.

A \code{data.frame}, or other object, will override the plot
data. All objects will be fortified to produce a data frame. See
\code{fortify()} for which variables will be created.

A \code{function} will be called with a single argument,
the plot data. The return value must be a \code{data.frame}, and
will be used as the layer data. A \code{function} can be created
from a \code{formula} (e.g. \code{~ head(.x, 10)}).}

\item{stat}{The statistical transformation to use on the data for this
layer, as a string.}

\item{position}{Position adjustment, either as a string, or the result of
a call to a position adjustment function.}

\item{...}{Other arguments passed on to \code{layer()}. These are
often aesthetics, used to set an aesthetic to a fixed value, like
\code{colour="red"} or \code{size=3}. They may also be parameters
to the paired geom/stat.}

\item{cont}{Vector of contour probabilities. Default value is \code{c(25,50,75}).}

\item{contperc}{Deprecated.}

\item{transp_neutral}{Flag to set neutral colour in diverging fill scale to be transparent.}

\item{breaks}{Numeric vector to set the contour breaks e.g. output from \code{contour_breaks}. Overrides \code{cont}.}

\item{digits}{Number of significant digits in legend keys. Default is 4.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped.
\code{FALSE} never includes, and \code{TRUE} always includes.
It can also be a named logical vector to finely select the aesthetics to
display.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics,
rather than combining with them. This is most useful for helper functions
that define both data and aesthetics and shouldn't inherit behaviour from
the default plot specification, e.g. \code{borders()}.}

\item{geom}{The geometric object to use display the data.}
}

\value{
    Similar output as the standard layer functions \code{ggplot2::geom_contour}, \code{geom_contour_filled} and  \code{ggplot2::stat_contour}, \code{stat_contour_filled}. 
}

\details{
    These layer functions are modifications of the standard layer functions \code{ggplot2::geom_contour}, \code{geom_contour_filled} and  \code{ggplot2::stat_contour}, \code{stat_contour_filled}. Their usage and output are similar, except that they require a tidy kernel estimate as the input, rather than the observations themselves, and that the underlying choice of the contour levels is different. For most cases, \code{geom_contour_ks} is equivalent to \code{geom_contour(stat="contour_ks")}, and likewise for \code{geom_contour_filled_ks}.

    The choice of the contour levels are based on probability contours. A 25\% 
    contour region is the smallest region that contains 25\% of the 
    probability mass defined by the kernel estimate. Probability contours offer a more intuitive approach to selecting the contour levels that reveal the pertinent characteristics of the kernel estimates. See Chacon & Duong (2018, Chapter 2.2). They are specified by the \code{cont} parameter: the default value is \code{cont=c(25,50,75)}, which computes the upper quartile, median and lower quartile probability contours. To display legend labels with the percent symbol, \code{contperc=TRUE} is deprecated. The preferred syntax is \code{after_stat(contperc)} in the aesthetic, though for \code{geom_contour_filled_ks}, it is already the default, and can be omitted. See examples below.

    Since these probability contours are computed for each group of the grouping variable in \code{data}, then these relative contour levels are different for each group. To produce a contour plot with fixed contour levels across all groups, then these can be supplied in \code{breaks}: a possible choice is provided by \code{\link{contour_breaks}}. 
}

\examples{
library(ggplot2)
data(crabs, package="MASS")
crabs2 <- dplyr::select(crabs, FL, CW)
t1 <- tidy_kde(crabs2)
gt <- ggplot(t1, aes(x=FL, y=CW))
gt + geom_contour_filled_ks() 
gt + geom_contour_ks(aes(colour=after_stat(contperc))) 
}

\seealso{\code{\link{contour_breaks}}}

\references{
  Chacon, J.E. & Duong, T. (2018) \emph{Multivariate Kernel Smoothing
  and Its Applications}. Chapman & Hall/CRC, Boca Raton.
}

\keyword{hplot}

\name{tidyst_kdde}
\alias{tidy_kdde}
\alias{st_kdde}
\title{Tidy and geospatial kernel density derivative estimates}

\description{
  Tidy and geospatial versions of kernel density derivative estimates for 1- and 2-dimensional data.
}

\usage{
tidy_kdde(data, deriv_order=1, ...)
st_kdde(x, deriv_order=1, ...)
}

\arguments{
    \item{data}{data frame/tibble of data values}
    \item{deriv_order}{derivative order. Default is 1.}
    \item{x}{sf object with point geometry}
    \item{...}{other parameters in \code{ks::kdde} function}
}

\value{
    --For \code{tidy_kdde}, the output is an object of class \code{tidy_ks}, which is a tibble with columns: 
    \item{x}{evaluation points in x-axis (name is taken from 1st input variable in \code{data})}
    \item{y}{evaluation points in y-axis (2-d) (name is taken from 2nd input variable in \code{data})}
    \item{estimate}{kernel density derivative estimate values}
    \item{deriv_order}{derivative order (same as input)}
    \item{deriv_ind}{index of partial derivative}
    \item{ks}{first row (within each \code{group}) contains the untidy kernel estimate from \code{ks::kde}}
    \item{tks}{short object class label derived from the \code{ks} object class}
    \item{label}{long object class label}
    \item{group}{grouping variable (if grouped input) (name is taken from grouping variable in \code{data})}
    \item{deriv_group}{additional derived grouping variable on partial derivative indices.}

    --For \code{st_kdde}, the output is an object of class \code{st_ks}, which is a list with fields:
    \item{tidy_ks}{tibble of simplified output (\code{deriv_ind}, \code{ks}, \code{tks}, \code{label}, \code{group}, \code{deriv_group}) from \code{tidy_kdde}}
    \item{grid}{sf object of grid of kernel density derivative estimate values, as polygons, with attributes \code{estimate}, \code{deriv_ind}, \code{group}, \code{deriv_group} copied from the \code{tidy_ks} object}
    \item{sf}{sf object of 5\%, 10\%, ..., 95\% 
    contour regions of the kernel density derivative estimate, as multipolygons, with attributes \code{estimate}, \code{contlabel}, \code{contperc}, \code{contregion}, \code{contline} derived from the contour level, and \code{deriv_ind}, \code{group}, \code{deriv_group} copied from the \code{tidy_ks} object.}
}

\details{
    The output from \code{*_kdde} have the same structure as the kernel density estimate from \code{\link[=st_kde]{*_kde}}, except that \code{estimate} is the kernel density derivative values at the grid points, and the additional derived grouping variable \code{deriv_group} is the index of the partial derivative, e.g. "deriv (1,0)" and "deriv (0,1)" for a first order derivative for 2-d data. The output is a grouped tibble, grouped by the input grouping variable (if it exists) and by \code{deriv_group}.

    For details of the computation of the kernel density derivative estimate and the bandwidth selector procedure, see \code{?ks::kdde}.

    In \code{geom_contour_filled_ks} and \code{st_get_contour}, the 100\% contour 
    is set to the bounding box of the data, and it is plotted in a transparent (if \code{data} is not missing) or neutral colour (if \code{data} is missing). For \code{geom_contour_filled_ks}, it is difficult to avoid displaying the boundary of this bounding box , though for \code{geom_sf}, using \code{aes(linetype=contline)} achieves this.     
    
    See \code{\link[=st_kde]{*_kde}} for details on the aesthetics based on \code{contperc}, \code{contregion}, \code{estimate}.


}

\examples{
library(ggplot2)
data(crabs, package="MASS")
## 1-d density curvature estimate
crabs1 <- dplyr::select(crabs, FL)
t1 <- tidy_kdde(crabs1, deriv_order=2)
gt1 <- ggplot(t1, aes(x=FL))
gt1 + geom_line(colour=1) 

## 2-d density gradient estimate
\donttest{crabs2 <- dplyr::select(crabs, FL, CW)
t2 <- tidy_kdde(crabs2, deriv_order=1)
tb1 <- contour_breaks(t2, group=FALSE)
tb2 <- contour_breaks(t2, group=TRUE)
gt2 <- ggplot(t2, aes(x=FL, y=CW)) 

## overall set of contour breaks suitable for both derivatives
## filled contour regions with density derivative labels w/o breaks
## displays bounding box by default
gt2 + geom_contour_filled_ks(data=t2, aes(fill=after_stat(contregion)), 
    colour=1, breaks=tb1) + facet_wrap(~deriv_group)
## contour lines with density derivative labels with breaks
gt2 + geom_contour_ks(data=t2, aes(group=deriv_group, 
    colour=after_stat(estimate)), breaks=tb1) + facet_wrap(~deriv_group)

## second partial derivative f^(0,1) only
t22 <- dplyr::filter(t2, deriv_ind==2)
tb22 <- dplyr::filter(tb2, deriv_ind==2)
## filled contour regions with density derivative labels with breaks
## aes(linetype=after_stat(contline)) suppresses displaying bounding box 
gt2 + geom_contour_filled_ks(data=t22, aes(fill=after_stat(contregion),
    linetype=after_stat(contline)), colour=1, breaks=tb22)
## contour lines with density derivative labels with breaks
gt2 + geom_contour_ks(data=t22, aes(colour=after_stat(estimate)), 
    breaks=tb22) 

## geospatial density derivative estimate
data(wa)
data(grevilleasf)
hakeoides <- dplyr::filter(grevilleasf, species=="hakeoides")
s1 <- st_kdde(hakeoides, deriv_order=1)
## overall set of contour breaks suitable for both derivatives
sb1 <- contour_breaks(s1, group=FALSE)
s1_cont1 <- st_get_contour(s1, breaks=sb1)
## set of contour breaks per derivative
sb2 <- contour_breaks(s1, group=TRUE)
sb2 <- dplyr::filter(sb2, deriv_ind==2)
s1_cont2 <- st_get_contour(s1)
s1_cont2 <- dplyr::filter(s1_cont2, deriv_ind==2)

## base R plot
## filled contour regions with density derivative labels with breaks
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)
plot(wa, xlim=xlim, ylim=ylim)
plot(s1, add=TRUE, which_deriv_ind=2, breaks=sb2)

## geom_sf plot
glim <- coord_sf(xlim=xlim, ylim=ylim)
gs <- ggplot(s1) + geom_sf(data=wa, fill=NA) + theme_sf()
## filled contours with density derivative "percentage" labels w/o breaks
## aes(linetype=contline) suppresses displaying bounding box 
gs + geom_sf(data=s1_cont2, aes(fill=contregion, linetype=contline), 
    colour=1) + scale_linetype_manual(values=c(0,1)) + glim 
   
## facet wrapped geom_sf plot for each partial derivative
## filled contours with density derivative labels with breaks
gs + geom_sf(data=s1_cont1, aes(fill=contregion, linetype=contline)) +
    scale_linetype_manual(values=c(0,1)) + glim + facet_wrap(~deriv_group)}
}

\keyword{smooth}

\name{tidyst_kda}
\alias{tidy_kda}
\alias{st_kda}
\title{Tidy and geospatial kernel discrimination analysis (classification)}

\description{
  Tidy and geospatial versions of kernel discrimination analysis (classification) for 1- and 2-dimensional data.
}

\usage{
tidy_kda(data, ...)
st_kda(x, ...)
}

\arguments{
    \item{data}{grouped tibble of data values}
    \item{x}{sf object with grouping attribute and with point geometry}
    \item{...}{other parameters in \code{ks::kda} function}
}
  
\value{
    --For \code{tidy_kda}, the output is an object of class \code{tidy_ks}, which is a tibble with columns: 
    \item{x}{evaluation points in x-axis (name is taken from 1st input variable in \code{data})}
    \item{y}{evaluation points in y-axis (2-d) (name is taken from 2nd input variable in \code{data})}
    \item{estimate}{weighted kernel density estimate values}
    \item{prior_prob}{prior probabilities for each group}
    \item{ks}{first row (within each \code{group}) contains the untidy kernel estimate from \code{ks::kda}}
    \item{tks}{short object class label derived from the \code{ks} object class}
    \item{label}{estimated KDA group label at \code{(x,y)}}
    \item{group}{grouping variable (same as input).}

    --For \code{st_kda}, the output is an object of class \code{st_ks}, which is a list with fields:
    \item{tidy_ks}{tibble of simplified output (\code{ks}, \code{tks}, \code{label}, \code{group}) from \code{tidy_kda}}
    \item{grid}{sf object of grid of weighted kernel density estimate values, as polygons, with attributes \code{estimate}, \code{label}, \code{group} copied from the \code{tidy_ks} object}
    \item{sf}{sf object of 5\%, 10\% ..., 95\% contour regions of weighted kernel density estimate, as multipolygons, with attributes \code{contlabel} derived from the contour level; and \code{estimate}, \code{group} copied from the \code{tidy_ks} object.}
}

\details{
    A kernel discriminant analysis (aka classification or supervised learning) assigns each grid point to the group with the highest density value, weighted by the prior probabilities. 

    The output from \code{*_kda} have the same structure as the kernel density estimate from \code{\link[=st_kde]{*_kde}}, except that \code{estimate} is the weighted kernel density values at the grid points (weighted by \code{prior_prob}), and \code{label} becomes the KDA grouping variable that indicates to which of the groups the grid points belong. The output is a grouped tibble, grouped by the input grouping variable.

    For details of the computation of the kernel discriminant analysis and the bandwidth selector procedure, see \code{?ks::kda}. The bandwidth matrix of smoothing parameters is computed as in \code{ks::kde} per group. 
}

\examples{
## tidy discriminant analysis (classification)
library(ggplot2)
data(hsct, package="ks")
hsct2 <- dplyr::as_tibble(hsct)
hsct2 <- dplyr::filter(hsct2, PE.Ly65Mac1 >0 & APC.CD45.2>0 &
    (subject==6 | subject==12))
hsct2 <- dplyr::select(hsct2, PE.Ly65Mac1, APC.CD45.2, subject)
hsct2 <- dplyr::mutate(hsct2, subject=factor(paste0("Subject #", subject), 
    levels=c("Subject #6", "Subject #12")))
hsct2 <- dplyr::group_by(hsct2, subject)
hsct1 <-dplyr::select(hsct2, PE.Ly65Mac1, subject)

## tidy 1-d classification
t1 <- tidy_kda(hsct1) 
gt1 <- ggplot(t1, aes(x=PE.Ly65Mac1)) 
gt1 + geom_line(aes(colour=subject)) + 
    geom_rug(aes(colour=label), sides="b", linewidth=1.5)

\donttest{## tidy 2-d classification
t2 <- tidy_kda(hsct2)
gt2 <- ggplot(t2, aes(x=PE.Ly65Mac1, y=APC.CD45.2))
gt2 + geom_contour_ks(aes(colour=subject)) + 
    geom_tile(aes(fill=label), alpha=0.2) + coord_fixed()}

## geospatial classification
\donttest{data(wa)
data(grevilleasf)
grevillea_gr <- dplyr::filter(grevilleasf, species=="hakeoides" |
    species=="paradoxa")
grevillea_gr <- dplyr::mutate(grevillea_gr, species=factor(species))  
grevillea_gr <- dplyr::group_by(grevillea_gr, species)
s1 <- st_kda(grevillea_gr)
s2 <- st_ksupp(st_kde(grevillea_gr))
s1$grid <- sf::st_filter(s1$grid, sf::st_convex_hull(sf::st_combine(s2$sf)))

## base R plot
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)
plot(wa, xlim=xlim, ylim=ylim)
plot(s1, which_geometry="grid", add=TRUE, alpha=0.1, border=NA)
plot(s1, add=TRUE, legend=FALSE, lwd=2)

## geom_sf plot
gs1 <- ggplot(s1) + geom_sf(data=wa, fill=NA) + theme_sf() + 
    geom_sf(data=s1$grid, aes(fill=label), alpha=0.1, colour=NA) 
gs1 + geom_sf(data=st_get_contour(s1), aes(colour=species), fill=NA) +
    coord_sf(xlim=xlim, ylim=ylim)}
}

\keyword{smooth}

\name{tidyst_kms}
\alias{tidy_kms}
\alias{st_kms}
\title{Tidy and geospatial kernel mean shift clustering}

\description{
    Tidy and geospatial versions of a kernel mean shift clustering for 1- and 2-dimensional data.
}

\usage{
tidy_kms(data, ...)
st_kms(x, ...)
}

\arguments{
    \item{data}{data frame/tibble of data values}
    \item{x}{sf object with point geometry}
    \item{...}{other parameters in \code{ks::kms} function}
}
  
\value{
    The output from \code{*_kms} have the same structure as the kernel density estimate from \code{\link[=st_kde]{*_kde}}, except that \code{x,y} indicate the data points rather than the grid points, and \code{estimate} indicates the mean shift cluster label of the data points, rather than the density values. 
}

\details{
    Mean shift clustering is a generalisation of \eqn{k}{k}-means clustering (aka unsupervised learning) which allows for non-ellipsoidal clusters and does not require the number of clusters to be pre-specified. The mean shift clusters are determined by following the initial observations along the density gradient ascent paths to the cluster centre.

    For details of the computation and the bandwidth selection procedure of the kernel mean shift clustering, see \code{?ks::kms}. The bandwidth matrix of smoothing parameters is computed as in \code{ks::kdde(deriv_order=1)}. 
}

\examples{
## tidy 2-d mean shift clustering 
library(ggplot2) 
data(crabs, package="MASS")
crabs2 <- dplyr::select(crabs, FL, CW)
t1 <- tidy_kms(crabs2)
## convex hulls of clusters
t2 <- dplyr::group_by(t1, label)
t2 <- dplyr::slice(t2, chull(FL,CW))

gt <- ggplot(t1, aes(x=FL, y=CW)) 
gt + geom_point(aes(colour=label)) +
    geom_polygon(data=t2, aes(fill=label), alpha=0.1, col=1, linetype="dotted")

## geospatial mean shift clustering 
data(wa)
data(grevilleasf)
hakeoides <- dplyr::filter(grevilleasf, species=="hakeoides")
s1 <- st_kms(hakeoides)
## convex hulls of clusters
s2 <- dplyr::group_by(s1$sf, label)
s2 <- dplyr::summarise(s2, geometry=sf::st_combine(geometry))
s2 <- sf::st_convex_hull(s2)

## base R plot
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)
plot(wa, xlim=xlim, ylim=ylim)
plot(s1, add=TRUE, pch=16)
plot(s2, add=TRUE, lty=3, pal=function(.){
    colorspace::qualitative_hcl(n=., palette="Set2", alpha=0.15)})

## geom_sf plot
gs <- ggplot(s1) + geom_sf(data=wa, fill=NA) + theme_sf()
gs + geom_sf(data=s1$sf, aes(colour=label), alpha=0.5) + 
    geom_sf(data=s2, aes(fill=label), linetype="dotted", alpha=0.15) + 
    coord_sf(xlim=xlim, ylim=ylim) 
}


\keyword{smooth}

\name{geom_point_ks}
\alias{geom_point_ks}
\alias{stat_point_ks}
\alias{geom_rug_ks}
\alias{stat_rug_ks}
\alias{GeomPointKs}
\alias{GeomRugKs}
\alias{StatPointKs}
\alias{StatRugKs}
\title{Rug and scatter plots for tidy kernel estimates}

\description{
    Rug and scatter plots for tidy kernel estimates for 1- and 2-dimensional data. 
}

\usage{
geom_point_ks(mapping=NULL, data=NULL, stat="point_ks", position="identity", 
    ..., na.rm=FALSE, jitter=FALSE, show.legend=NA, inherit.aes=TRUE) 
stat_point_ks(mapping=NULL, data=NULL, geom="point_ks", position="identity", 
    ..., na.rm=FALSE, show.legend=NA, inherit.aes=TRUE)
geom_rug_ks(mapping=NULL, data=NULL, stat="rug_ks", position="identity", 
    ..., outside=FALSE, sides="bl", length=unit(0.03, "npc"), na.rm=FALSE, 
    jitter=FALSE, show.legend=NA, inherit.aes=TRUE) 
stat_rug_ks(mapping=NULL, data=NULL, geom="rug_ks", position="identity", 
    ..., na.rm=FALSE, show.legend=NA, inherit.aes=TRUE)   
}

\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{aes()}. If specified and \code{inherit.aes = TRUE} (the
default), it is combined with the default mapping at the top level of the
plot. You must supply \code{mapping} if there is no plot mapping.}

\item{data}{The data to be displayed in this layer. There are three
options:

If \code{NULL}, the default, the data is inherited from the plot
data as specified in the call to \code{ggplot()}.

A \code{data.frame}, or other object, will override the plot
data. All objects will be fortified to produce a data frame. See
\code{fortify()} for which variables will be created.

A \code{function} will be called with a single argument,
the plot data. The return value must be a \code{data.frame}, and
will be used as the layer data. A \code{function} can be created
from a \code{formula} (e.g. \code{~ head(.x, 10)}).}

\item{stat}{The statistical transformation to use on the data for this
layer, as a string.}

\item{position}{Position adjustment, either as a string, or the result of
a call to a position adjustment function.}

\item{...}{Other arguments passed on to \code{layer()}. These are
often aesthetics, used to set an aesthetic to a fixed value, like
\code{colour="red"} or \code{size=3}. They may also be parameters
to the paired geom/stat.}

\item{na.rm}{If \code{FALSE}, the default, missing values are removed with
a warning. If \code{TRUE}, missing values are silently removed.}

\item{jitter}{Flag to jitter data before plot. Default value is FALSE.}

\item{outside}{logical that controls whether to move the rug tassels outside of the plot area. Default is off (FALSE). You will also need to use \code{coord_cartesian(clip = "off")}. When set to TRUE, also consider changing the sides argument to "tr". See examples.}

\item{sides}{A string that controls which sides of the plot the rugs appear on.
It can be set to a string containing any of \code{"trbl"}, for top, right,
bottom, and left.}

\item{length}{A \code{grid::unit()} object that sets the length of the rug lines. Use scale expansion to avoid overplotting of data.}

\item{show.legend}{logical. Should this layer be included in the legends?
\code{NA}, the default, includes if any aesthetics are mapped.
\code{FALSE} never includes, and \code{TRUE} always includes.
It can also be a named logical vector to finely select the aesthetics to
display.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics,
rather than combining with them. This is most useful for helper functions
that define both data and aesthetics and shouldn't inherit behaviour from
the default plot specification, e.g. \code{borders()}.}

\item{geom}{The geometric object to use display the data}
}

\value{
    Similar output as the standard layer functions \code{ggplot2::geom_point}, \code{ggplot2::geom_rug} and \code{ggplot2::stat_point}.
}

\details{
    These layer functions are modifications of the standard layer functions \code{ggplot2::geom_point}, \code{ggplot2::geom_rug} and \code{ggplot2::stat_point}. Their usage and output are similar, except that they require a tidy kernel estimate as the input, rather than the observations themselves. For most cases, \code{geom_rug_ks} is equivalent to \code{geom_rug(stat="rug_ks")}, and likewise for \code{geom_point_ks}. 
}

\examples{library(ggplot2)
data(crabs, package="MASS")

## rug plot for tidy 1-d kernel density estimate
crabs1 <- dplyr::select(crabs, FL)
t1 <- tidy_kde(crabs1)
g1 <- ggplot(t1, aes(x=FL)) + geom_line()
g1 + geom_rug_ks(colour=4) 

## scatter plot for tidy 2-d kernel density estimate
crabs2 <- dplyr::select(crabs, FL, CW)
t2 <- tidy_kde(crabs2)
g2 <- ggplot(t2, aes(x=FL, y=CW)) 
g2 + geom_contour_ks(aes(colour=after_stat(estimate))) + 
    geom_point_ks(colour=1) 
}

\keyword{hplot}   

\name{eks-internal}
\alias{as_tidy_ks}
\alias{dim_ks}
\alias{unfactor}
\alias{label_percent}
\alias{tidy_ks}
\alias{scale_asymmetric}
\alias{scale_transp_neutral}
\alias{st_ks}
\alias{st_add_coordinates}
\alias{st_remove_long_segment}

\title{Internal functions in the eks library}

\description{These functions are user-level but which the user is not required to use directly.}

\value{The user is not required to use directly these outputs.}

\keyword{internal}

\name{eks-package}
\alias{eks}
\alias{eks-package}
\docType{package}
\title{eks}

\description{
    This package extends the functionality of the kernel smoothing functions from the \pkg{ks} package in base \R to the tidyverse and to GIS (Geographical Information Systems) ecosystems.     
}

\details{
    As the kernel smoothers from the \pkg{ks} package are prefixed as \code{k*}, their equivalents in \pkg{eks} are systematically named as follows:   
    \itemize{
        \item \code{tidy_k*} for 1- and 2-d tidy data 
        \item \code{st_k*} for 2-d geospatial data.
    }
      
    The output data tibbles (tidy data frames provided by the \pkg{tibble} package) from \code{tidy_k*} can be visualised within the \pkg{ggplot2} graphical interface, using the usual layer functions and the custom ones supplied in this package. These \code{tidy_k*} functions are analogous to those in the \pkg{broom} and related packages, though the latter tend to focus on tidying the summary diagnostic output from model fitting (and not on tidying the underlying estimates themselves), whereas \code{tidy_k*} are more substantive since they do compute tidy estimates.   
  
    The output simple feature geometries (provided by the \pkg{sf} package) from \code{st_k*} can be visualised in the (i) \pkg{ggplot2} graphical interface using primarily the \code{ggplot2::geom_sf} layer function, or (ii) in the base \R graphical interface using the \code{plot} method supplied in this package. These simple feature geometries can also be exported as standard geospatial formats (e.g. shapefile, GEOS geometry) for use in external GIS software such as ArcGIS and QGIS. 
}   
    
\author{ Tarn Duong }

\references{
  Chacon, J.E. & Duong, T. (2018) \emph{Multivariate Kernel Smoothing
  and Its Applications}. Chapman & Hall/CRC Press, Boca Raton. 

  Duong, T. (2025) Statistical visualisation for tidy and geospatial data in R via kernel smoothing methods in the eks package. \emph{Computational Statistics} \bold{40}, 2825--2847.
}

\keyword{package}

\name{tidyst_kroc}
\alias{tidy_kroc}
\alias{st_kroc}
\title{Tidy and geospatial kernel receiver operating characteristic (ROC) curve}

\description{
  Tidy and geospatial versions of kernel receiver operating characteristic (ROC) curve for 1- and 2-dimensional data.
}

\usage{
tidy_kroc(data1, data2, ...)
st_kroc(x1, x2, ...)
}

\arguments{
    \item{data1, data2}{data frames/tibbles of data values}
    \item{x1,x2}{sf objects with point geometry}
    \item{...}{other parameters in \code{ks::kroc} function}
}
  
\value{
    The output has the same structure as the 1-d kernel distribution estimate from \code{\link[=st_kcde]{*_kcde}}, except that \code{fpr} (\eqn{x}{x}-variable) is the false positive rate (complement of specificity) and \code{estimate} is the true positive rate (sensitivity), rather than the usual estimation grid points and cdf values. 
}

\details{
    A kernel ROC curve is a modification of the standard kernel distribution estimate where the two data samples are compared. For details of the computation and the bandwidth selection procedure of the kernel density ROC curve, see \code{?ks::kroc}. The bandwidth matrix of smoothing parameters is computed as in \code{ks::kcde} per data sample.
}

\examples{
## 2-d kernel ROC curve between unsuccessful and successful grafts
library(ggplot2)
data(hsct, package="ks")
hsct <- dplyr::as_tibble(hsct)
hsct <- dplyr::filter(hsct, PE.Ly65Mac1 >0 & APC.CD45.2>0)
hsct6 <- dplyr::filter(hsct, subject==6)   ## unsuccessful graft 
hsct6 <- dplyr::select(hsct6, PE.Ly65Mac1, APC.CD45.2)
hsct12 <- dplyr::filter(hsct, subject==12) ## successful graft 
hsct12 <- dplyr::select(hsct12, PE.Ly65Mac1, APC.CD45.2)
t1 <- tidy_kroc(data1=hsct6, data2=hsct12)
ggplot(t1, aes(x=fpr)) + geom_line(colour=1) 

## geospatial ROC curve between Grevillea species
data(wa)
data(grevilleasf)
hakeoides <- dplyr::filter(grevilleasf, species=="hakeoides")
paradoxa <- dplyr::filter(grevilleasf, species=="paradoxa")
s1 <- st_kroc(x1=hakeoides, x2=paradoxa)
ggplot(s1, aes(x=fpr)) + geom_line(colour=1) 
}

\keyword{smooth}

\name{tidyst_as_kde}
\alias{tidy_as_kde}
\alias{st_as_kde}

\title{Tidy and geospatial quasi kernel density estimates from gridded data}

\description{
    Tidy and geospatial versions of quasi kernel density estimates for 1- and 2-dimensional gridded data.
}

\usage{
tidy_as_kde(data, density, ...)
st_as_kde(x, attrib=1, density, cont, ...)
}

\arguments{
    \item{data}{data frame/tibble, where first d columns are evaluation points of an estimation grid, and last column is estimate value}
    \item{x}{sf object, where \code{attrib} is estimate value and estimation grid is rectangle polygon geometry}
    \item{attrib}{name or position of estimate variable. Default is 1.}
    \item{density}{flag to compute bona fide density. Default is TRUE.}
    \item{cont}{vector of contour probabilities. If missing, the default is seq(5, 95, by=5). Usually not required to be set explicitly by the user.}
    \item{...}{other parameters}
}

\value{
    The input gridded data should be a tidy form of a complete Cartesian product that forms a regular rectangular grid with no missing values or grid points/polygons. The input is converted to a quasi density estimate so that it can be subsequently treated like an output from \code{tidy_kde} or \code{st_kde}.  
}

\details{
    If \code{density=TRUE} then all the estimate values are non-negative. If \code{density=FALSE} then the estimate values can be positive or negative.
}

\examples{
## tidy quasi density estimate
library(ggplot2)
data(faithfuld, package="ggplot2")
t1 <- tidy_as_kde(faithfuld)
## probabilistic density contour levels
ggplot(t1, aes(x=eruptions, y=waiting)) + 
    geom_contour_filled_ks(colour=1, aes(fill=after_stat(contperc)))

## non-probabilistic contour levels in ggplot2::geom_contour_filled
ggplot(faithfuld, aes(x=eruptions, y=waiting, z=density)) +
    geom_contour_filled(bins=4) + 
    colorspace::scale_fill_discrete_sequential(palette="Heat")

## geospatial quasi density estimate
## see example in ? st_intergrid 
}

\keyword{smooth}

\name{tidyst_kde_local_test}
\alias{tidy_kde_local_test}
\alias{st_kde_local_test}
\title{Tidy and geospatial kernel density based local two-sample comparison tests}

\description{
  Tidy and geospatial versions of kernel density based local two-sample comparison tests for 1- and 2-dimensional data.
}

\usage{
tidy_kde_local_test(data1, data2, labels, ...)
st_kde_local_test(x1, x2, labels, ...)
}

\arguments{
    \item{data1, data2}{data frames/tibbles of data values}
    \item{x1,x2}{sf objects with point geometry}
    \item{labels}{flag or vector of strings for legend labels}
    \item{...}{other parameters in \code{ks::kde.local.test} function}
}
  
\value{
    The output has the same structure as the kernel density estimate from \code{\link[=st_kde]{*_kde}}, except that \code{estimate} is the difference between the density values \code{data1-data2} rather than the density values, and \code{label} becomes an indicator factor of the local comparison test result: "f1<f2" = \code{data1 < data2}, 0 = \code{data1 = data2}, "f2>f1" = \code{data1 > data2}.

    The output from \code{st_kde_local_test} has two contours, with \code{contlabel=-50} (for f1<f2) and \code{contlabel=50} (for f1>f2), as multipolygons which delimit the significant difference regions.
}

\details{
    A kernel local density based two-sample comparison is a modification of the standard kernel density estimate where the two data samples are compared. A Hochberg procedure is employed to control the significance level for multiple comparison tests. 

    For details of the computation of the kernel local density based two-sample comparison test and the bandwidth selector procedure, see \code{?ks::kde.local.test}. The bandwidth matrix of smoothing parameters is computed as in \code{ks::kde} per data sample. 

    If \code{labels} is missing, then the first sample label is taken from \code{x1}, and the second sample label from \code{x2}. If \code{labels="default"} then these are "f1" and "f2". Otherwise, they are assigned to the values of the input vector of strings.
}

\examples{
## tidy local test between unsuccessful and successful grafts
library(ggplot2)
data(hsct, package="ks")
hsct <- dplyr::as_tibble(hsct)
hsct <- dplyr::filter(hsct, PE.Ly65Mac1 >0 & APC.CD45.2>0)
hsct6 <- dplyr::filter(hsct, subject==6)   ## unsuccessful graft 
hsct6 <- dplyr::select(hsct6, PE.Ly65Mac1, APC.CD45.2)
hsct12 <- dplyr::filter(hsct, subject==12) ## successful graft 
hsct12 <- dplyr::select(hsct12, PE.Ly65Mac1, APC.CD45.2)
t1 <- tidy_kde_local_test(data1=hsct6, data2=hsct12)
gt <- ggplot(t1, aes(x=PE.Ly65Mac1, y=APC.CD45.2)) + coord_fixed() 
gt + geom_contour_filled_ks(aes(fill=after_stat(contregion)))

\donttest{t2 <- tidy_kfs(hsct12)
gt + geom_contour_ks(data=t1, aes(colour=after_stat(contregion))) +
    geom_contour_filled_ks(data=t2, aes(fill=after_stat(contregion)), colour=1) + 
    scale_fill_manual(values=7)

## geospatial local test between Grevillea species
data(wa)
data(grevilleasf)
hakeoides <- dplyr::filter(grevilleasf, species=="hakeoides")
paradoxa <- dplyr::filter(grevilleasf, species=="paradoxa")
s1 <- st_kde_local_test(x1=hakeoides, x2=paradoxa)
s2 <- st_kfs(hakeoides)

## base R plot
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)
plot(wa, xlim=xlim, ylim=ylim)
plot(s1, add=TRUE)

plot(wa, xlim=xlim, ylim=ylim)
plot(s1, col=NA, add=TRUE,
     border=colorspace::qualitative_hcl(n=2, palette="Dark2", rev=TRUE))
plot(s2, col=7, add=TRUE, pos="bottom")

## geom_sf plot
gs <- ggplot(s1) + geom_sf(data=wa, fill=NA) + theme_sf() 
gs + geom_sf(data=st_get_contour(s1), aes(fill=contregion)) +
    coord_sf(xlim=xlim, ylim=ylim)
gs + geom_sf(data=st_get_contour(s1), aes(colour=contregion), fill=NA) +
    geom_sf(data=st_get_contour(s2), aes(fill=contregion), colour=1) + 
    scale_fill_manual(values=7) + coord_sf(xlim=xlim, ylim=ylim)}
}

\keyword{smooth}

\name{tidyst_kdr}
\alias{tidy_kdr}
\alias{st_kdr}
\title{Tidy and geospatial kernel density ridge estimates}

\description{
  Tidy and geospatial versions of kernel density ridge estimates for 2-dimensional data.
}

\usage{
tidy_kdr(data, dTolerance, ...)
st_kdr(x, dTolerance, ...)
}

\arguments{
    \item{data}{data frame/tibble of data values}
    \item{x}{sf object with point geometry}
    \item{dTolerance}{tolerance parameter in \code{sf::st_simplify} for reducing complexity of density ridge}
    \item{...}{other parameters in \code{ks::kdr} function}
}

\value{
    The output from \code{*_kdr} have the same structure as the kernel density estimate from \code{\link[=st_kde]{*_kde}}, except that \code{x,y} indicate the points on the density ridge, rather than the grid points themselves, and \code{estimate} becomes NA. For \code{st_kdr}, the density ridge is stored as a multipoints \code{sf} object.
}

\details{
    A density ridge can be interpreted as the line connecting the peaks in the kernel density estimate, like for a mountain range. It can also be interpreted as the filament generalisation of 2-d principal components.  For details of the computation and the bandwidth selection procedure of the kernel density ridge estimate, see \code{?ks::kdr}. The bandwidth matrix of smoothing parameters is computed as in \code{ks::kdde(deriv_order=2)}.

    To reduce the complexity of the density ridge, a call to \code{sf::st_simplify(,dTolerance)} is made. If \code{dTolerance} is missing, then it defaults to approximately the mean distance between each pair of consecutive points in each segment of the density ridge. If \code{dTolerance=0} then this step of Ramer-Douglas-Peucker simplification is not carried out.
}

\examples{## tidy density ridge estimate
library(ggplot2)
\donttest{data(cardio, package="ks")
cardio <- dplyr::as_tibble(cardio[,c("ASTV","Mean")])
set.seed(8192)
cardio <- cardio[sample(1:nrow(cardio), round(nrow(cardio)/4,0)),]
## gridsize=c(21,21) is for illustrative purposes only
## remove for more complete KDR
t1 <- tidy_kdr(cardio, gridsize=c(21,21))
gt <- ggplot(t1, aes(x=ASTV, y=Mean))
gt + geom_point_ks(colour=3) + 
    geom_path(aes(colour=label, group=segment))} 

## geospatial density ridge estimate
data(wa)
data(grevilleasf)
hakeoides <- dplyr::filter(grevilleasf, species=="hakeoides")
## gridsize=c(21,21) is for illustrative purposes only 
## remove for more complete KDR
s1 <- st_kdr(hakeoides, gridsize=c(21,21))

## base R plot
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)
plot(wa, xlim=xlim, ylim=ylim)
plot(sf::st_geometry(hakeoides), add=TRUE, col=3, pch=16, cex=0.5)
plot(s1, add=TRUE)

## geom_sf plot
gs <- ggplot(s1) + geom_sf(data=wa, fill=NA) + theme_sf()
gs + geom_sf(data=hakeoides, colour=3, alpha=0.5) +
    geom_sf(data=s1$sf, aes(colour=contregion)) + 
    coord_sf(xlim=xlim, ylim=ylim) 
}

\keyword{smooth}

\name{theme_sf}
\alias{theme_sf}

\title{Theme for geospatial maps}

\description{
    Theme for geospatial maps. 
}

\usage{
theme_sf()
}

\details{
    A partial copy of \code{ggthemes::theme_map()}.
}

\examples{
library(ggplot2)
data(wa)
data(grevilleasf)
hakeoides <- dplyr::filter(grevilleasf, species=="hakeoides")
ggplot() + theme_sf() + geom_sf(data=wa, fill=NA) + 
    geom_sf(data=hakeoides, alpha=0.4) 
}

\keyword{hplot}

\name{tidyst_kde}
\alias{tidy_kde}
\alias{st_kde}
\title{Tidy and geospatial kernel density estimates}

\description{
    Tidy and geospatial versions of kernel density estimates for 1- and 2-dimensional data.
}

\usage{
tidy_kde(data, ...)
st_kde(x, ...)
}

\arguments{
    \item{data}{data frame/tibble of data values}
    \item{x}{sf object with point geometry}
    \item{...}{other parameters in \code{ks::kde} function}
}

\value{
    --For \code{tidy_kde}, the output is an object of class \code{tidy_ks}, which is a tibble with columns:
    \item{x}{evaluation points in x-axis (name is taken from 1st input variable in \code{data})}
    \item{y}{evaluation points in y-axis (2-d) (name is taken from 2nd input variable in \code{data})}
    \item{estimate}{kernel estimate values}
    \item{ks}{first row (within each \code{group}) contains the untidy kernel estimate from \code{ks::kde}}
    \item{tks}{short object class label derived from the \code{ks} object class}
    \item{label}{long object class label}
    \item{group}{grouping variable (if grouped input) (name is taken from grouping variable in \code{data}).}

    --For \code{st_kde}, the output is an object of class \code{st_ks}, which is a list with fields:
    \item{tidy_ks}{tibble of simplified output (\code{ks}, \code{tks}, \code{label}, \code{group}) from \code{tidy_kde}}
    \item{grid}{sf object of grid of kernel density estimate values, as polygons, with attributes \code{estimate}, \code{group} copied from the \code{tidy_ks} object}
    \item{sf}{sf object of 5\%, 10\%, ..., 95\% 
    contour regions of kernel density estimate, as multipolygons, with attributes \code{estimate}, \code{contlabel}, \code{contperc}, \code{contregion} derived from the contour level, and \code{group} copied from the \code{tidy_ks} object.}
}

\details{
    For \code{tidy_kde}, the first columns of the output tibble are copied from \code{aes(x)} (1-d) or \code{aes(x,y)} (2-d). These columns are the evaluation grid points. The \code{estimate} column is the kernel density values at these grid points. The \code{group} column is a copy of the grouping variable of the input data. The \code{ks} column is a copy of the untidy kernel estimate from \code{ks::kde}, since the calculations for the layer functions \code{\link{geom_contour_ks}}, \code{\link{geom_contour_filled_ks}} require both the observations \code{data} and the kernel estimate as a \code{kde} object. For this reason, it is advised to compute a tidy kernel estimate first and then to create a \code{ggplot} with this tidy kernel estimate as the default \code{data} in the layer. 

    For \code{st_kde}, the output list contains the field \code{tidy_ks} which is the output from \code{tidy_ks}. The field \code{grid} is the kernel estimate values, with rectangular polygons. The field \code{sf} is the 5\%, 10\%, ..., 95\% probability contour regions as multipolygons, with the derived attribute \code{contlabel = 100\%-cont}.

    The structure of the \code{tidy_kde} output is inherited from the input, i.e. if the input is a data frame/ (grouped) tibble then the output is a data frame/(grouped) tibble. Likewise for the \code{sf} object outputs for \code{st_kde}.

    The default bandwidth matrix is the unconstrained plug-in selector \code{ks::Hpi}, which is suitable for a wide range of data sets, since it is not restrained to smoothing along the coordinate axes. This produces a kernel estimate which is more representative of the data than with the default bandwidth in \code{geom_density_2d} and \code{geom_density_2d_filled}. For further details of the computation of the kernel density estimate and the bandwidth selector procedure, see \code{?ks::kde}.

    If \code{breaks} is missing, then for \code{geom_contour_ks}, \code{geom_filled_contour_ks}, the density contours at \code{cont} percent are displayed, and \code{contperc} is the default aesthetic, which shows legend keys with percentages. 
    For \pkg{eks} \eqn{\geq}{>=} 1.1.0, the aesthetic \code{label_percent(contlabel)} is deprecated in favour of \code{contperc}. The other derived quantities \code{contregion}, \code{estimate} computed via \code{stat_filled_contour_ks}, \code{stat_contour_ks}, are also suitable aesthetics and these display the contour levels and \eqn{\geq}{>=} contour levels in the legend keys respectively. For \code{geom_sf}, the default aesthetic is not suitable, and should be chosen from \code{contperc}, \code{contregion}, \code{estimate}, to produce analogous plots.  

    If \code{breaks} is not missing, then the density contours at \code{breaks} are displayed, and \code{estimate} is the default aesthetic. The other derived quantity \code{contregion} is also suitable, whereas \code{contperc} is NA so is not suitable. 

    For facetted plots, it is recommended to set the same \code{breaks} for all facets, and to use \code{estimate} or \code{contregion} as aesthetics. The same density contour percentages \code{cont} can lead to different contour levels for different densities in different facets which is not suitable for a facetted plot.
}

\examples{
## tidy density estimates
library(ggplot2)
data(crabs, package="MASS")
## tidy 1-d density estimate per species
crabs1 <- dplyr::select(crabs, FL, sp)
crabs1 <- dplyr::group_by(crabs1, sp)
t1 <- tidy_kde(crabs1)
gt1 <- ggplot(t1, aes(x=FL)) 
gt1 + geom_line(colour=1) + geom_rug_ks(colour=4) + facet_wrap(~sp)

## tidy 2-d density estimate
## suitable smoothing matrix gives bimodal estimate
crabs2 <- dplyr::select(crabs, FL, CW)
t2 <- tidy_kde(crabs2)
gt2 <- ggplot(t2, aes(x=FL, y=CW)) 

## filled contour regions with density percentage labels
gt2 + geom_contour_filled_ks(colour=1) 
## filled contour regions with density level labels
gt2 + geom_contour_filled_ks(colour=1, aes(fill=after_stat(estimate)))
gt2 + geom_contour_filled_ks(colour=1, aes(fill=after_stat(contregion)))

## contour lines only with density percentage labels
gt2 + geom_contour_ks(aes(colour=after_stat(contperc)))
## contour lines only with density level labels
gt2 + geom_contour_ks(aes(colour=after_stat(estimate)))
gt2 + geom_contour_ks(aes(colour=after_stat(contregion)))

## default smoothing in geom_density_2d_filled() gives unimodal estimate
gt3 <- ggplot(crabs2, aes(x=FL, y=CW)) 
gt3 + geom_density_2d_filled(bins=4, colour=1) +
    colorspace::scale_fill_discrete_sequential(palette="Heat") +
    guides(fill=guide_legend(title="Density", reverse=TRUE))

\donttest{## facet wrapped geom_sf plot with fixed contour levels for all facets
crabs3 <- dplyr::select(crabs, FL, CW, sex)
t3 <- tidy_kde(dplyr::group_by(crabs3, sex))
tb <- contour_breaks(t3, group=FALSE)
gt4 <- ggplot(t3, aes(x=FL, y=CW)) + facet_wrap(~sex)
## filled contours
gt4 + geom_contour_filled_ks(colour=1, breaks=tb, aes(fill=after_stat(estimate))) 
gt4 + geom_contour_filled_ks(colour=1, breaks=tb, aes(fill=after_stat(contregion))) 
## contour lines 
gt4 + geom_contour_ks(breaks=tb, aes(colour=after_stat(estimate)))
gt4 + geom_contour_ks(breaks=tb, aes(colour=after_stat(contregion))) 

## geospatial density estimate
data(wa)
data(grevilleasf)
hakeoides <- dplyr::filter(grevilleasf, species=="hakeoides")
hakeoides_coord <- data.frame(sf::st_coordinates(hakeoides))
s1 <- st_kde(hakeoides)

## base R plot
## filled contour regions with density percentage labels
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)
plot(wa, xlim=xlim, ylim=ylim)
plot(s1, add=TRUE)

## geom_sf plot
## suitable smoothing matrix gives optimally smoothed contours
gs1 <- ggplot(s1) + geom_sf(data=wa, fill=NA) + theme_sf() 
glim <- coord_sf(xlim=xlim, ylim=ylim)
## filled contour regions with density percentage labels
gs1 + geom_sf(data=st_get_contour(s1), aes(fill=contperc)) + glim 
## filled contour regions with density level labels
gs1 + geom_sf(data=st_get_contour(s1), aes(fill=estimate)) + glim
gs1 + geom_sf(data=st_get_contour(s1), aes(fill=contregion)) + glim

## default smoothing in geom_density_2d_filled() is oversmoothed
ggplot(hakeoides_coord) + geom_sf(data=wa, fill=NA) + theme_sf() + 
    colorspace::scale_fill_discrete_sequential(palette="Heat") + 
    geom_density_2d_filled(aes(x=X, y=Y), bins=4, colour=1) +
    guides(fill=guide_legend(title="Density", reverse=TRUE)) + glim
    
## facet wrapped geom_sf plot with fixed contour levels for all facets
grevillea_gr <- dplyr::filter(grevilleasf, species=="hakeoides" |
    species=="paradoxa")
grevillea_gr <- dplyr::mutate(grevillea_gr, species=factor(species))  
grevillea_gr <- dplyr::group_by(grevillea_gr, species)
s2 <- st_kde(grevillea_gr)
sb <- contour_breaks(s2, group=TRUE)

gs3 <- ggplot(s2) + geom_sf(data=wa, fill=NA) + theme_sf() +
    facet_wrap(~species)
## filled contour regions with density percentage labels w/o breaks
gs3 + geom_sf(data=st_get_contour(s2), aes(fill=contperc)) +
    glim 
## filled contour regions with density level labels with breaks
gs3 + geom_sf(data=st_get_contour(s2, breaks=sb), aes(fill=estimate)) +
    glim
gs3 + geom_sf(data=st_get_contour(s2, breaks=sb), aes(fill=contregion)) +
    glim} 

\dontrun{## export as geopackage for external GIS software
sf::write_sf(wa, dsn="grevillea.gpkg", layer="wa")
sf::write_sf(hakeoides, dsn="grevillea.gpkg", layer="hakeoides")
sf::write_sf(gs1_cont, dsn="grevillea.gpkg", layer="hakeoides_cont")
sf::write_sf(s1$grid, dsn="grevillea.gpkg", layer="hakeoides_grid")}
}

\keyword{smooth}

\name{tidyst_kcurv}
\alias{tidy_kcurv}
\alias{st_kcurv}
\title{Tidy and geospatial kernel summary density curvature estimates}

\description{
  Tidy and geospatial versions of kernel summary density curvature estimates for 2-dimensional data.
}

\usage{
tidy_kcurv(data, ...)
st_kcurv(x, ...)
}

\arguments{
  \item{data}{tidy kernel density curvature estimate (output from \code{\link{tidy_kdde}(deriv_order=2)})}
  \item{x}{geospatial density curvature estimate (output from \code{\link{st_kdde}(deriv_order=2)})}
  \item{...}{other parameters in \code{ks::kcurv} function}
}

\value{
    The output from \code{*_kcurv} have the same structure as the input kernel density curvature estimate from \code{\link[=st_kdde]{*_kdde}}, except that \code{estimate} indicates the summary curvature values rather than the density curvature values, and that \code{deriv_group} for each of the partial derivatives is collapsed into a single grouping.
}

\details{
    A kernel density summary curvature estimate is a modification of a kernel density curvature estimate where the matrix of second order partial derivative values is summarised as a scalar value. For details of the computation of the kernel density summary curvature estimate, see \code{?ks::kcurv}. The bandwidth matrix of smoothing parameters is computed as in \code{ks::kdde(deriv_order=2)}. 
}

\examples{
\donttest{## tidy kernel summary density curvature estimate
library(ggplot2)
data(crabs, package="MASS")
crabs2 <- dplyr::select(crabs, FL, CW)
t1 <- tidy_kdde(crabs2, deriv_order=2)
t2 <- tidy_kcurv(t1)
gt1 <- ggplot(t2, aes(x=FL, y=CW)) 
gt1 + geom_contour_filled_ks(colour=1) 
gt1 + geom_contour_ks(aes(colour=after_stat(contperc))) 

## geospatial kernel summary density curvature estimate
data(wa)
data(grevilleasf)
hakeoides <- dplyr::filter(grevilleasf, species=="hakeoides")
s1 <- st_kdde(hakeoides, deriv_order=2)
s2 <- st_kcurv(s1)

## base R plot
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)
plot(wa, xlim=xlim, ylim=ylim)
plot(s2, add=TRUE)

## geom_sf plot
gs1 <- ggplot(s2) + geom_sf(data=wa, fill=NA) + theme_sf()
gs1 + geom_sf(data=st_get_contour(s2), aes(fill=contperc)) +
    coord_sf(xlim=xlim, ylim=ylim)}
}

\keyword{smooth}

\name{tidyst_kde_balloon}
\alias{tidy_kde_balloon}
\alias{tidy_kde_sp}
\alias{st_kde_balloon}
\alias{st_kde_sp}
\title{Tidy and geospatial kernel density estimates with variable kernels}

\description{
    Tidy and geospatial versions of kernel density estimates with variable kernels for 2-dimensional data.
}

\usage{
tidy_kde_balloon(data, ...)
tidy_kde_sp(data, ...)
st_kde_balloon(x, ...)
st_kde_sp(x, ...)
}

\arguments{
    \item{data}{data frame/tibble of data values}
    \item{x}{sf object with point geometry}
    \item{...}{other parameters in \code{ks::kde.balloon}, \code{ks::kde.sp} functions}
}
  
\value{
    The outputs from \code{*_kde_balloon}, \code{*_kde_sp} have the same structure as the standard kernel density estimate from \code{\link[=st_kde]{*_kde}}. 
}

\details{
    A variable kernel density estimate is a modification of the standard density estimate where the bandwidth matrix is variable. There are two main types: balloon kernel estimates (\code{*_kde_balloon}) where the bandwidth varies with the grid point, and sample point kernel estimates (\code{*_kde_sp}) where the bandwidth varies with the data points. For details of the computation of the variable kernel estimates and of the bandwidth selector procedure, see \code{ks::kde.balloon}, \code{ks::kde.sp}.
}

\examples{
## tidy variable density estimates
\donttest{library(ggplot2)
data(worldbank, package="ks")
worldbank <- dplyr::as_tibble(worldbank)
wb2 <- na.omit(worldbank[,c("GDP.growth", "inflation")])
xmin <- c(-70,-25); xmax <- c(25,70)

## standard density estimate
t1 <- tidy_kde(wb2, xmin=xmin, xmax=xmax)
## sample point variable density estimate
t2 <- tidy_kde_sp(wb2, xmin=xmin, xmax=xmax)
## balloon variable density estimate
## gridsize=c(21,21) only for illustrative purposes
t3 <- tidy_kde_balloon(wb2, xmin=xmin, xmax=xmax, gridsize=c(21,21))
t4 <- rbind(t1, t2, t3, labels=c("Standard KDE","Sample point KDE","Balloon KDE"))
tb <- contour_breaks(t4, group=FALSE)
gt <- ggplot(t4, aes(x=GDP.growth, y=inflation))
gt + geom_contour_filled_ks(breaks=tb, colour=1) + facet_wrap(~group)

## geospatial variable density estimates
data(wa)
data(grevilleasf)
hakeoides <- dplyr::filter(grevilleasf, species=="hakeoides")

## standard density estimate
s1 <- st_kde(hakeoides)
## sample point variable density estimate
s2 <- st_kde_sp(hakeoides)   
s3 <- c(s1, s2, labels=c("Standard KDE","Sample point KDE"))
sb <- contour_breaks(s3, group=FALSE)
bcols <- colorspace::sequential_hcl(nrow(sb), palette="Heat", rev=TRUE)

## base R plot
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)
layout(matrix(1:2, ncol=2))
par(mar=(c(0,0,0,0)))
plot(wa, xlim=xlim, ylim=ylim)
plot(s1, add=TRUE, col=bcols[1:2], breaks=sb, legend=FALSE)
par(mar=(c(0,0,0,0)))
plot(wa, xlim=xlim, ylim=ylim)
plot(s2, add=TRUE, col=bcols, breaks=sb)
layout(1)

## geom_sf plot
gs <- ggplot(s3) + geom_sf(data=wa, fill=NA) + theme_sf()
gs + geom_sf(data=st_get_contour(s3, breaks=sb), aes(fill=estimate)) +
    coord_sf(xlim=xlim, ylim=ylim) + facet_wrap(~group)}
}

\keyword{smooth}

\name{grevilleasf}
\docType{data}
\alias{grevilleasf}
\alias{wa}
\title{Geographical locations of Grevillea plants in Western Australia}

\description{
The \code{wa} data set contains the polygon of the administrative boundary of Western Australia (excluding islands). The \code{grevillea} data set contains the locations of 22303 grevillea plants in Western Australia. 
}

\usage{data(wa)
data(grevilleasf)
}

\format{\code{wa} is an sf object, whose geometry is the polygon in the 
  EPSG:7850 (GDA2020/MGA zone 50) projection.

  \code{grevilleasf} is an sf object with 22303 rows and 2 attributes.
  Each row corresponds to an observed plant. 
  The first column is the full scientific name,
  the second is the species name.
  The geometry is the point location of the plant in the EPSG:7850 (GDA2020/MGA zone 50) projection. This is a superset of the \code{grevillea} dataset in the \pkg{ks} package. 
} 

\source{
Atlas of Living Australia (2021) \href{https://www.ala.org.au/}{Grevillea occurrence}. Accessed on 2021-08-18. 

Geoscape Australia (2021) \href{https://data.gov.au/data/dataset/wa-state-boundary-geoscape-administrative-boundaries}{WA State Boundary -- Geoscape Administrative Boundaries}. Accessed on 2021-08-18. 
}

\keyword{datasets}

\name{tidyst_kcde}
\alias{tidy_kcde}
\alias{tidy_kcopula}
\alias{st_kcde}
\title{Tidy and geospatial kernel cumulative distribution and copula estimates}

\description{
    Tidy and geospatial versions of kernel cumulative distribution estimates for 1- and 2-dimensional data, and kernel copula estimates for 2-dimensional data.
}

\usage{
tidy_kcde(data, ...)
tidy_kcopula(data, ...)
st_kcde(x, ...)
}

\arguments{
    \item{data}{data frame/tibble of data values}
    \item{x}{sf object with point geometry}
    \item{...}{other parameters in \code{ks::kcde}, \code{ks::kcopula} functions}
}

\value{
    The outputs from \code{*_kcde} have the same structure as the kernel density estimate from \code{\link[=st_kde]{*_kde}}, except that \code{estimate} indicates the cumulative distribution rather than the density values. Likewise for \code{tidy_kcopula}.
}

\details{
 	For details of the computation of the kernel distribution and copula estimates, and of the bandwidth selector procedures, see \code{?ks::kcde}, \code{?ks::kcopula}.
}

\examples{library(ggplot2)
data(crabs, package="MASS")
## tidy 1-d distribution estimate 
crabs1 <- dplyr::select(crabs, FL)
t1 <- tidy_kcde(crabs1)
gt2 <- ggplot(t1, aes(x=FL)) 
gt2 + geom_line(colour=1) + geom_rug_ks(colour=4) 

## tidy 2-d copula estimate
crabs2 <- dplyr::select(crabs, FL, CW)
t2 <- tidy_kcopula(crabs2)
gt2 <- ggplot(t2, aes(x=FL, y=CW)) 
gt2 + geom_contour_filled_ks(colour=1, cont=seq(10,90,by=10))

## geospatial distribution estimate
\donttest{data(wa)
data(grevilleasf)
paradoxa <- dplyr::filter(grevilleasf, species=="paradoxa")
paradoxa_bbox <- sf::st_bbox(c(xmin=4e5, xmax=8e5, ymin=6.4e6, ymax=6.65e6), 
    crs=sf::st_crs(paradoxa))
xminb <- paradoxa_bbox[1:2]; xmaxb <- paradoxa_bbox[3:4]
paradoxa_bbox <- sf::st_as_sfc(paradoxa_bbox)
paradoxa_crop <- sf::st_filter(paradoxa, paradoxa_bbox)
s1 <- st_kcde(paradoxa_crop, xmin=xminb, xmax=xmaxb)

## base R plot
xlim <- c(1.2e5, 1.1e6); ylim <- c(6.1e6, 7.2e6)
plot(wa, xlim=xlim, ylim=ylim)
plot(paradoxa_bbox, add=TRUE, lty=3, lwd=2)
plot(s1, add=TRUE)

## geom_sf plot
gs1 <- ggplot(s1) + geom_sf(data=wa, fill=NA) + theme_sf() 
gs1 + geom_sf(data=paradoxa_bbox, linewidth=1.2, linetype="dotted", fill=NA) +
    geom_sf(data=st_get_contour(s1), aes(fill=contperc)) +
    coord_sf(xlim=xlim, ylim=ylim)}
}

\keyword{smooth}

\name{tidyst_kdcde}
\alias{tidy_kdcde}
\alias{st_kdcde}
\title{Tidy and geospatial kernel deconvolved density estimates}

\description{
    Tidy and geospatial versions of kernel deconvolved density estimates for 1- and 2-dimensional data.
}

\usage{
tidy_kdcde(data, ...)
st_kdcde(x,...)
}

\arguments{
    \item{data}{data frame/tibble of data values}
    \item{x}{sf object with point geometry}
    \item{...}{other parameters in \code{ks::kdcde} function}
}
  
\value{
    The output from \code{*_kdcde} have the same structure as the standard kernel density estimate from \code{\link[=st_kde]{*_kde}}. 
}

\details{
    A deconvolved kernel density estimate is a modification of the standard density estimate for data observed with error. This version is based on a weighted kernel density estimate. For details of the computation of the kernel deconvolved density estimate and the bandwidth selector procedure, see \code{?ks::kdcde}. 
}

\examples{\donttest{## tidy 2-d deconvolved density estimate
library(ggplot2)
data(air, package="ks")
air <- na.omit(air[, c("time","co2","pm10")])
air <- dplyr::filter(air, time=="20:00")
air <- dplyr::select(air, co2, pm10)
## for details on computation of Sigma.air, see ?ks::kdcde
Sigma.air <- diag(c(6705.765, 957.664)) 

t1 <- tidy_kde(air)
t2 <- tidy_kdcde(air, Sigma=Sigma.air, reg=0.00021)
t3 <- rbind(t1, t2, labels=c("Standard KDE","Deconvolved KDE"))
tb <- contour_breaks(t3, cont=seq(10,90,by=20), group=FALSE)

## deconvolved estimate is more clearly bimodal than standard KDE 
gt <- ggplot(t3, aes(x=co2, y=pm10)) 
gt + geom_contour_filled_ks(colour=1) + facet_wrap(~group)
gt + geom_contour_filled_ks(colour=1, breaks=tb) + facet_wrap(~group)} 
}

\keyword{smooth}
